// Version 6.07  21.01.2011 Markus Greither
// Copyright (c) 1991-2011 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file cpxpsr.h
    \brief Header for the function parser object and related items.
*/

#ifndef __CPXPSR__
#define __CPXPSR__

#include <vector>
#include <signal.h>

using namespace std;

// ----------------------------------------------------------------------------
// Definitions for exception handling
// ----------------------------------------------------------------------------

#ifdef __BORLANDC__
#define SIGNAL_CALLING_CONVENTION _USERENTRY
#elif _MSC_VER
#define SIGNAL_CALLING_CONVENTION __cdecl
#else
#define SIGNAL_CALLING_CONVENTION
#endif

typedef void (SIGNAL_CALLING_CONVENTION *fptr)(int);

//! Reset the floating point unit in case of an overflow
void SIGNAL_CALLING_CONVENTION FPECatcher(int *);

//! Set up signal() FPE exceptions
fptr SetFPESignal();

//! Clear signal() FPE exceptions
fptr ClearFPESignal(fptr oldhndl);

// ----------------------------------------------------------------------------
// MarkerObj
// ----------------------------------------------------------------------------

//! Expression type Object.
/*!
    The definitions are used to group expressions into categories. This
    is used for function calculation speed optimisation as well as for the
    solver.
*/
class MarkerObj
{
 public:
  enum Markers
  {
    //! Function expression indicator
    FuMarker = _TCHAR('='),
    //! Number expression indicator
    ZahlMarker = _TCHAR('#'),
    //! Label indicator
    LabelMarker = _TCHAR('%'),
    //! Potential expression indicator
    PotMarker = _TCHAR('^'),
    //! Variable expression indicator
    VarMarker = _TCHAR('$')
  };
  //! Add a marker to a string, consisting of one of the markers plus index
  static void AddMarker(MGPStr &AkStr,_TCHAR Marker,int Nummer)
  {
    _TCHAR A[3];

    A[2] = 0;
    A[1] = _TCHAR(Nummer);
    A[0] = Marker;
    AkStr.AddStr(A);
  }
};

// ----------------------------------------------------------------------------
// Cpx
// ----------------------------------------------------------------------------

//! Class for complex numbers.
/*!
    Based on the std::complex class.
*/
class Cpx :public complex<double>
{
 public:
  //! Ctor
  Cpx() : complex<double>() {}
  //! Ctor
  Cpx(double r, double i) : complex<double>(r,i) {}
  //! Ctor
  Cpx(double x) : complex<double>(x) {}
  //! Copy Ctor
  Cpx(const complex<double> &num) : complex<double>(num) {}
  //! Necessary for double version of Parser: returning only real part
  operator double ()
  {
    if (imag() != 0)
      throw "Conversion not possible!";
    else
      return real();
  }
};

// ----------------------------------------------------------------------------
// ParamsObj
// ----------------------------------------------------------------------------

//! Class for function parameters.
struct ParamsObj
{
 public:
  //! Parameter names
  vector<MGPStr> Name;
  //! Parameter values
  vector<Cpx> Wert;
};

// ----------------------------------------------------------------------------
// CpxObj
// ----------------------------------------------------------------------------

//! Class for functions.
/*!
    Has separate members for the name, parameters and term of the function.
*/
class CpxObj
{
 public:

  //! Function Term, Name and Parameter strings
  MGPStr Term,Namen,ParamStr;
  //! Copy Ctor
  CpxObj(const CpxObj &Other)
  {
    Params = 0;
    Status = 0;
    *this = Other;
  }
  //! Ctor
  CpxObj()
  {
    Params = 0;
    Status = 0;
  }
  //! Dtor, deletes current parameter object
  virtual ~CpxObj()
  {
    DelParams();
  }
  //! = operator
  void operator = (const CpxObj &Other)
  {
    Term = Other.Term;
    Namen = Other.Namen;
    ParamStr = Other.ParamStr;
    Ergebnis = Other.Ergebnis;
    Status = Other.Status;
    DelParams();
  }

//! Status flags.
/*!
      These flags are used for calculation optimisation and
      generation of error messages.
*/
  enum StatusFlags
  {
    Berechnet = 1,
    DiffErg = 2,
    LaufVariable = 4,
    Modified = 8
  };

  //! Function status
  int Status;
  //! Cached function result if Status & Berechnet
  Cpx Ergebnis;
  //! Function parameter object
  ParamsObj *Params;

  //! Delete current parameter object
  void DelParams()
  {
    delete Params;
    Params = 0;
  }
  //! Create new parameter object
  void CreateParams()
  {
    DelParams();
    Params = new ParamsObj;
  }
  //! Set new parameter object
  void SetParams(ParamsObj *par)
  {
    DelParams();
    Params = par;
  }
};

// ----------------------------------------------------------------------------
// FunctionArrObj
// ----------------------------------------------------------------------------

//! Array of CpxObjs.
/*!
    This array is used for storing sets of functions. Most members are
    accessors, others are for setting / getting status flags.
*/
class FunctionArrObj : public KettObj<CpxObj>
{
  public:
   //! Return calculated flag
   bool Berechnet()
   {
     return ((GetAkEint().Status & CpxObj::Berechnet) > 0);
   }
   //! Return integration flag
   bool Diff()
   {
     return ((GetAkEint().Status & CpxObj::DiffErg) > 0);
   }
};

//! Global error variable for matherr
extern int AkErr;

// ----------------------------------------------------------------------------
// ParserObj
// ----------------------------------------------------------------------------

//! Base class for parsers
/*!
    There are members for parsing the functions without calculation
    (usage e. g. TEFSolver, TEFEdit)
*/
class ParserObj : public ZahlFormatObj, protected MarkerObj
{
 public:
  enum ParserKonsts
  {
    //! Maximum number of recursive calls
    MaxAufrufe = 20,
    //! Value for invalid function.
    InvalidFkt = 255
  };

//! Class for function errors
/*!
    Holds the line number, line position and error type information.
*/
  class FehlerRec
  {
   public:
    //! Error type
    int Art;
    //! Error character position
    unsigned int Pos;
    //! Number of function where the error occured
    int FuNummer;
    //! Flag for X error value available
    bool HasXValue;
    //! X error value
    double XVal;
    //! Ctor with initialisation
    FehlerRec(int A,unsigned int P,int FN,bool HasX = false,double XV = 0)
    {
      Art = A;
      Pos = P;
      FuNummer = FN;
      HasXValue = HasX;
      XVal = XV;
    }
    //! Ctor
    FehlerRec()
    {
      ClearError();
    }
    //! Reset error structure
    void ClearError()
    {
      Art = NoError;
      Pos = 0;
      FuNummer = 0;
      HasXValue = false;
      XVal = 0;
    }
  };

//! Structure for function definition.
/*!
    This structure is used for defining functions. Function name, arcus
    function (if available), Spreadsheet name,
    parameter count, group and description are stored here for each function.
*/
  struct FunkDefType
  {
    //! Function name
    const _TCHAR *FuName;
    //! Arcus function name (if available)
    const _TCHAR *ArcusName;
    //! Spreadsheet function name (if available)
    const _TCHAR *SpreadName;
    //! Number of parameters
    unsigned char AnzParam;
    //! Function category group
    unsigned char Mode;
    //! Function descriptions
    const _TCHAR *Desc;
  };

//! Structure for function group definition.
/*!
    To each group in ParserKats, a description is defined here.
*/
  struct FuTypType
  {
    //! Function type string
    const _TCHAR *FuTyp;
    //! Function type description string
    const _TCHAR *Desc;
  };

//! Categories for functions / operators.
/*!
    This array defines groups of functions. These groups are used for the
    function select dialog. Each defined function / operator is assigned to
    one of these groups.
*/
  enum ParserKats
  {
    FuAlle      = 0,
    FuAllgemein,
    FuBezeichner,
    FuKomplex,
    FuKonst,
    FuLaufvar,
    FuObsolete,
    FuOperator,
    FuUmwandeln,
    FuWinkel,
    FuTypAnz
  };

//!Function count per function name length.
/*!
   For each possible length of function names (1-MaxFuLength), the count of
   functions is stored here. When adding functions, these values _need_ to be
   updated !
*/
  enum FuAnzNumbers
  {
    //! Function count with 1 character
    FuAnz1 = 9,
    //! Function count with 2 characters
    FuAnz2 = 9,
    //! Function count with 3 characters
    FuAnz3 = 31,
    //! Function count with 4 characters
    FuAnz4 = 8,
    //! Function count with 5 characters
    FuAnz5 = 0,
    //! Function count with 6 characters
    FuAnz6 = 7
  };

//! Element count values.
/*!
    These constants hold the counts for available functions, function name
    length, and defined operators. MaxFuStr is computed for the FuAnzNumbers,
    the other definitions _need_ to be updated when adding operators / allow
    longer function names !
*/
  enum FuLimits
  {
    //! Maximum function index
    MaxFuStr = FuAnz1+FuAnz2+
               FuAnz3+FuAnz4+
               FuAnz5+FuAnz6-1,
    //! Maximum used function character length
    MaxFuLength = 6,
    //! Number of defined operators
    MaxOpStr    = 16
  };

//! ParserObj error flags.
/*!
    These flags are used to report errors.
*/
  enum ParserFehler
  {
    DivByZero = MaxZahlFehler,
    UnknownCommand,
    WrongSymbol,
    WrongLabel,
    BracketError,
    Domain,
    Singularity,
    UnknownError,
    Recursion,
    InvalidCommand,
    ParameterCount,
    SyntaxError,
    GraphNotFound,
    Convergence,
    ImagResult
  };
  //! Ctor
  ParserObj() 
  {
    KonstFeld = 0;
    Funktion = 0;
  }
  //! Ctor
  ParserObj(const NumListFormat &LDV)
  : ZahlFormatObj(LDV)
  {
    KonstFeld = 0;
    Funktion = 0;
  }
  //! Dtor
  virtual ~ParserObj() 
  {
    free(KonstFeld);
  }
  //! Find and return the = position in this string
  static int IstGleichPos(const _TCHAR *Zeile,bool *Fehler = 0);
  //! Skip blanks and comments
  static unsigned int LeerZNKomment(const _TCHAR *AkStr,unsigned int &p)
  {
    int tmp = p;
    unsigned int ret = LeerZNKomment(AkStr,tmp);
    p = tmp;
    return ret;
  }
  //! Skip blanks and comment at the current position
  static unsigned int LeerZNKomment(const _TCHAR *AkStr,int &p);
  //! Determine function type index in Zeile at pos p
  static unsigned char FuNummer(unsigned int &p,const _TCHAR *Zeile);
  //! Check for parenthesis error in function array
  static bool PruefeKlammern(FunctionArrObj *Funktion);
  //! Search character c in Zeile starting at p, ignore comments and labels
  static bool SucheZeichen(const _TCHAR *Zeile,unsigned int &p,_TCHAR c);
  //! Find next character out of comments and labels
  static void NextFunkZeichen(const _TCHAR *Zeile,unsigned int &p);
  //! Static array for function type descriptions
  static FuTypType FuTypArr[FuTypAnz];
  //! Static array for function definitions
  static FunkDefType FuDefArr[MaxFuStr+MaxOpStr+1];
  //! Label string
  static const _TCHAR *BezString;
  //! Find and return the beginning of the parameter string in this string
  int GetParamsAnfPos(const _TCHAR *Zeile);
  //! Read next value (inc mult. functions and operators) w/o determining type
  bool NextAkWert(const _TCHAR *AkStr,unsigned int &p);
  //! Read next value (inc multiple functions and operators) and return type
  _TCHAR AkWert(MGPStr &AkStr,unsigned int &p,bool DelComment = false,
              bool ReadOps = true,bool ResolveOps = true);
  //! Read next expression, return type, Num is function type index if function
  _TCHAR AkWertLesen(const _TCHAR *AkStr,unsigned int &p,unsigned char &Num);
  //! Reset functions array to calculate another function
  void ResetFunktions();
  //! Set new functions array and initialize
  void InitFunktions(FunctionArrObj *Funk);
  //! Determine highest function index with function of name at current pos
  int HoleLabelFuNummer(unsigned int &p);
  //! Determine highest function index with function name AkStr
  int BezeichnerFuNummer(const _TCHAR *AkStr);
  //! Read parameter names out of ParamStr into Par
  bool GetParamsNamen(const _TCHAR *ParamStr = 0,ParamsObj *Par = 0);
  //! Determine if AkStr[p] is operator and evaluate it
  bool Operator(MGPStr &AkStr,unsigned int &p,
                bool DelComment = false,bool ResolveOps = true);
  //! Resolve | (parallel) operator
  bool ParallelAufloesen(MGPStr &AkStr,unsigned int PAnf, unsigned int POp,
                         unsigned int &p);
  //! Resolve ^ operator
  bool PotenzAufloesen(MGPStr &AkStr,unsigned int PAnf,unsigned int POp,
                       unsigned int &p,
                       bool DelComment = false,bool ResolveOps = true);
  //! Determines error type from math library
  void GetFehler();
  //! Calculate currently selected function in functions array
  virtual bool FunktionsBerechnung(Cpx &/*gz*/,double /*ix*/,Cpx /*iy*/,FehlerRec &/*iFehler*/,
                                   double /*iHi*/,bool /*iDe*/,vector<bool> * /*BerFus*/ = 0)
  {
    return false;
  }
  //! Calculate currently selected function in function array with x value
  virtual bool Ergebnis(double /*x*/,Cpx /*ywert*/,Cpx &/*Erg*/)
  {
    return false;
  }
  //! Calculate function Funktion with x value
  virtual double Ergebnis(const _TCHAR * /*Funktion*/,double /*x*/)
  {
    return 0;
  }
  //! Determine if currently selected function in array has term
  bool AkFunktionVorhanden()
  {
    return Funktion && (Funktion->GetAkEint().Term.Length() > 0);
  }

 protected:
  //! Array with counts of functions for each available function length
  static const unsigned char FuAnzArr[];
  //! Start index for functions for each available function length
  static const unsigned char FuAnfArr[];

  //! Currently set dynamic functions array
  FunctionArrObj *Funktion;
  //! Shortcut to current function term
  const _TCHAR *AkTerm;
  //! Data array for cached constants
  double *KonstFeld;
  //! Count for currently cached constants in KonstFeld
  unsigned char AkZahl;
  //! Current function position
  unsigned int p;
};

//! Function parser class.
/*!
    This class parses and computes functions stored in FunctionArrObjs.
    It is templatized, so that usage with both double and Cpx values
    is possible.
*/
template<class T> class ParserBaseObj :public ParserObj
{
 protected:
  //! Current x value
  double x;
  //! Current Hi (helper variable) value
  double Hi;
  //! Bracket counter
  int KlZaehler;
  //! Current y value, intermediate result and last value for integration
  T ywert,z,LastY;
  //! Debug and integration flags
  bool Debug,Diff;
  //! Temporary string
  MGPStr tmp;
  //! Count for recursive calls
  unsigned int AnzAufrufe;
  //! Last x value for integration
  double LastX,FirstX;

  //! Helper function for T = Cpx, returns imaginary
  inline double imag(Cpx x)
  {
    return x.imag();
  }
  //! Helper function for T = Cpx, returns real part
  inline double real(Cpx x)
  {
    return x.real();
  }
  //! Helper function for T = Cpx, returns complex<double>::arg(x)
  inline double arg(Cpx x)
  {
    return ::arg(x);
  }
  //! Helper function for T = double, returns imaginary of double
  inline double imag(double x)
  {
    return 0;
  }
  //! Helper function for T = double, returns real part of double
  inline double real(double x)
  {
    return x;
  }
  //! Helper function for T = double, returns zero
  inline double arg(double x)
  {
    return 0;
  }
  //! Numeric integration subroutine
  void Integration(unsigned int orgp, bool avg = false);
  //! Calculate next sum expression
  void SummenBerechnung();
  //! Calculate next value
  void WertBerechnung();
  //! Calculate next expression
  void AusdruckBerechnung();
  //! Call next function
  void FunktionsAufruf();
  //! Evaluate function parameters
  ParamsObj *GetParamsWerte();
  //! Find and return index of next parameter
  int GetParamIndex();
  //! Determines already calculated functions in functions array
  void GetBerechneteFunktionen(vector<bool> &BerFus);
  //! Set complex value
  T SetValue(double real,double imag)
  {
    return Cpx(real,imag);
  }
  //! Get real part of value
  double Real(T &val)
  {
    return real(val);
  }
  //! Get imaginary part of value
  double Imag(T &val)
  {
    return imag(val);
  }
  //! Convert from Cpx to current type
  T Convert(Cpx &val)
  {
    return val;
  }
  //! Complex arcus sinus
  T asin(T z)
  {
     T ret;
     if (Imag(z) == 0)
       ret = SetValue(::asin(Real(z)),0);
     else
     if (Real(z) == 0)
       ret = SetValue(0,log(Imag(z)+sqrt(Imag(z)*Imag(z)+1)));
     else
       throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
     return ret;
  }
  //! Complex arcus cosine
  T acos(T z)
  {
     T ret;
     if (Imag(z) == 0)
       ret = SetValue(::acos(Real(z)),0);
     else
     if ((Real(z) == 0) && (Imag(z) >= 1))
       ret = SetValue(0,log(Imag(z)+sqrt(Imag(z)*Imag(z))-1));
     else
       throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
     return ret;
  }
  //! Complex arcus tangens
  T atan(T z)
  {
    T ret;
    if (Imag(z) == 0)
      ret = SetValue(::atan(Real(z)),0);
    else
    if ((Real(z) == 0) && (abs(Imag(z)) < 1))
      ret = SetValue(Real(z),0.5*log((1+Imag(z))/(1-Imag(z))));
    else
      throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
    return ret;
  }
  //! Numeric differentiation subroutine
  void Differentiation()
  {
    if (!Diff)
    {
      Diff = true;
      if (!Funktion->Diff())
      { // Start value of differential equation
        LastY = z; // save y value for next point
        LastX = x; // Start value x
        (*Funktion)->Status |= CpxObj::DiffErg;
        z = SetValue(0,0);
      }
      else
      {
        T tmp = z;
        z = (z-LastY)/(x-LastX);
        LastY = tmp;
        LastX = x;
      }
    }
    else // only one differentation or integration per calculation
      throw FehlerRec(InvalidCommand,p,Funktion->AkEintNummer());
  }

 public:
  //! Ctor
  ParserBaseObj() : ParserObj() {}
  //! Ctor, sets number format
  ParserBaseObj(const NumListFormat &LDV) : ParserObj(LDV) {}

  //! Calculate currently selected function in functions array
  bool FunktionsBerechnung(Cpx &gz,double ix,Cpx iy,FehlerRec &iFehler,
                           double iHi,bool iDe,vector<bool> *BerFus = 0);
  //! Calculate function Funktion with x value, return TRUE if no error occured
  bool Ergebnis2(const _TCHAR *Funktion,double x,double &Erg)
  {
    FehlerRec Fehler;
    Cpx ywert,ZWert;
    // dynamic allocation because auto doesn't work with DMC (?!)
    FunctionArrObj *Fkt = new FunctionArrObj;
    double Hi = 0;

    (*Fkt)->Term.SetStr(Funktion);
    InitFunktions(Fkt);
    Fehler.ClearError();
    Fkt->Begin();
    FunktionsBerechnung(ZWert,x,ywert,Fehler,Hi,false);
    delete Fkt;
    Erg = real(ZWert);
    if (Fehler.Art != NoError)
      return false;
    else
      return true;
  }
  //! Calculate currently selected function in function array with x value
  bool Ergebnis(double x,Cpx ywert,Cpx &Erg)
  {
    FehlerRec Fehler;
    double Hi = 0;

    Fehler.ClearError();
    FunktionsBerechnung(Erg,x,ywert,Fehler,Hi,false);
    return (Fehler.Art == NoError);
  }
  //! Calculate function Funktion with x value
  double Ergebnis(const _TCHAR *Funktion,double x)
  {
    double ZWert;

    if (Ergebnis2(Funktion,x,ZWert))
      return ZWert;
    else
      return 0;
  }
};

template<class T>
bool ParserBaseObj<T>::FunktionsBerechnung(Cpx &gz,double ix,Cpx iy,
                         FehlerRec &Fehler,double iHi,bool iDe,
                         vector<bool> *BerFus)
{
  int Ak;
  int AltFunk;

  Hi = iHi;
  x = ix;
  ywert = Convert(iy);
  Debug = iDe;
  AkErr = 0;
  KlZaehler = 0;
  AnzAufrufe = 0;
  AltFunk = Funktion->AkEintNummer();
  for (Ak = 0;Ak <= Funktion->GetMaxEintrag();Ak++)
  {
    Funktion->SetAkEint(Ak);
    if ((*Funktion)->Status & CpxObj::LaufVariable)
      (*Funktion)->Status &= ~CpxObj::Berechnet;
  }
  Funktion->SetAkEint(AltFunk);
  AkTerm = (*Funktion)->Term.c_str();
  p = 0;
  Diff = false;

  bool ret = true;
  fptr oldhndl = SetFPESignal();
  // Catch C++ exeptions
  try
  {
    SummenBerechnung();
    (*Funktion)->Status |= CpxObj::Berechnet;
  }
  catch (FehlerRec &Fehl)
  {
    Fehler = Fehl;
    ret = false;
  }
  // Catch C exceptions (VC)
  catch (...)
  {
    Fehler.Art = Overflow;
    Fehler.Pos = p;
    Fehler.FuNummer = Funktion->AkEintNummer();
    z = 0;
    gz = 0;
    ret = false;
  }
  ClearFPESignal(oldhndl);
  // Check for result overflow manually
  if ((real(z) > numeric_limits<double>::max()) || 
      (real(z) < -numeric_limits<double>::max()) ||
      (imag(z) > numeric_limits<double>::max()) || 
      (imag(z) < -numeric_limits<double>::max()))
  {
    Fehler.Art = Overflow;
    Fehler.Pos = p;
    Fehler.FuNummer = Funktion->AkEintNummer();
    z = 0;
    gz = 0;
    ret = false;
  }
  if (!ret)
    return false;

  gz = z;
  if ((KlZaehler != 0) || (p < _tcslen(AkTerm))) // Klammer-Fehler
  {
    if (Fehler.Art == NoError)
    {
      Fehler.Art = BracketError;
      Fehler.Pos = p;
      Fehler.FuNummer = Funktion->AkEintNummer();
    }
  }
  if (!Debug)
    Fehler.Pos = 0;
  (*Funktion)->Ergebnis = gz;
  if (BerFus)
    GetBerechneteFunktionen(*BerFus);
  return (((*Funktion)->Status & CpxObj::Modified) > 0);
}

template<class T>
void ParserBaseObj<T>::Integration(unsigned int orgp,bool avg)
{
  (*Funktion)->Status |= CpxObj::LaufVariable;
  if (!Diff)
  { // no differential equation in this line
    bool startvalueavail;
    
    LeerZNKomment(AkTerm,p);
    // Check for separator
    if (!_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                GetLDV().ListSeparator.Length()))
    {
      p += GetLDV().ListSeparator.Length(); // skip separator
      startvalueavail = true;
    }
    else
    {
//      throw FehlerRec(InvalidCommand,p,Funktion->AkEintNummer());
      startvalueavail = false;
    }
    LeerZNKomment(AkTerm,p);
    Diff = true;
    // Define start value for differential equation or integration
    if (!Funktion->Diff())
    { 
      if (startvalueavail)
      {
        LastY = z; // previous y value for next point
        ywert = z; // Start value y
      }
      else
      {
        LastY = SetValue(0,0); // previous y value for next point
        ywert = SetValue(0,0);
      }
      LastX = x; // Start value x
      FirstX = x; // remember 1. x value 
      (*Funktion)->Status |= CpxObj::DiffErg;
      SucheZeichen(AkTerm,p,')');
      if (AkTerm[p] == ')')
      {
        p++;
        KlZaehler--;
      }
    }
    else
    {
      if (avg && (LastX == FirstX)) // Correct AVG start value
        ywert = ywert*(x-FirstX);
      int i = 2; // Start with two points per value
      unsigned int HiP;
      // Store start position of term
      if (startvalueavail)
      {
        HiP = p;
        KlZaehler--;
      }
      else
        HiP = orgp;
      T Hz = 0;
      double Step;
      do
      {
        Step = (x-LastX)/i;
        double OrigX = x;
        x = LastX; // Set x to previous value
        ywert = LastY;
        // Approximate values in the steps defined by i
        for (int ZchPz = 0;ZchPz < i;ZchPz++)
        { 
          x += Step; // x += delx
          Hz = z; // Save previous result
          p = HiP; // Reset to start of term
          if (!startvalueavail)
          {
            if (AkTerm[p] == '(')
            { // brackets around parameters
              p++;
              KlZaehler++;
              SummenBerechnung();
            }
            else
              WertBerechnung();
          }
          else
          {
            KlZaehler++;
            SummenBerechnung(); // Compute term
          }
          T v1 = z*Step;
          ywert += v1; // Add range of x from x to x+delx approximated
        }
        x = OrigX; // Reset x to orignal value
        i <<= 1; // compute next smaller stepping
      }
      while ((i < 65536) && // Maximum stepping
             // additional area is maximum 1/1000 of current Y
             (abs(z)*Step > (abs(LastY) > 1e-9 ? abs(LastY)/1000 : 1e-12)) &&
             // term differs more than 1/1000 of current result
             ((abs(Hz) > abs(z)*1.001) || (abs(z) > abs(Hz)*1.001)));
      if (i >= 65536)
        throw FehlerRec(Convergence,p,Funktion->AkEintNummer(),true,x);
      z = ywert;
      LastY = z; // Store current Y for next point
      LastX = x; // Store current X for next point
    }
  }
  else // only one differential calculation per term
    throw FehlerRec(InvalidCommand,p,Funktion->AkEintNummer());
}

template<class T>
void ParserBaseObj<T>::AusdruckBerechnung()
{
  T hz;

  WertBerechnung();
  while ((AkTerm[p] == '^') || (AkTerm[p] == '|'))
  {
    switch (AkTerm[p])
    {
      case '^':
        p++;
        hz = z;
        AusdruckBerechnung();
        if ((Imag(hz) != 0) || (Imag(z) != 0))
          z = pow(hz,z);
        else
          z = SetValue(pow(Real(hz),Real(z)),0);
        if (AkErr > 0)
          GetFehler();
        break;
      case '|':
        while (AkTerm[p] == '|')
          p++;
        hz = z;
        AusdruckBerechnung();
        T v1 = (hz*z);
        T v2 = (hz+z);
        z = v1/v2;
        break;
    }
    LeerZNKomment(AkTerm,p);
  }
}

template<class T>
void ParserBaseObj<T>::SummenBerechnung()
{
  bool Negativ;
  T hz,GesZ;

  if (AnzAufrufe++ > MaxAufrufe)
  {

    z = SetValue(1,1);
    throw FehlerRec(Recursion,p,Funktion->AkEintNummer());
  }
  z = SetValue(0,0);
  GesZ = SetValue(0,0);
  do
  {
    Negativ = false;
    LeerZNKomment(AkTerm,p); // skip blanks and comments
    if (AkTerm[p] == '+')
      p++; // skip +
    else
    if (AkTerm[p] == '-')
    {
      p++;
      Negativ = true;
    } // store -
    AusdruckBerechnung();
    switch (AkTerm[p])
    {
     case '*' :
     case '/' :
     case '\\' :
     {
       do
       {
          bool Mal = (AkTerm[p] == '*');
          p++;
          hz = z;
          AusdruckBerechnung();
          if (Mal)
            z *= hz;
          else
          if ((Real(z) != 0) || (Imag(z) != 0))
            z = hz/z;
          else
            throw FehlerRec(DivByZero,p,Funktion->AkEintNummer());
       }
       while ((AkTerm[p] != 0) && ((AkTerm[p] == '*') ||
              (AkTerm[p] == '/') ||
              (AkTerm[p] == '\\')));
       break;
     }
    }
    if (Negativ)
      GesZ -= z;
    else
      GesZ += z;
    switch (AkTerm[p])
    {
      case '=':
        p++;
        hz = GesZ;
        AusdruckBerechnung();
        GesZ = SetValue((hz == z),0);
        break;
      case '<':
        p++;
        hz = GesZ;
        if (AkTerm[p] == '>')
        { // different
          p++;
          AusdruckBerechnung();
          GesZ = SetValue((hz != z),0);
        }
        else
        if (AkTerm[p] == '=')
        { // smaller or equal
          p++;
          AusdruckBerechnung();
          GesZ = SetValue((Real(hz) <= Real(z)),0);
        }
        else
        { // smaller
          AusdruckBerechnung();
          GesZ = SetValue((Real(hz) < Real(z)),0);
        }
        break;
      case '>':
        p++;
        hz = GesZ;
        if (AkTerm[p] == '=')
        { // larger or equal
          p++;
          AusdruckBerechnung();
          GesZ = SetValue((Real(hz) >= Real(z)),0);
        }
        else
        { // larger
          AusdruckBerechnung();
          GesZ = SetValue((Real(hz) > Real(z)),0);
        }
        break;
      case  0 :
      case '+':
      case '-':
      case ')':
        break; // valid operators
      default:
        if (_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                    GetLDV().ListSeparator.Length()))
          throw FehlerRec(UnknownCommand,p,Funktion->AkEintNummer());
    }
    z = SetValue(0,0);
  }
  while ((AkTerm[p] != 0) && (AkTerm[p] != ')') &&
         _tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                 GetLDV().ListSeparator.Length()));
  z = GesZ;
  if (AkTerm[p] == ')')
  {
    p++;
    KlZaehler--;
  }
  AnzAufrufe--;
}

template<class T>
void ParserBaseObj<T>::WertBerechnung()
{
  T Hz,Hz1;
  int i;
  int ZchPz,HiP;
  unsigned char FuTyp;
  bool Negativ;
  _TCHAR a;

  Negativ = false;
  LeerZNKomment(AkTerm,p);
  if (AkTerm[p] == '-')
  { // save -
    p++;
    Negativ = true;
  }
  else
  if (AkTerm[p] == '+') // skip +
    p++;
  LeerZNKomment(AkTerm,p); // skip blanks and comments
  FuTyp = InvalidFkt;
  if (_TCHAR(AkTerm[p]) == _TCHAR(FuMarker))
  { // FuMarker found
    p++;
    FuTyp = (unsigned char)(AkTerm[p]);
    p++;
    a = FuMarker;
  }
  else
  if (IST_BUCHS(AkTerm[p]))
  { // Letters found -> function start
    i = p;
    FuTyp = FuNummer(p,AkTerm);
    a = FuMarker;
    if ((Debug == false) && (FuTyp != InvalidFkt))
    {
      tmp.SetStr(_TEXT(""));
      tmp.AddNChar(AkTerm,i);
      tmp.AddChar(_TCHAR(FuMarker));
      tmp.AddChar(_TCHAR(FuTyp));
      tmp.AddStr(&AkTerm[p]);
      (*Funktion)->Term.SetStr(tmp);
      AkTerm = (*Funktion)->Term.c_str();
      p = i+2; // update counter
    }
  }
  else
    a = AkTerm[p];
  if ((IST_ZAHL(a)) || (a == '-') ||
      (a == GetLDV().DecSeparator[0]))
  { // beginning of number found
    i = p;
    FehlerRec Fehl;
    z = SetValue(Zahl(AkTerm,p,Fehl.Art),0);
    if (Fehl.Art != NoError)
    {
      Fehl.Pos = p;
      Fehl.FuNummer = Funktion->AkEintNummer();
      throw Fehl;
    }
    else
    if ((Debug == false) && (AkZahl < 255))
    { // if no debug, replace number by ZahlMarker and index
      tmp.SetStr(_TEXT(""));
      tmp.AddNChar(AkTerm,i);
      tmp.AddChar(_TCHAR(ZahlMarker));
      tmp.AddChar(_TCHAR(AkZahl));
      tmp.AddStr(&AkTerm[p]);
      KonstFeld = (double *)realloc(KonstFeld,sizeof(double)*(AkZahl+1));
      KonstFeld[AkZahl] = Real(z);
      if (AkZahl < 255)
        AkZahl++;
      (*Funktion)->Term.SetStr(tmp);
      AkTerm = (*Funktion)->Term.c_str();
      p = i+2; // update counter
    }
  }
  else
  switch (a)
  {
     case ZahlMarker :
     { // Read cached number from KonstFeld
       p++;
       if ((unsigned char)(AkTerm[p]) <= AkZahl)
         z = SetValue(KonstFeld[(unsigned char)(AkTerm[p])],0);
       else
         throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
       p++;
       break;
     }
     case LabelMarker :
     case '[' :
     { // LabelMarker or [
       FunktionsAufruf();
       break;
     }
     case FuMarker :
     {
       if (FuTyp != InvalidFkt)
       { // valid function
         unsigned int orgp = p;
         if (FuDefArr[FuTyp].AnzParam > 0)
         { // has parameters
           if (AkTerm[p] == '(')
           { // brackets around parameters
             p++;
             KlZaehler++;
             SummenBerechnung();
           }
           else
             WertBerechnung();
         }
         switch(FuTyp)
         {
          case 1 :
          case 2 :
          case 3 : //X,F,T
             z = SetValue(x,0);
             (*Funktion)->Status |= CpxObj::LaufVariable;
             break;
          case 4 : //W
             z = SetValue(x*2*M_PI,0);
             (*Funktion)->Status |= CpxObj::LaufVariable;
             break;
          case 5 ://J
             z = SetValue(0,1);
             break;
          case 6 ://Y
             z = ywert;
             (*Funktion)->Status |= CpxObj::LaufVariable;
             break;
          case 7 ://E
             z = SetValue(M_E,0);
             break;
          case 8 ://S
             Integration(orgp);
             break;
          case 9 ://D
             Differentiation();
             break;
          case FuAnz1+
          1 :  // Pi
             z = SetValue(M_PI,0);
             break;
          case FuAnz1+
          2 :  //HI
             z = SetValue(Hi,0);
             (*Funktion)->Status |= CpxObj::LaufVariable;
             break;
          case FuAnz1+
          3 :  //IM
             z = SetValue(Imag(z),0);
             break;
          case FuAnz1+
          4 :  //RE
             z = SetValue(Real(z),0);
             break;
          case FuAnz1+
          5 :  //dB
             z = SetValue(abs(z),0);
             if (Real(z)>1e-30)
               z = SetValue(log10(Real(z))*20,0);
             else
               z = SetValue(-600,0);
             break;
          case FuAnz1+
          6 :  //LN
             z = log(z);
             break;
          case FuAnz1+
          7 : //XC
              if (Imag(z) == 0)
              {
                if ((Real(z) != 0) && (x != 0))
                {
                  z = SetValue(0,-1.0/2/M_PI/Real(z)/x);
                  (*Funktion)->Status |= CpxObj::LaufVariable;
                }
                else
                  throw FehlerRec(DivByZero,p,Funktion->AkEintNummer());
              }
              else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
              break;
          case FuAnz1+
          8 :   //XL
              if (Imag(z) == 0)
              {
                z = SetValue(0,2*M_PI*Real(z)*x);
                (*Funktion)->Status |= CpxObj::LaufVariable;
              }
              else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
              break;
          case FuAnz1+
          9 :   //IF
              if (Imag(z) == 0)
              {
                Hz = z;
                if (!_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                             GetLDV().ListSeparator.Length()))
                  p += GetLDV().ListSeparator.Length();
                if (Real(z) > 0)
                { // Condition valid, first value is result
                  SummenBerechnung();
                  if (!_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                               GetLDV().ListSeparator.Length()))
                    p += GetLDV().ListSeparator.Length();
                  SucheZeichen(AkTerm,p,')');
                  if (AkTerm[p] == ')')
                  {
                    p++;
                    KlZaehler--;
                  }
                }
                else
                { // second value is result
                  SucheZeichen(AkTerm,p,GetLDV().ListSeparator[0]);
                  if (!_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                               GetLDV().ListSeparator.Length()))
                    p += GetLDV().ListSeparator.Length();
                  SummenBerechnung();
                }
              }
              else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
              break;
          case FuAnz1+FuAnz2+
          1 :  //CPX
             Hz = z;
             if (!_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                          GetLDV().ListSeparator.Length()))
               p += GetLDV().ListSeparator.Length();
             SummenBerechnung();
             z = SetValue(Real(Hz),Real(z));
             break;
          case FuAnz1+FuAnz2+
          2 :   //CDV
             Hz = z;
             if (!_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                          GetLDV().ListSeparator.Length()))
               p += GetLDV().ListSeparator.Length();
             SummenBerechnung();
             if ((Real(Hz) != 0) || (Imag(Hz) != 0))
             {
                T v2 = (Hz+z);
                z = Hz/v2;
             }
             else
                throw FehlerRec(DivByZero,p,Funktion->AkEintNummer());
             break;
          case FuAnz1+FuAnz2+
          3 :  //CPA
             HiP = KlZaehler;
             do
             {
                if (!_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                             GetLDV().ListSeparator.Length()))
                  p += GetLDV().ListSeparator.Length();
                Hz = z;
                SummenBerechnung();
                T v1 = (Hz*z);
                T v2 = (Hz+z);
                z = v1/v2;
             }
             while ((KlZaehler == HiP) &&
                    !_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                             GetLDV().ListSeparator.Length()) &&
                    (AkTerm[p] != 0));
             break;
          case FuAnz1+FuAnz2+
          4 :  //MAX
             Hz = z;
             if (!_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                          GetLDV().ListSeparator.Length()))
               p += GetLDV().ListSeparator.Length();
             SummenBerechnung();
             if (Real(Hz) > Real(z))
               z = SetValue(Real(Hz),Imag(z));
             if (Imag(Hz) > Imag(z))
               z = SetValue(Real(z),Imag(Hz));
             break;
          case FuAnz1+FuAnz2+
          5 :  //MIN
             Hz = z;
             if (!_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                          GetLDV().ListSeparator.Length()))
               p += GetLDV().ListSeparator.Length();
             SummenBerechnung();
             if (Real(Hz) < Real(z))
               z = SetValue(Real(Hz),Imag(z));
             if (Imag(Hz) < Imag(z))
               z = SetValue(Real(z),Imag(Hz));
             break;
          case FuAnz1+FuAnz2+
          6 :  //LOG
#ifdef __BORLANDC__
             z = log10(z);
#else
             z = log(z)/log(10.0);
#endif
             break;
          case FuAnz1+FuAnz2+
          7 :  //SIN
             z = sin(z);
             break;
          case FuAnz1+FuAnz2+
          8 :  //COS
             z = cos(z);
             break;
          case FuAnz1+FuAnz2+
          9 :  //TAN
#ifdef __BORLANDC__
             z = tan(z);
#else
             z = sin(z)/cos(z);
#endif
             break;
          case FuAnz1+FuAnz2+
          10:  //EXP
             z = exp(z);
             break;
          case FuAnz1+FuAnz2+
          11:  //SQT
             z = sqrt(z);
             break;
          case FuAnz1+FuAnz2+
          12:  //SQR
             Hz = z;
             z = SetValue(Real(Hz)*Real(Hz)-Imag(Hz)*Imag(Hz),2*Real(Hz)*Imag(Hz));
             break;
          case FuAnz1+FuAnz2+
          13:  //ABS
#ifdef __DMC__
             z = SetValue(abs(z),0);
#else
             z = abs(z);
#endif
             break;
          case FuAnz1+FuAnz2+
          14:  //FAC
             if ((Real(z) >= 0) && (long(Real(z)) == Real(z)) && (Imag(z) == 0))
             {
                Hz = SetValue(1,0);
                for (i = 1;(i <= long(Real(z))) && (AkErr == 0);i++)
                  Hz = SetValue(Real(Hz)*double(i),0);
                z = Hz;
             }
             else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             break;
          case FuAnz1+FuAnz2+
          15:  //SGN
             if (Imag(z) == 0)
             {
                if (Real(z) != 0)
                  z = SetValue(fabs(Real(z))/Real(z),0);
                else
                  z = SetValue(0,0);
             }
             else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             break;
          case FuAnz1+FuAnz2+
          16:  //ODD
             if ((int(Real(z)) == Real(z)) && (Imag(z) == 0))
             {
                if ((Real(z)/2 == int(Real(z)/2)) == 0)
                  z = SetValue(-1,0);
                else
                  z = SetValue(1,0);
             }
             else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             break;
          case FuAnz1+FuAnz2+
          17:  //INT
             z = SetValue(floor(Real(z)),floor(Imag(z)));
             if (Real(z) < 0)
               z = SetValue(Real(z)+1,Imag(z));
             if (Imag(z) < 0)
               z = SetValue(Real(z),Imag(z)+1);
             break;
          case FuAnz1+FuAnz2+
          18:  //CXC
             if (Imag(z) == 0)
             {
                if ((Real(z) != 0) && (x != 0))
                {
                  z = SetValue(0,double(-1.0)/2/M_PI/x/Real(z));
                  (*Funktion)->Status |= CpxObj::LaufVariable;
                }
                else
                  throw FehlerRec(DivByZero,p,Funktion->AkEintNummer());
             }
             else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             break;
          case FuAnz1+FuAnz2+
          19:  //CXL
             if (Imag(z) == 0)
             {
                z = SetValue(0,2*M_PI*x*Real(z));
                (*Funktion)->Status |= CpxObj::LaufVariable;
             }
             else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             break;
          case FuAnz1+FuAnz2+
          20:  //VAL
             ZchPz = Funktion->AkEintNummer();
             Funktion->SetAkEint((unsigned char)(Real(z)+1));
             if (Funktion->AkEintNummer() != ZchPz)
             {
                if (Diff || (Funktion->Berechnet() == false))
                {
                  HiP = p;
                  p = 0;
                  AkTerm = (*Funktion)->Term.c_str();
                  SummenBerechnung();
                  p = HiP;
                  (*Funktion)->Status  |= CpxObj::Berechnet;
                  (*Funktion)->Ergebnis = z;
                }
                else
                  z = Convert((*Funktion)->Ergebnis);
             }
             HiP = (*Funktion)->Status;
             if (HiP & (CpxObj::LaufVariable+CpxObj::Modified))
             { // transfer running variable/modified in calling function
               Funktion->SetAkEint(ZchPz);
               (*Funktion)->Status |= HiP &
                                     (CpxObj::LaufVariable+CpxObj::Modified);
             }
             else
               Funktion->SetAkEint(ZchPz);
             AkTerm = (*Funktion)->Term.c_str();
             break;
          case FuAnz1+FuAnz2+
          21:  //CEX
             z = exp(z);
             break;
          case FuAnz1+FuAnz2+
          22:  //CAB
#ifdef __DMC__
             z = SetValue(abs(z),0);
#else
             z = abs(z);
#endif
             break;
          case FuAnz1+FuAnz2+
          23:  //CTI
             z = SetValue(-arg(z)/x/2/M_PI,0);
             break;
          case FuAnz1+FuAnz2+
          24:  //CAN
             z = SetValue(arg(z)*180/M_PI,0);
             break;
          case FuAnz1+FuAnz2+
          25:  //RAD
             z = SetValue(Real(z)*M_PI/180,0);
             break;
          case FuAnz1+FuAnz2+
          26:  //DEG
             z = SetValue(Real(z)*180/M_PI,0);
             break;
          case FuAnz1+FuAnz2+
          27 : //XCF
              if ((Imag(z) == 0) &&
                  !_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                           GetLDV().ListSeparator.Length()))
              {
                Hz = z;
                p += GetLDV().ListSeparator.Length();
                SummenBerechnung();
                if ((Real(Hz) != 0) && (Real(z) != 0))
                  z = SetValue(0,-1.0/2/M_PI/Real(z)/Real(Hz));
                else
                  throw FehlerRec(DivByZero,p,Funktion->AkEintNummer());
              }
              else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
              break;
          case FuAnz1+FuAnz2+
          28 :   //XLF
              if ((Imag(z) == 0) &&
                  !_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                           GetLDV().ListSeparator.Length()))
              {
                Hz = z;
                p += GetLDV().ListSeparator.Length();
                SummenBerechnung();
                z = SetValue(0,2*M_PI*Real(z)*Real(Hz));
              }
              else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
              break;
          case FuAnz1+FuAnz2+
          29:  //ANG
             z = SetValue(arg(z)*180/M_PI,0);
             break;
          case FuAnz1+FuAnz2+
          30 : //AVG
            Integration(orgp,true);
            if (x != FirstX)
              z = z / (x-FirstX);
            break;
          case FuAnz1+FuAnz2+
          31:  //ARG
             z = SetValue(arg(z),0);
             break;
          case FuAnz1+FuAnz2+FuAnz3+
          1 :  //ASIN
             z = asin(z);
             break;
          case FuAnz1+FuAnz2+FuAnz3+
          2 :  //ACOS
             z = acos(z);
             break;
          case FuAnz1+FuAnz2+FuAnz3+
          3 :  //ATAN
             z = atan(z);
             break;
          case FuAnz1+FuAnz2+FuAnz3+
          4 :  //SINH
             z = sinh(z);
             break;
          case FuAnz1+FuAnz2+FuAnz3+
          5 :  //COSH
             z = cosh(z);
             break;
          case FuAnz1+FuAnz2+FuAnz3+
          6 :  //TANH
#ifdef __BORLANDC__
             z = tanh(z);
#else
             z = sinh(z)/cosh(z);
#endif
             break;
          case FuAnz1+FuAnz2+FuAnz3+
          7 :  //SQRT
             z = sqrt(z);
             break;
          case FuAnz1+FuAnz2+FuAnz3+
          8 :  //DIFF
             Integration(orgp);
             break;
          case FuAnz1+FuAnz2+FuAnz3+FuAnz4+FuAnz5+
          1 :  //ARSINH
             if (Imag(z) == 0)
             {
                z = SetValue(log(Real(z)+sqrt(Real(z)*Real(z)+1)),0);
             }
             else
             if (Real(z) == 0)
             {
                if ((Imag(z) < 1) && (Imag(z) > -1))
                  z = SetValue(0,::atan(Imag(z)/sqrt(1-Imag(z)*Imag(z))));
                else
                if (Imag(z) == 1)
                  z = SetValue(0,M_PI_2);
                else
                if  (Imag(z) == -1)
                  z = SetValue(0,-M_PI_2);
                else
                  throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             }
             else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             break;
          case FuAnz1+FuAnz2+FuAnz3+FuAnz4+FuAnz5+
          2 :  //ARCOSH
             if (Imag(z) == 0)
             {
                if (Real(z) >= 1)
                  z = SetValue(log(Real(z)+sqrt(Real(z)*Real(z)-1)),0);
                else
                  throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             }
             else
             if (Real(z) == 0)
             {
                if ((Imag(z) < 1) && (Imag(z) > -1))
                  z = SetValue(0,M_PI_2-::atan(Imag(z)/sqrt(1-Imag(z)*Imag(z))));
                else
                if (Imag(z) == 1)
                  z = SetValue(0,0);
                else
                if (Imag(z) == -1)
                  z = SetValue(0,M_PI);
                else
                  throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             }
             else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             break;
          case FuAnz1+FuAnz2+FuAnz3+FuAnz4+FuAnz5+
          3 :  //ARTANH
             if (Imag(z) == 0)
             {
                if ((Real(z) > -1) && (Real(z) < 1))
                  z = SetValue(0.5*log((1+Real(z))/(1-Real(z))),0);
                else
                  throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             }
             else
             if (Real(z) == 0)
                z = SetValue(0,::atan(Imag(z)));
             else
                throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
             break;
          case FuAnz1+FuAnz2+FuAnz3+FuAnz4+FuAnz5+
          4 :  //POLDEG
          case FuAnz1+FuAnz2+FuAnz3+FuAnz4+FuAnz5+
          5 :  //POLRAD
             Hz = z;
             z = SetValue(sqrt(Real(Hz)*Real(Hz)+Imag(Hz)*Imag(Hz)),0);
             if (Real(Hz) != 0)
             {
                Hz1 = SetValue(::atan(Imag(Hz)/Real(Hz)),0);

                if (Imag(Hz) < 0)
                {
                  if (Real(Hz) < 0)
                    z = SetValue(Real(z),Real(Hz1)-M_PI);
                  else
                    z = SetValue(Real(z),Real(Hz1));
                }
                else
                {
                  if (Real(Hz) < 0)
                    z = SetValue(Real(z),M_PI+Real(Hz1));
                  else
                    z = SetValue(Real(z),Real(Hz1));
                }
             }
             else
               z = SetValue(Real(z),M_PI_2);
             if (FuTyp == FuAnz1+FuAnz2+FuAnz3+FuAnz4+FuAnz5+4)
               z = SetValue(Real(z),Imag(z)*180/M_PI);
             break;
          case FuAnz1+FuAnz2+FuAnz3+FuAnz4+FuAnz5+
          6 :  //KRTDEG
             Hz = z;
             z = SetValue(cos(Imag(Hz)*M_PI/180)*Real(Hz),
                     sin(Imag(Hz)*M_PI/180)*Real(Hz));
             break;
          case FuAnz1+FuAnz2+FuAnz3+FuAnz4+FuAnz5+
          7 :  //KRTRAD
             Hz = z;
             z = SetValue(cos(Imag(Hz))*Real(Hz),sin(Imag(Hz))*Real(Hz));
             break;
          default :
             throw FehlerRec(UnknownCommand,p,Funktion->AkEintNummer());
         }
       }
       else
         throw FehlerRec(UnknownCommand,p,Funktion->AkEintNummer());
      }
      break;
     case '(' :
     {
       p++;
       HiP = KlZaehler;
       KlZaehler++;
       SummenBerechnung();
       if ((HiP != KlZaehler) &&
           _tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                   GetLDV().ListSeparator.Length()))
         throw FehlerRec(BracketError,p,Funktion->AkEintNummer());
     }
     break;
     default :
       throw FehlerRec(WrongValue,p,Funktion->AkEintNummer());
  }
  if (Negativ)
    z = SetValue(-Real(z),-Imag(z));
  LeerZNKomment(AkTerm,p);
  if (AkErr > 0)
    GetFehler();
}

template<class T>
void ParserBaseObj<T>::FunktionsAufruf()
{
  int ZchPz = Funktion->AkEintNummer();
  int HiP;

  if (AkTerm[p] == (int)LabelMarker)
  { // if LabelMarker, call function of index
    HiP = (unsigned char)AkTerm[p+1];
    p += 2;
  }
  else
  { // evaluate label
    int i = p;
    HiP = GetParamIndex(); // skip parameters and evaluate
    if (HiP >= 0)
    {
      z = Convert((*Funktion)->Params->Wert[HiP]);
      return;
    }
    p = i;
    HiP = HoleLabelFuNummer(p); // skip label and evaluate
    if (HiP == 0)
      throw FehlerRec(WrongLabel,p,ZchPz); // invalid label
    else
    { // valid label
      if ((Debug == false) && (HiP < 256))
      { // if no debug mode, replace label with LabelMarker
        tmp.SetStr(_TEXT(""));
        tmp.AddNChar(AkTerm,i);
        tmp.AddChar(_TCHAR(LabelMarker));
        tmp.AddChar(_TCHAR(HiP));
        tmp.AddStr(&AkTerm[p]);
        (*Funktion)->Term.SetStr(tmp);
        AkTerm = (*Funktion)->Term.c_str();
        p = i+2; // update counter
      }
    }
  }
  if (HiP != ZchPz)
  { // No error, other function called
    ParamsObj *Pars = GetParamsWerte(); // get parameters
    Funktion->SetAkEint(HiP); // select called function
    if (Pars)
    { // call with parameters
      (*Funktion)->Status &= ~CpxObj::Berechnet;
      size_t AnzP = Pars->Wert.size(); // save parameter length
      (*Funktion)->SetParams(Pars); // transfer parameters
      if (GetParamsNamen() &&
          (AnzP != (*Funktion)->Params->Name.size()))
        throw FehlerRec(ParameterCount,p,ZchPz);
    }
    if (Diff || (Funktion->Berechnet() == false))
    { // no differential equation or not calculated yet
      HiP = p;
      p = 0;
      int i = KlZaehler;
      KlZaehler = 0;
      AkTerm = (*Funktion)->Term.c_str();
      SummenBerechnung(); // Calculate function
      if ((KlZaehler != 0) || (p < _tcslen(AkTerm))) // Parenthesis mismatch
        throw FehlerRec(BracketError,p,Funktion->AkEintNummer());
      KlZaehler = i;
      p = HiP;
      (*Funktion)->Status |= CpxObj::Berechnet;
      (*Funktion)->Ergebnis = z;
    }
    else
      z = Convert((*Funktion)->Ergebnis);
  }
  else // error recursion
    throw FehlerRec(Recursion,0,Funktion->AkEintNummer());
  HiP = (*Funktion)->Status;
  if (HiP & (CpxObj::LaufVariable+CpxObj::Modified))
  { // copy modified / running variable flags to calling function
    Funktion->SetAkEint(ZchPz);
    (*Funktion)->Status |= HiP & (CpxObj::LaufVariable+CpxObj::Modified);
  }
  else
    Funktion->SetAkEint(ZchPz);
  AkTerm = (*Funktion)->Term.c_str();
}

template<class T>
int ParserBaseObj<T>::GetParamIndex()
{
  unsigned int AnfPos = p;

  if (!(*Funktion)->Params ||
      !(*Funktion)->Params->Name.size() ||
      (AkTerm[p] != '['))
    return -1;
  while (AkTerm[p] && (AkTerm[p] != ']'))
    p++;
  if (AkTerm[p])
    p++;
  else
    return -1;
  size_t AkP = 0;
  while (AkP < (*Funktion)->Params->Name.size())
  {
    unsigned int AkL = (*Funktion)->Params->Name[AkP].Length();
    if ((p-AnfPos == AkL) &&
        !_tcsncmp((*Funktion)->Params->Name[AkP].c_str(),&AkTerm[AnfPos],AkL))
      return AkP;
    AkP++;
  }
  return -1;
}

template<class T>
ParamsObj *ParserBaseObj<T>::GetParamsWerte()
{
  LeerZNKomment(AkTerm,p);
  if (AkTerm[p] != '(')
    return 0; // no parameters
  ParamsObj *Par = new ParamsObj;
  int AkKl = KlZaehler;
  KlZaehler++;
  do
  { // read all parameters
    p += GetLDV().ListSeparator.Length();
    SummenBerechnung();
    Par->Wert.push_back(z);
    LeerZNKomment(AkTerm,p);
  }
  while ((p < _tcslen(AkTerm)) && (KlZaehler > AkKl) &&
         !_tcsncmp(&AkTerm[p],GetLDV().ListSeparator.c_str(),
                  GetLDV().ListSeparator.Length()));
  if (AkKl != KlZaehler)
  {
    delete Par;
    return 0;
  }
  else
    return Par;
}

template<class T>
void ParserBaseObj<T>::GetBerechneteFunktionen(vector<bool> &BerFus)
{
  int Old = Funktion->AkEintNummer();
  Funktion->Begin();
  BerFus.resize(Funktion->GetMaxEintrag()+1);
  while (Funktion->NextEint())
  {
    if ((*Funktion)->Status & CpxObj::Berechnet)
      BerFus[Funktion->AkEintNummer()] = true;
  }
  Funktion->SetAkEint(Old);
}

//! Parser for complex (Cpx) numbers
typedef ParserBaseObj<Cpx> ParserCpxObj;

//! Parser for real (double) numbers
typedef ParserBaseObj<double> ParserDoubleObj;

#endif

