// Version 6.06  29.11.2009 Markus Greither
// Copyright (c) 1991-2009 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file diagramm.h
    \brief Header for the diagram class and related items
*/

#ifndef __EFDIAGRAMM__
#define __EFDIAGRAMM__

#include "diagtype.h"

// ----------------------------------------------------------------------------
// TextAlignObj
// ----------------------------------------------------------------------------

//! Object for text alignment
/*!
    This objects allows left, right, bottom, top, center (both directions)
    for text output, supports vertical text (simulates WIN32 SetTextAlign())
*/
class TextAlignObj
{
  //! Alignment settings for X and Y direction
  int AlX,AlY;
  //! Text to align
  MGPStr text;
  //! Array that containts the size of each character in text (pixels)
  ShortArrObj CharLen;
  //! Total length of the string (pixels)
  int GesLen;
  //! Height of the string (pixels)
  int Height;
  //! Flag for vertical text output
  bool Vertikal;

 public:
//! Mode flags for alignment
  enum Modes
  {
    taLEFT = 1,
    taRIGHT,
    taCENTER,
    taTOP,
    taBOTTOM
  };

  //! Ctor, sets X alignment to left and Y alignment to top as default
  TextAlignObj()
  {
    AlX = taLEFT;
    AlY = taTOP;
    Vertikal = false;
  }
  //! Changes the text alignment setting
  void SetTextAlign(int xal,int yal)
  {
    AlX = xal;
    AlY = yal;
  }
  //! Sets new text
  void SetText(const _TCHAR *Text)
  {
    text.SetStr(Text);
  }
  //! Get text
  const MGPStr &GetText()
  {
    return text;
  }
  //! Sets vertical mode
  void SetVertikal(bool vert)
  {
    Vertikal = vert;
  }
  //! Perform text size calculations based on font in given wxDC
  wxSize CalcText(wxDC &dc);
  //! Return top-left position of text
  wxPoint GetTopLeftPos(const wxPoint &refpos)
  {
    return GetTopLeftPos(refpos.x,refpos.y);
  }
  //! Return top-left position of text
  wxPoint GetTopLeftPos(int xpos,int ypos);
  //! Draw aligned text at position outx,outy
  void TextOut(wxDC &dc,const wxPoint &pos)
  {
    TextOut(dc,pos.x,pos.y);
  }
  //! Draw aligned text at position outx,outy
  void TextOut(wxDC &dc,int outx,int outy);
  //! Returns height of current text
  int GetHeight()
  {
    return Height;
  }
  //! Returns length of current text
  int GetLength()
  {
    return GesLen;
  }
  //! Return dimension of string
  wxSize GetTextSize()
  {
    return wxSize(Vertikal ? Height : GesLen,Vertikal ? GesLen : Height);
  }
};

//! Array of TextAlignObj
class TextAlignArr : public ListObj<TextAlignObj> {};

// ----------------------------------------------------------------------------
// GraphObj
// ----------------------------------------------------------------------------

//! Base graphic object
/*!
    Handles zooming.
*/
class GraphObj : public GraphBaseObj
{
 protected:
  //! Zoom factor (1.0 means 100%)
  double Zoom;
  //! Scale factor for font sizes
  float FontFakt;

 public:
  //! Screen coordinates of this object
  BildType BildBer;

  //! Set zoom and set font size factor
  void SetZoom(double Z)
  {
// Schriftgroessen-Faktor
// Theorie : (1/72 Zoll)/((0.01mm/25.4mm)) = 35.4/(96/72) = 26.4
    Zoom = Z;
    FontFakt = 13.0*Zoom;
  };
  //! Default Ctor, sets zoom to 100%
  GraphObj() {SetZoom(1);};
  //! Return true if given coordinate is within this object
  bool PunktInObject(int X,int Y);
};

// ----------------------------------------------------------------------------
// SkalObj
// ----------------------------------------------------------------------------

//! Scaling object
/*!
    Holds value range and viewport size and pos, handles line widths,
    Has methods for conversion between viewport and value range coords.
*/
class SkalObj :public GraphObj
{
 protected:
  //! Axis
  AxisDataArr *XAxis,*YAxis;

 public:
  //! Evaluate line width and set minimum width
  static int MinLinie(double min, double x)
  {
    return (min > x ? int(min) : int(x));
  }
  //! Line or point width values vor various items
  int FettLinie,RahmenLinie;
  //! Dimensions of marker lines and thin lines
  int TickLen,MarkLaenge,MarkAbst,MarkLinie,DuennLinie;
  //! Diagram units for MM_HIMETRIC, Diagram width 10cm, Diagram height 7.07cm
  void SetZoom(double Z)
  {
    GraphObj::SetZoom(Z);
    FettLinie = MinLinie(2,30.0*Zoom);
    DuennLinie = MinLinie(1,15.0*Zoom);
    RahmenLinie = FettLinie;
    TickLen = MinLinie(2,100.0*Zoom);
    MarkLaenge = MinLinie(2,150.0*Zoom);
    MarkAbst = MinLinie(6,250.0*Zoom);
    MarkLinie = MinLinie(2,38.0*Zoom);
  }
  //! Default Ctor, sets Zoom to 100%
  SkalObj()
  {
    SetZoom(1);
    XAxis = 0;
    YAxis = 0;
  }
  //! Initialize, set ranges and log flags
  void GrInit(BildType BildB,AxisDataArr &XA,AxisDataArr &YA);
  //! Converts x value Wert into x screen coordinate
  long XPos(double Wert);
  //! Converts y value Wert into y screen coordinate
  long YPos(double Wert);
  //! Converts x screen coordinate into x value
  double XWert(long Pos);
  //! Converts y screen coordinate into y value
  double YWert(long Pos);
  //! Sets the length of the value lines extending out of the diagram
  inline void SetTickLen(int len)
  {
    TickLen = len;
  }
  //! Recalculate scale factors for given new screen coordinates
  virtual void UpdateBildBer(BildType BB);
  //! Dtor
  virtual ~SkalObj();
};

// ----------------------------------------------------------------------------
// DiagObj
// ----------------------------------------------------------------------------

//! Diagram object
/*!
    Draws the diagram and labels.
*/
class DiagObj :public SkalObj,ZahlFormatObj
{
  //! Diagram screen size including labels
  BildType DiagBer;
  //! Flag for CLineTo if last point was in diagram
  bool InBer;
  //! Last point values for CLineTo
  double AkX,AkY;
  //! Labels for x and y axis, title
  TextAlignObj TText;
  //! Axis label aligned labels
  TextAlignArr XText,YText;
  //! Fonts for axis labels, axis numbers and title
  FontType Achse,Titel,AchsLabels;
  //! Diagram options
  int Mode;
  //! Number of digits for axis numbers
  int NumDigits;
  //! Points for CLineTo
#if wxCHECK_VERSION(2,9,0)
  wxPointList *Points;
#else
  wxList *Points;
#endif
  //! Flag for using splines in CLineTo
  bool splines;
  //! Array of y axis labelling x dimensions
  int xposarr[AxisDataArr::MaxValue+1];
  //! Background colour
  unsigned long m_Colour;

  //! Draws x line for diagram
  void XLinieZeichnen(wxDC &DC,int x,bool Solid);
  //! Draws y line for diagram
  void YLinieZeichnen(wxDC &DC,int y,bool Solid,int textpos,bool OppositeSide);
  //! Draws axis numbers
  void LabelOut(bool YLinien, int AkPos,const _TCHAR *Zeile,
                TextAlignObj &TextAlign,int pos,bool OppositeSide,wxDC &DC);
  //! Formats number Wert for output
  void FormatWert(double Wert,MGPStr &Zeile,bool YLinien,int Format,bool OppositeSide);
  //! Draws diagram in given direction with linear scale
  void LinLinien(AxisData &ad,bool YLinien,wxDC &DC,bool Darst,int textpos);
  //! Draws diagram in given direction in log scale
  void LogLinien(AxisData &ad,bool YLinien,wxDC &DC,bool Darst,int textpos);
  //! Get x text len for current y axis
  int GetXTextLen();
  //! Diagram title position
  wxPoint DiagTextPos()
  {
    return wxPoint(BildBer.XMin+(BildBer.XMax-BildBer.XMin)/2,
                   BildBer.YMin-TText.GetHeight()/4);
  }
  //! Draw lines or splines, delete points list
  void DrawLines(wxDC &DC);
  //! Delete list of points
  void DeletePoints();
  //! Position of X axis label text
  wxPoint XTextPos()
  {
    wxPoint pt(BildBer.XMin+(BildBer.XMax-BildBer.XMin)/2,
               BildBer.YMax+(*XAxis)->TextH+XText->GetHeight()*11/10);
    return pt;
  }
  //! Position of Y axis label text
  wxPoint YTextPos()
  {
    wxPoint pt(YText->GetHeight()/5,
               BildBer.YMin+(BildBer.YMax-BildBer.YMin)/2);
    return pt;
  }

 protected:
  //! Diagram title
  MGPStr DiagText;

 public:
  //! Default Ctor
  DiagObj();
  //! Dtor
  virtual ~DiagObj();
  //! Creates wxFont from FontType
  wxFont *GetFont(wxDC &dc,FontType Font);
  //! Initialize diagram
  void GrInit(BildType &BildB,AxisDataArr &XA,AxisDataArr &YA,const FontType &A,
              const FontType &T,const FontType &AL,const unsigned long &col,
              const NumListFormat &iLDV,int M,int ND,
              wxDC &DC);
  //! Calculates the text sizes and reduces diagram size accordingly
  void SetTextSize(wxDC &DC,const MGPStr &Zeile,bool YLinien);
  //! Draw diagram lines
  void Linien(wxDC &DC,bool Darst);
  //! Draw eventual standing out lines
  void FlushCLineTo(wxDC &DC);
  //! Draw clipped graph from AkX, AkY to X, Y
  void CLineTo(wxDC &DC,double X, double Y);
  //! Initialize CLineTo to first point of graph
  void ResetCLineTo(double X, double Y,bool spl = false);
  //! Sets line widths
  inline void SetLineWidths(int Fett,int Duenn,int Rahmen)
  {
    FettLinie = Fett;
    DuennLinie = Duenn;
    RahmenLinie = Rahmen;
  }
  //! Sets diagram options
  inline void SetMode(int M)
  {
    Mode = M;
  }
  //! Sets font for axis labels
  inline void SetAchseFont(FontType &A)
  {
    Achse = A;
  }
  //! Sets font for diagram title
  inline void SetTitelFont(FontType &T)
  {
    Titel = T;
  }
  //! Sets number of digits for axis numbers
  inline void SetNumDigits(int N)
  {
    NumDigits = N;
  }
  //! Check whether position is within diagram
  bool PunktInDiagramm(int X,int Y)
  {
    return((X >= DiagBer.XMin) && (X <= DiagBer.XMax) &&
           (Y >= DiagBer.YMin) && (Y <= DiagBer.YMax));
  }
  //! Check whether position is within diagram title
  bool PunktInLabel(int X,int Y)
  {
    wxRect DTRect(TText.GetTopLeftPos(DiagTextPos()),TText.GetTextSize());
    return (X >= DTRect.GetLeft()) && (X <= DTRect.GetRight()) &&
           (Y >= DTRect.GetTop()) && (Y <= DTRect.GetBottom());
  }
  //! Get y axis number under given point
  int GetYAxisNumber(int X,int Y);
  
  //! Get the pixel delta on y axis so x scaling matches y scaling
  int GetXDeltaForSameXYScaling();
};

// ----------------------------------------------------------------------------
// MarkObj
// ----------------------------------------------------------------------------

//! Marking Object
/*!
    This class is used for markers for diagrams. It has a x/y position
    and can be hidden / shown.
*/
class MarkObj :public SkalObj
{
  //! Color of marker
  long Farbe;
  //! Flags whether x and / or y line is visible
  bool XMarke,YMarke,visible;
  //! Current mark position
  PosType MarkPunkt;

 public:
  //! Draws x line of marker
  void MarkLinieX(wxDC &DC,long MarkX,int Min,int Max);
  //! Draws y line of marker
  void MarkLinieY(wxDC &DC,long MarkY,int Min,int Max);
  //! Show marker at position x, y
  void Mark(long X, long Y,wxDC &DC);
  //! Hide marker
  void UnMark(wxDC &DC);
  //! Draw marker
  void Draw(wxDC &DC,bool force = false);
  //! Set visible state of marker
  void SetVisible(bool visible,wxDC &DC);
  //! Default Ctor, maker is hidden
  MarkObj() {XMarke = false;YMarke = false;};
  //! Initialize marker
  void Init(BildType BildB,AxisDataArr &XA,AxisDataArr &YA,long Frb);
  //! Returns true if marker is shown
  bool Markiert();
  //! Set new screen coordinates
  virtual void UpdateBildBer(BildType BB);
  //! Set marker style options
  inline void SetMarkStyle(int Laenge,int Abst,int width)
  {
    MarkLaenge = Laenge;
    MarkAbst = Abst;
    MarkLinie = width;
  }
  //! Sets marker color
  inline void SetFarbe(long frb)
  {
    Farbe = frb;
  }
  //! Dtor
  virtual ~MarkObj();
};

#endif

