// Version 3.34  21.02.2004 Markus Greither
// Copyright (c) 1991-2004 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file diagtype.h
    \brief Header for the GUI independent data types for the diagram
*/

#ifndef __EFDIAGTYPE__
#define __EFDIAGTYPE__

// ----------------------------------------------------------------------------
// FontType
// ----------------------------------------------------------------------------

//! Font definition
/*!
    This is a structure to hold font information.
    Has SetString member for Font instead of MGPStr so that it
    can easily copied to clipboard
*/
struct FontType
{
//! Available font styles (StrikeOut currently not supported)
  enum FontStyles
  {
    Bold      = 1,
    Italic    = 2,
    Underline = 4,
    StrikeOut = 8
  };

  //! Font name
  SetString Font;
  //! Font settings
  int Size,Style;
  //! Font color
  unsigned long Color;

  //! Initializes font, color, size and style
  void Init(const _TCHAR *Ft,int Sz,int St,unsigned long Co)
  {
    Font.SetStr(Ft);
    Size = Sz;
    Style = St;
    Color = Co;
  }
  //! Ctor, sets font, color, size and style
  FontType(const _TCHAR *Ft,int Sz,int St,unsigned long Co)
  {
    Init(Ft,Sz,St,Co);
  }
  //! Default Ctor, sets default font
  FontType()
  {
    Font.SetStr(_TEXT("Arial"));
    Size = 12;
    Style = 0;
    Color = 0;
  }
  //! Copy operator
  void operator = (const FontType &f)
  {
    Font.SetStr(f.Font);
    Size = f.Size;
    Style = f.Style;
    Color = f.Color;
  }
  //! Equality operator
  bool operator != (const FontType &f) const
  {
    return (_tcscmp(Font.c_str(),f.Font.c_str()) ||
            (Size != f.Size) ||
            (Style != f.Style) ||
            (Color != f.Color));
  }
};

// ----------------------------------------------------------------------------
// GraphBaseObj
// ----------------------------------------------------------------------------

//! Diagram graphics base object, holds non-GUI constants
class GraphBaseObj
{
 public:
  //! Structure for vertices in WerType coords
  struct VectorType
  {
    //! X and Y positions
    double X,Y;
  };

  //! Structure for diagram current viewport size and pos
  struct BildType
  {
    //! X and Y screen coordinates
    int XMin,XMax,YMin,YMax;
  };

  //! Structure for vertices in BildType coords
  struct PosType
  {
    //! Screen coordinates
    long X,Y;
  };

  //! Flags for diagram options
  enum DiagOptionen
  {
    DiaTitel = 1,
    DiaRahmen = 2,
    DiaAlleLinienX = 4,
    DiaAlleLinienY = 8,
    DiaXVertikal = 16,
    DiaEndLabels = 32
  };
  //! Function for clipping point calculation for lines in diagrams
  /*! This function calculates the clipping point of
      the line specified by BildX/BildY and AkX,AkY and the
      clipping rectangle BildBer.
  */
  static bool ClipPoint(double &NewX,double &NewY,
                        const double BildX,const double BildY,
                        const double AkX,const double AkY,
                        const BildType BildBer,const bool InBer);

};

// ----------------------------------------------------------------------------
// AxisDataArr
// ----------------------------------------------------------------------------

//! Data for an axis
class AxisData
{
  //! Options
  enum OptionFlags
  {
    Log  =         1,
    Visible =      2,
    OppositeSide = 4
  };
  //! Options
  unsigned int Options;

 public:
  //! Label of axis
  SetString Label;
  //! Range
  double Min,Max;
  //! Text sizes of axis
  int TextL,TextH;
  //! Scaling variables
  double Fak,Sum;
  //! Ctor
  AxisData()
  {
    TextL = 0;
    TextH = 0;
    Options = 0;
  }
  //! Set log option
  void SetLog(bool l)
  {
    if (l)
      Options |= Log;
    else
      Options &= ~Log;
  }
  //! Get log option
  bool GetLog() const
  {
    return bool(Options & Log);
  }
  //! Set visible option
  void SetVisible(bool v)
  {
    if (v)
      Options |= Visible;
    else
      Options &= ~Visible;
  }
  //! Get visible option
  bool GetVisible() const
  {
    return (Options & Visible) != 0;
  }
  //! Set opposite side option
  void SetOppositeSide(bool s)
  {
    if (s)
      Options |= OppositeSide;
    else
      Options &= ~OppositeSide;
  }
  //! Get opposite side option
  bool GetOppositeSide() const
  {
    return (Options & OppositeSide) != 0;
  }
  //! Set Options
  void SetOptions(unsigned int O)
  {
    Options = O;
  }
  //! Get options
  const unsigned int &GetOptions() const
  {
    return Options;
  }
};

//! Axis Data array
class AxisDataArr
{
 public:
  //! Constant values
  enum values
  {
    MaxValue = 4
  };
 protected:
  //! Axis data
  AxisData data[MaxValue+1];
  //! Iterator
  size_t index;
 public:
   //! Ctor
   AxisDataArr()
   {
     index = 0;
   }
   //! Set index to beginning
   void Begin()
   {
     index = 0;
   }
   //! Return data count
   size_t GetMaxEintrag()
   {
     return MaxValue;
   }
   //! Increase index
   bool NextEint()
   {
     if (index < GetMaxEintrag())
     {
       index++;
       return true;
     }
     else
       return false;
   }
   //! Return data of current index
   AxisData &GetAkEint()
   {
     return data[index];
   }
   //! Set index to value
   void SetAkEint(size_t i)
   {
     if (i <= GetMaxEintrag())
       index = i;
   }
   //! Reset data
   void ResetKette()
   {
   }
   //! Create new data item
   void CreateAkEint(size_t i)
   {
     if (i <= GetMaxEintrag())
       index = i;
   }
   //! Get current index
   size_t AkEintNummer()
   {
     return index;
   }
   //! Return data of current index
   AxisData *operator ->()
   {
     return &data[index];
   }
};


#endif
