// Version 6.02  20.05.2007 Markus Greither
// Copyright (c) 1991-2007 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file ef.h
    \brief Header for the application main class and top level windows
*/

#ifdef __GNUG__
// #pragma interface "docview.h"
#endif

#ifndef __EF__
#define __EF__

#include <wx/docview.h>
#include <wx/docmdi.h>
#include <wx/help.h>
#include <wx/splash.h>

#if wxCHECK_VERSION(2,8,0)
#include <wx/aui/aui.h>
#endif

#ifdef EFHTMLHELP
#ifdef __WXMSW__
#include "wx/msw/helpchm.h"
#else
#include "wx/html/helpctrl.h"
#endif
#endif

// ----------------------------------------------------------------------------
// EFComboBox
// ----------------------------------------------------------------------------

//! Combo box that handles WXR_RETURN by sending EVT_COMMAND_ENTER to parent
class EFZoomBox : public wxComboBox
{
 public:
  //! Zoom values count
  enum EFZoomBoxValues
  {
    AnzZooms = 7
  };
  //! Zoom values used for zoom box in toolbar
  static wxString Zooms[AnzZooms];

  //! Ctor
  EFZoomBox(wxWindow* parent, wxWindowID id, const wxString& value = wxT(""),
             const wxPoint& pos = wxDefaultPosition,
             const wxSize& size = wxDefaultSize)
   : wxComboBox(parent,id,value,pos,size,AnzZooms,Zooms) {};
  //! Handle WXK_RETURN and send zoom event to parent
  void OnKeyDown(wxKeyEvent &event);
  //! Makes sure that WXK_RETURN is sent to OnKeyDown (?)
  void OnChar(wxKeyEvent &event);
  DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// EFDocManager
// ----------------------------------------------------------------------------

//! Document manager that calls GetCurrentView()->LoadData() for new docs
class EFDocManager : public wxDocManager
{
 public:
  //! Ctor
  EFDocManager(long flags = wxDEFAULT_DOCMAN_FLAGS, bool initialize = TRUE)
   : wxDocManager(flags,initialize) {};
  //! Create document and and load data into view
  wxDocument *CreateDocument(const wxString &path,long flags);
};

// ----------------------------------------------------------------------------
// MyFrame
// ----------------------------------------------------------------------------

//! Easy Funktion main frame.
/*!
    Handles events for the TEFOptionsDlg and TEFAboutDlg, Zooming.
    Tracks window size.
*/
class MyFrame: public wxDocMDIParentFrame
{
 public:
  //! Ctor
  MyFrame(wxDocManager *manager, wxFrame *frame, wxWindowID id,
          const wxString& title, const wxPoint& pos, const wxSize& size,
          const long type);
  //! Dtor
  virtual ~MyFrame();
  //! Respond to zoom events from zoom box
  void OnZoomChanged(wxCommandEvent &);
  //! Show about dialog
  void OnAbout(wxCommandEvent &);
  //! Show options dialog
  void OnOptionen(wxCommandEvent &);
  //! Get last window size and maximized mode before closing
  void OnCloseWindow(wxCloseEvent& event);
  //! Resize child windows
  void OnSize(wxSizeEvent &event);
  //! Patch for english window menu in wxWindows (still necessary ?)
  void SetWindowMenu(wxMenu *menu);
  //! Shows help window
  void OnHelp(wxCommandEvent &);
  //! GetClientArea begin
  wxPoint GetClientAreaOrigin() const;
  //! Overwrite this function in order to avoid annoying "window zapping" when dialog boxes close
  void SetFocusFromKbd() {};

  //! AUI manager
#if wxCHECK_VERSION(2,8,0)
  wxAuiManager m_frmmgr;
#endif
  DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// EFMDIChildFrame
// ----------------------------------------------------------------------------

//! Easy Funktion document child frame.
/*!
    MDI child frame for documents. Tracks window size.
*/
class EFMDIChildFrame :public wxDocMDIChildFrame
{
  //! Plain text viewer for function editor
  wxWindow *edplvw;
  //! Function editor
  wxWindow *edit;
  //! Currently active panel
  wxPanel *m_curpnl;
  //! Cached children counter
  size_t m_childcount;
 public:
  //! Ctor
  EFMDIChildFrame(wxDocument* doc, wxView* view, wxMDIParentFrame* parent,
                  wxWindowID id, const wxString& title,
                  const wxPoint& pos = wxDefaultPosition,
                  const wxSize& size = wxDefaultSize,
                  long style = wxDEFAULT_FRAME_STYLE,
                  const wxString& name = wxT("frame"));
  //! Auto resize if number of children changed
  void OnIdle(wxIdleEvent &);
  //! Arrange child windows
  void LayoutChildren(wxPanel *pnl = 0);
  //! Set editor child and create plain view
  void SetEditor(wxWindow *ed);
  //! Dtor, removes this frames menu from FileHistory
  virtual ~EFMDIChildFrame();
  //! Get current window size and maximized mode
  void OnSize(wxSizeEvent &event);
  //! Event handling and breaking recursion loops
  bool ProcessEvent(wxEvent& event);
  //! Overwrite this function in order to avoid annoying "window zapping" when dialog boxes close
  void SetFocusFromKbd() {};

  DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// MyApp
// ----------------------------------------------------------------------------

//! Easy Funktion application module.
/*!
    Provides Locale support, help controller, loads/writes program options
    to/from registry, loads XRC resources.
*/
class MyApp: public wxApp
{
 protected:
  //! Creates main frame menu
  wxMenuBar *CreateFrameMenu();
  //! Creates child frame menu
  wxMenuBar *CreateViewMenu();
  //! Creates help menu
  wxMenu *HilfeMenu();
  //! Creates file menu
  wxMenu *DateiMenu(bool MainMenu);
  //! Inserts clipboard commands into menu
  void InsertClipboardCommands(wxMenu *menu,bool selectall = false);
  //! Read information from system
  void EFGetResource(const char *section,const char *entry,wxString &value,
                     wxChar *def);

  //! Path for file open dialog
  wxString StdFilePath;
  //! Splash screen shown at program start
  wxSplashScreen *OpenWindow;
  //! Bitmap for splash screen
  wxBitmap *SplashBitmap;
  //! Number format
  NumListFormat LDV;
  //! Obtain the LDV information from locale
  void GetLocaleLDV();

 public:
  enum GlobalOptions
  {
    //! Flag for functions lower case (MyApp::Options)
    keyLowerCase = 1,
    //! Flag for INI file found (MyApp::Options) obsolete
    keyIniVorhanden = 2,
    //! Flag for main frame maximized at startup (MyApp::Options)
    keyMaximized = 4,
    //! Flag for child frame maximized at startup (MyApp::Options)
    keyMaximChild = 8,
    //! Flag for bitmap print (obsolete<=4.63)
    keyBitmapPrint = 16,
    //! Flag for displaying solver save warning (MyApp::Options)
    keyWarnSolve = 32,
    //! Flag for displaying diagram/graph delete warning (MyApp::Options)
    keyWarnDelete = 64,
    //! Flag for XML file save format (obsolete<=5.19)
    keyXMLSaveFormat = 128,
    //! Flag for EMF file save format for EMF graphics
    keySaveEMFAsBitmap = 256,
    //! Flag for real number results
    keyOnlyRealResults = 512,
    //! Flag for function selector automatically visible
    keyFuncSelVisible = 1024
  };
  enum DiagOptions
  {
    //! Flag for new worksheet with diagram
    DiagInNewWorksheet = 1,
    //! Flag for new diagrams with frame
    DiagHasFrame = 2,
    //! Flag for scale end labels displayed
    DiagEndLabels = 4,
    //! Flag for X values vertical
    DiagXValuesVert = 8
  };
  enum EditOptions
  {
    //! Flag for updating the editor window while scrolling
    EditScrollTrack = 1
  };

  //! Ctor
  MyApp();
  //! Set up application
  bool OnInit();
  //! Clean up application
  int OnExit();
  //! Run application
  int OnRun();

  //! Clipboard data formats for diagram and graph
  wxDataFormat graphformat,diagformat;
  //! Dummy clipboard format for editor text
  wxDataFormat editformat;
  //! Create child frame
  EFMDIChildFrame *CreateChildFrame(wxDocument *doc, wxView *view);
  //! Creates diagram menu
  wxMenu *CreateDiagramMenu(bool popup);
  //! Creates calculation menu
  wxMenu *CreateCalcMenu(bool popup);
  //! Locale for intl support
  wxLocale *Locale;
  //! Doc/View document manager
  wxDocManager* m_docManager;
  //! Flag for using windows number format (otherwise english)
  long WinZahlFormat;
  //! Last / new main frame size and position
  wxRect IniRect;
  //! Graph default format
  int GraphOpt,GraphSymb,GraphDarst,DiagFlags;
  //! Editor options
  int EditFlags;
  //! Option flags
  long Options;
  //! Flag for solver debug output
  long SolverDbgOut;
  //! Flag for number format (Engineering / scientific)
  long ZahlForm;
  //! Number of undo operations to store
  long m_UndoCount;
  //! Pointer to status bar
  wxStatusBar *sb;
  //! Pointer to tool bars
  wxToolBar *tb,*tb2;
  //! Pointer to zoom box in toolbar
  wxComboBox *zoombox;
  //! Default graph write format
  long GraphWriteForm;
  //! Current zoom value
  int Zoom;
  //! Style for function editor labels
  int LFontStyle;
  //! Style for function editor comments
  int KFontStyle;
  //! Color for function editor labels
  unsigned long LFontColour;
  //! Color for function editor comments
  unsigned long KFontColour;
  //! Colour for diagram background
  unsigned long DiagColour;
  //! The programs data directory
  wxString m_datadir;
  //! Returns the default path for file open
  inline wxString &GetStdFilePath()
  {
    return StdFilePath;
  }
  //! Sets the default path for file open
  inline void SetStdFilePath(const wxString &n)
  {
    StdFilePath = n;
    if (n[0] != 0)
      wxSetWorkingDirectory(n);
  }
  //! Returns number format
  const NumListFormat &GetLDV()
  {
    if (WinZahlFormat)
      return LDV;
    else
      return ZahlFormatObj::DefaultFormat;
  }
  //! Default graph import data
  GraphladenData m_graphdata;
  //! Returns toolbar zoom in button state
  bool GetZoomIn();
  //! Reads configuration values frome config
  void GetIniValues(bool getdefaults = false);
  //! Writes configuration values to config
  void SetIniValues();
  //! Retrieves the applications data directory
  inline void GetAppDataDir(wxString &dir)
  {
    dir = m_datadir;
  }
#ifdef EFHTMLHELP
#ifdef __WXMSW__
  //! MS HTML Help controller
  wxCHMHelpController HelpCtrl;
#else
  //! wxWindows HTML Help controller
  wxHtmlHelpController HelpCtrl;
#endif
#else
  //! WinHelp controller
  wxHelpController HelpCtrl;
#endif
};

DECLARE_APP(MyApp)

#endif

