// Version 5.20  05.07.2003 Markus Greither
// Copyright (c) 1991-2003 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efbrcalc.h
    \brief Header for the fraction format calculation object of the function editor
*/

#ifndef __EFBRCALC__
#define __EFBRCALC__

// ----------------------------------------------------------------------------
// FormatStrObj
// ----------------------------------------------------------------------------

//! Extension for formatted text to MGPStr
/*
    Supports text formats (bold, italic, underline, font size).
*/
class FormatStrObj :public MGPStr
{
public:
  //! Returns string from character position Pos
  const _TCHAR *FromMarkPos(unsigned int Pos) const
  {
    return &AkStr[ModMarkPos(Pos)];
  }
  //! Returns character at character position pos
  const _TCHAR &operator ()(unsigned int pos) const
  {
    return AkStr[ModCharPos(pos)];
  }
  //! Ctor
  FormatStrObj() : MGPStr() {};
  //! Ctor
  FormatStrObj(const FormatStrObj &other) : MGPStr() {SetStr(other.c_str());}
  //! Returns (character) length of string
  unsigned int FLength();
  //! Removes font marker at Pos
  unsigned int DelFontMarker(unsigned int Pos);
  //! Eventually removes current font marker and inserts new one at Pos
  void InsFontMarker(unsigned int Pos,_TCHAR C,unsigned char H);
  //! Insert character at character position Pos
  bool FInsChar(unsigned int Pos,_TCHAR C)
  {
    return MGPStr::InsChar(ModCharPos(Pos),C);
  }
  //! Replace character at character position Pos
  void FReplChar(unsigned int Pos,const _TCHAR C)
  {
    MGPStr::ReplChar(ModCharPos(Pos),C);
  }
  //! Delete n chars at character position Pos
  bool FDelNChar(unsigned int Pos,int Len)
  {
    int p = ModMarkPos(Pos);
    return MGPStr::DelNChar(p,ModLen(p,Len));
  }
  //! Add String and character position Pos
  bool FAddAtStr(unsigned int Pos,const _TCHAR *AStr)
  {
    return MGPStr::AddAtStr(ModCharPos(Pos),AStr);
  }
  //! Truncate string to Pos characters (w/o font markers)
  bool FShortStr(unsigned int Pos)
  {
    return MGPStr::ShortStr(ModMarkPos(Pos));
  }
  //! Truncate string after font marker at character position Pos
  bool ShortAfterMarker(unsigned int Pos)
  {
    return MGPStr::ShortStr(ModCharPos(Pos));
  }
  //! Insert string AStr at character position Pos
  bool FInsStr(unsigned int Pos,const _TCHAR *AStr)
  {
    return MGPStr::InsStr(ModCharPos(Pos),AStr);
  }
  //! Insert len chars from AStr at character position Pos
  bool FInsNChar(unsigned int Pos,const _TCHAR *AStr,int Len)
  {
    return MGPStr::InsNChar(ModCharPos(Pos),AStr,Len);
  }
  //! Calculates unformatted string and returns last found font format
  unsigned short GetTextStr(MGPStr &AStr) const;
  //! Returns character position from string position AFTER font marker at Pos
  unsigned int ModCharPos(unsigned int Pos) const;
  //! Returns character position from string position BEFORE font marker at Pos
  unsigned int ModMarkPos(unsigned int Pos) const;
  //! Returns formatted length of string at Pos from character length L
  unsigned int ModLen(unsigned int Pos,unsigned int L) const;
};

// ----------------------------------------------------------------------------
// BruchType
// ----------------------------------------------------------------------------

//! Data structure for lines with formatted functions.
struct BruchType
{
  //! Strings midline y positions
  ShortArrObj MidZ;
  //! Strings X positions
  LongArrObj X;
  //! Counts of lines above and below mid line for this function
  ShortArrObj ObZ,UntZ;
  //! Midline index for intermediate positions
  ShortArrObj RefMidZ;
  //! Formatted string
  MGPStr BZeile;
  //! Unformattted string
  FormatStrObj FZeile;
  //! If not zero, holds error position where formatting failed
  unsigned int FehlerPos;
  //! Flags for formatting error and comment line
  bool FormatError,Commentline;
  //! Modify flag for this function
  bool Modified;
  //! Line length in pixels
  int LineLength;
  //! Ctor
  BruchType();
};

// ----------------------------------------------------------------------------
// BruchArrObj
// ----------------------------------------------------------------------------

//! Dynamic array of BruchType
class BruchArrObj : public ListObj<BruchType>
{
 public:
  //! Returns currently active line
  BruchType &AkZ() {return GetAkEint();};
  //! Returns modify flag for line Nr
  bool IsModified(int Nr);
  //! Clear modify flag for line Nr
  void ClearModified(int Nr);
  //! Set modify flag for line numbers set in SetFus
  void SetModified(bool set,vector<bool> &SetFus);
};

// ----------------------------------------------------------------------------
// TEFTextCalc
// ----------------------------------------------------------------------------

//! Base class for TEFBruch.
/*
    Provides font generation and calculations.
*/
class TEFTextCalc : public wxWindow
{
 protected:
//! Structure for font properties
  struct FontMetricStruct
  {
    //! Font properties
    int OrgHeight,OrgAveCharW,Style,Weight,ExternalLeading,Descent;
    //! Underline flag
    bool Underlined;
    //! Font name
    wxString FontName;
    //! Font size in points
    int FPoints;
  };

  //! Calculated caret position for ScrollToCaret
  int CaretXPos,CaretYPos;
  //! Font properties
  FontMetricStruct FontMetric;
  //! Line height adjustment value
  float FSizeFak;
  //! Current zoom factor
  float Zoom,OrgZoom;
  //! Internally used DC for screen / print
  wxDC *DC;
  //! Internally used DC for clipboard / print
  wxDC *CDC;
  //! Font size in pixels
  int FHeight;
  //! Current normal font
  wxFont *WFont;
  //! Current label font
  wxFont *IFont;
  //! Current comment font
  wxFont *CFont;
  //! Average character width for scroll
  int AveChrW;
  //! Screen resolution
  int PixPerInchX,PixPerInchY;
  //! Bracket sizes for current font
  wxSize KlSize;
  //! Style for label font
  int LabelFontStyle;
  //! Style for comment font
  int KommentFontStyle;
  //! Color for label font
  long LabelFontColour;
  //! Color for comment font
  long KommentFontColour;
  //! Print flag
  bool Printing;

  //! Creates font with given format and size
  wxFont *CreateFormatFont(char Format,unsigned char H = 0) const;
  //! Determine font properties
  void GetFontMetrics(FontMetricStruct &FM,const wxFont &font);
  //! Creates new fonts with specified format
  void SetFormatFont(char Format,unsigned char H = 0,bool force = false);
  //! Changes font size in fixed increments
  void ChangeFontSize(bool up);
  //! Calculate string length
  wxSize ZeileExtent(const _TCHAR *string, int stringLen) const;

 public:
  //! Ctor
  TEFTextCalc() : wxWindow(),
    DC(0),CDC(0),WFont(0),IFont(0),CFont(0),Printing(false)
  {
  }
  //! Ctor
  TEFTextCalc(wxWindow *parent,wxWindowID id,const wxPoint &pos,
              const wxSize &size,long style,const wxString &title);
  //! Dtor, cleans up fonts
  virtual ~TEFTextCalc()
  {
    if (DC)
      DelDC();
    delete WFont;
    delete IFont;
    delete CFont;
  }
  //! Constants for font formatting / line positions
  enum BZeileMarkers
  {
    FontMarker  = '\f', // Marker fuer anderes Schrift-Format
    PageMarker  = 0x50, // 'P' Marker fuer Seiten-Umbruch
    FontHeight  = 0x40, // '@'
    Bold        = 0x31, // '1'
    Italic      = 0x32, // '2'
    Underline   = 0x34, // '4'
    Strikeout   = 0x38, // '8'
    Normal      = 0x30, // '0'
    PosMarker   = 1,    // neue Position in BZeile
    BruchMarker = 2     // bearbeiteter Bruchstrich in BZeile
  };

  //! Hide caret
  void HideCaret();
  //! Show caret
  void ShowCaret();
  //! Create DC and fonts
  void GetDC();
  //! Delete DC
  void DelDC();
  //! Calculate string dimensions
  inline void EFTextExtent(wxDC &dc,const wxString &string,
                           wxCoord *w,wxCoord *h,
                           wxCoord *desc = 0,wxCoord *elead = 0) const
  {
    dc.GetTextExtent(string,w,h,desc,elead);
  }
  //! Return currently set zoom
  float GetZoom() {return OrgZoom;}
  //! Return actually internal used zoom
  float GetInternalZoom() {return Zoom;}
  //! Calculate formatted string dimensions
  wxSize EFTabbedTextExtent(const _TCHAR *string,
                            int stringLen, bool kursiv) const;
  //! Determines first position from f that is no font marker
  void FindFirstChar(const _TCHAR *Zeile,unsigned int &f) const;
  //! Delete enclosing brackets of string at f, even if Comment/Blanks trailing
  void KlammernEntfernen(MGPStr &Akstr,unsigned int f) const;
  //! Return current label font style
  int GetLabelFontStyle() const
  {
    return LabelFontStyle;
  }
  //! Set current label font style
  void SetLabelFontStyle(int Style)
  {
    LabelFontStyle = Style;
  }
  //! Return current comment font style
  void SetKommentFontStyle(int Style)
  {
    KommentFontStyle = Style;
  }
  //! Set current label font color
  void SetLabelFontColour(unsigned long Colour)
  {
    LabelFontColour = Colour;
  }
  //! Set current comment font color
  void SetKommentFontColour(unsigned long Colour)
  {
    KommentFontColour = Colour;
  }

  DECLARE_DYNAMIC_CLASS(TEFTextCalc)
};

// ----------------------------------------------------------------------------
// BruchCalcObj
// ----------------------------------------------------------------------------

//! Class for function formatting.
/*
    Calculates X/Y formats.
*/
class BruchCalcObj
{
 private:
  //! Function formatting calculation object
  const TEFTextCalc *TxtCalc;
  //! Current line
  BruchType &BrAkZ;
  //! Parser
  ParserObj &Psr;
  //! Flags for show brackets, mark labels
  bool KlammerDarst,MarkLabels;
  //! Current index for InZeile/OutZeile
  int AkOutZeile,AkInZeile;
  //! Current position
  int AkZeile;
  //! Create new inner line
  inline void AddInZeile(const int AkMidZeile);
  //! Create new outer line
  inline void AddOutZeile(const int AkMidZeile);
  //! Create new mid line
  void NewMidZeile(bool Inc,const int AkMidZeile);
  //! Read outer line and calculate height
  inline int FindObBruch(const _TCHAR *Zeile,unsigned int AkPos,
                         const int AkMidZeile,int AkAbsch,int LinePos)
  {
    bool PGef = false,UGef = false;
    int ThisZeile = AkZeile+1;// Ersten Eintrag fuer diese Zeile speichern
    int ret = FindBruch(Zeile,AkPos,PGef,UGef,ThisZeile,true);
    short ZH = short(BrAkZ.ObZ[ThisZeile]+BrAkZ.UntZ[ThisZeile]+LinePos);
    if (ZH > BrAkZ.ObZ[AkMidZeile])
      BrAkZ.ObZ.SetValue(AkMidZeile,ZH);
    if (ZH > BrAkZ.ObZ[AkAbsch])
      BrAkZ.ObZ.SetValue(AkAbsch,ZH);
    return ret;
  }

  //! Read inner line and calculate height
  inline int FindUntBruch(const _TCHAR *Zeile,unsigned int AkPos,
                          const int AkMidZeile,int AkAbsch)
  {
    bool PGef = false,UGef = false;
    int ThisZeile = AkZeile+1;// Ersten Eintrag fuer diese Zeile speichern
    int ret = FindBruch(Zeile,AkPos,PGef,UGef,ThisZeile,true);
    short ZH = short(BrAkZ.ObZ[ThisZeile]+BrAkZ.UntZ[ThisZeile]+1);
    if (ZH > BrAkZ.UntZ[AkMidZeile])
      BrAkZ.UntZ.SetValue(AkMidZeile,ZH);
    if (ZH > BrAkZ.UntZ[AkAbsch])
      BrAkZ.UntZ.SetValue(AkAbsch,ZH);
    return ret;
  }
  //! Replaces * with point in str
  void ZeileWithoutAsterix(MGPStr &str);
  //! Moves lines to fit new inner lines
  void ZeilenVerschieben(bool BruchstrichOut,const int ThisInZeile,
                         const int ThisOutZeile,const int AkMidZeile);
  //! Read line , evaluate expressions, calculate height
  int FindBruch(const _TCHAR *Zeile,unsigned int AkPos,bool &PGef,bool &UGef,
                const int AkMidZeile,bool GanzeZeile);
  //! Insert new position in BZeile
  bool InsertPos(int &AkZeile,int AkMidZeile,int AkAbsch,unsigned int Pos);
  //! Inserts Zeile in BZeile, sets positions for brackets and comments
  bool ZeileInAbschnitten(const _TCHAR *Zeile,unsigned int SLen,int &AkZeile,
                          int AkMidZeile,int &AkAbsch,ShortArrObj &Abschs,
                          unsigned int AkPos,int &Len,bool InsPos = true);
  //! Insert lines to fit new inner lines
  void ZeilenEinfuegen(int AkMidZeile,bool PGef,bool AkPGef,
                                      bool UGef,bool AkUGef);
  //! Calculate minimum distance of this line from next outer line
  int GetInAbst(int AkMidZeile);
  //! Calculate minimum distance of this line from next inner line
  int GetOutAbst(int AkMidZeile);
  //! Get inner lines (mirrored !)
  int GetInZeilen(int AkMidZeile);
  //! Get outer lines (mirrored !)
  int GetOutZeilen(int AkMidZeile);
  //! Calculate string dimensions for underlength
  int UnderscoreTextExtent(const _TCHAR *Zeile,unsigned int SLen,int &AkZeile,
                           const int AkMidZeile,int &AkAbsch,
                           ShortArrObj &Abschs,int GesLen);
  //! Checks whether last line was out
  bool BruchStrichOut(int AkMidZeile)
  {
    return (abs(BrAkZ.MidZ[AkMidZeile-1]) > abs(BrAkZ.MidZ[AkMidZeile]));
  }
  //! Checks whether inner line must be added
  bool AddIn(int AkMidZeile,bool PGef,bool AkPGef,bool UGef,bool AkUGef)
  {
    return ((BrAkZ.MidZ[AkMidZeile] > 0) && (!PGef && AkPGef) ||
            (BrAkZ.MidZ[AkMidZeile] <= 0) && (!UGef && AkUGef));
  }
  //! Checks whether outer line must be added
  bool AddOut(int AkMidZeile,bool PGef,bool AkPGef,bool UGef,bool AkUGef)
  {
    return ((BrAkZ.MidZ[AkMidZeile] <= 0) && (!PGef && AkPGef) ||
            (BrAkZ.MidZ[AkMidZeile] > 0) && (!UGef && AkUGef));
  }
  //! Set error position
  void SetFehlerPos(const _TCHAR *Zeile,unsigned int Pos)
  {
    BrAkZ.FehlerPos = Zeile-BrAkZ.FZeile.c_str()+Pos;
    BrAkZ.FormatError = true;
  }
  //! Calculates low part of line position
  char GetBIndexLow(int x)
  {
    return char((x & 127)+1);
  }
  //! Calculates high part of line position
  char GetBIndexHigh(int x)
  {
    return char((x >> 7)+1);
  }
  //! Evaluate characters
  bool EvalChars(const _TCHAR *Zeile,unsigned int &Pos,int &Pos1,ShortArrObj &Anfs,int &KlZaehler,
                 unsigned int &AnfLastComment,unsigned int &Anf,int &End,int &BruchKl,
                 bool &FirstChar,int &AnfKlE,bool &AkBPGef,bool &Vorzeichen,bool &RechZ,
                 unsigned int &AkUPos,const int AkMidZeile,bool &GanzeZeile);
  //! Insert base of exponential expression
  void InsertBase(const _TCHAR *Zeile,int &Pos1,unsigned int &Anf,bool &PGef,bool &UGef,
                  ShortArrObj &Abschs,int &AkAbsch,int &PotMidPos,int &Len,
                  const int AkMidZeile,unsigned int &AkPos)
  {
    int tmp = AkZeile+2;
    // insert in same line
    BrAkZ.MidZ.SetValue(AkZeile+1,short(BrAkZ.MidZ[AkZeile]));
    // make sure that eventually already inserted lines for exponents are recognized
    PGef = true;
    // Pass copy of base to FindBruch(),
    // not whole line from Anf, in order to avoid recursion with e. g. (^2) !
    MGPStr tmps;
    tmps.AddNChar(&Zeile[Anf],Pos1-Anf);
    Len += FindBruch(tmps.c_str(),AkPos+Len,PGef,UGef,AkMidZeile,false);
    // Transfer line height
    if (BrAkZ.UntZ[Abschs[AkAbsch]] < BrAkZ.UntZ[tmp])
      BrAkZ.UntZ.SetValue(Abschs[AkAbsch],BrAkZ.UntZ[tmp]);
    if (BrAkZ.ObZ[Abschs[AkAbsch]] < BrAkZ.ObZ[tmp])
      BrAkZ.ObZ.SetValue(Abschs[AkAbsch],BrAkZ.ObZ[tmp]);
    PotMidPos = BrAkZ.ObZ[tmp];
    if (!PotMidPos)
      PotMidPos = 1;
    BrAkZ.BZeile.AddChar(' '); // replaces '^'
  }
  //! Insert fraction or exponent in line, with the divider line at Pos1
  bool InsertFraction(const _TCHAR *Zeile,int &Pos1,unsigned int &Anf,int &RekCount,const int AkMidZeile,
                      int &PotMidPos,ShortArrObj &Abschs,int &AkAbsch,int &Len,unsigned int &AkPos,
                      unsigned int &Pos);

 public:
  //! Recursion limit for FindBruch
  enum BruchCalcConsts
  {
    MAXREK = 50
  };
  //! outer and inner line positions
  ShortArrObj OutZeile,InZeile;
  //! Recursion counter
  int RekCount;
  //! Ctor
  BruchCalcObj(BruchType &B,ParserObj &P,const TEFTextCalc *T,bool K,bool M)
   : BrAkZ(B),
     Psr(P)
  {
    KlammerDarst = K;
    MarkLabels = M;
    TxtCalc = T;
    AkOutZeile = 0; // No outer line yet
    AkInZeile = 0; // No inner line yet
    OutZeile.Clear(); // releasea memory (?!)
    InZeile.Clear();
    OutZeile.SetValue(0,0); // Initialize first field
    InZeile.SetValue(0,0);
    BrAkZ.MidZ.Clear();
    BrAkZ.X.Clear();
    BrAkZ.UntZ.Clear();
    BrAkZ.ObZ.Clear();
    AkZeile = 0;
    RekCount = 0;
  }
  //! Calculates index at this line position
  static int GetBIndex(MGPStr &BZeile,int x)
  {
    return ((unsigned char)(BZeile[x])-1)+
           (((unsigned char)(BZeile[x+1])-1) << 7);
  }
  //! Inline call to FindBruch with Gef and PGef false
  inline int FindBruch(const _TCHAR *Zeile,unsigned int AkPos)
  {
    bool PGef = false,UGef = false;
    int AkMidZeile = AkZeile+1;// Store first entry for this line
    return FindBruch(Zeile,AkPos,PGef,UGef,AkMidZeile,true);
  }
};

#endif

