// Version 5.25  08.03.2004 Markus Greither
// Copyright (c) 1991-2004 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efdlgs.h
    \brief Header for the dialogs not related to a specific window
*/

#ifndef __EFDLGS__
#define __EFDLGS__

#include "efdlgbas.h"
#include "wx/progdlg.h"

// ----------------------------------------------------------------------------
// UmformenData
// ----------------------------------------------------------------------------

//! Data structure for TEFUmformenDlg
struct UmformenData
{
  //! True if current marking is solved instead of current function
  int AkZeileOpt;
  //! Currently active solver template
  AuflVorlageStruct AkSett;
  //! Pointer to EFComboBoxes
  EFComboData *ErgH,*ParH,*ZusH;
  //! Solver templates
  AuflVorlageObj *Vorlage;
};

// ----------------------------------------------------------------------------
// TEFUmformenDlg
// ----------------------------------------------------------------------------

//! Dialog for solving functions
/*
    This dialog lets the user set the various options for function solving.
    When OK is pressed, the TEFSolverDlg is invoked that displayes the
    results of function solving.
*/
class TEFUmformenDlg :public TEFHilfeDlg
{
  //! Passed in data
  UmformenData &Data;
  //! Pointers to EFComboBoxes
  TEFComboBox *Erg,*Par,*Zus,*Tem;
  //! Currently selected index in solver templates
  int AkDataIndex;

 public:
  //! Ctor
  TEFUmformenDlg(UmformenData &D);
  //! Dtor
  virtual ~TEFUmformenDlg();
  //! Create dialog
  void Create(wxWindow *parent,wxWindowID Id,const wxString &title);
  //! Respond to OK button press
  void OnOk(wxCommandEvent &event);
  //! Respond to cancel button press
  void OnCancel(wxCommandEvent &event);
  //! Respond to new template selection
  void OnTemplate(wxCommandEvent &);
  //! Respond to template deletion
  void OnDelTemplate(wxCommandEvent &);
  //! Set new template
  void SetData(int i);

  DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// TEFNaeherungDlg
// ----------------------------------------------------------------------------

//! Dialog for approximation functions
class TEFNaeherungDlg :public TEFHilfeDlg
{
  //! Passed in data
  NaeherungDataType &Data;
  //! Vertices for approx. function
  wxListBox *Param;
  //! Respond to button clicks
  void BnClicked();
  //! Respond to button clicks
  void OnBnClicked(wxCommandEvent &);
  //! Update enable state of controls
  void EnChanged();
  //! Update enable state of controls
  void OnEnChanged(wxCommandEvent &);
  //! Respond to function type selection
  void OnFunctionType(wxCommandEvent &);

 public:
  //! Ctor
  TEFNaeherungDlg(NaeherungDataType &D);
  //! Dtor
  virtual ~TEFNaeherungDlg();
  //! Create dialog
  void Create(wxWindow *parent,wxWindowID Id,const wxString &title);
  //! Respond to OK button press
  void OnOk(wxCommandEvent &event);
  //! Respond to cancel button press
  void OnCancel(wxCommandEvent &event);
  //! Respond to add button press
  void OnYes(wxCommandEvent &event);
  //! Respond to delete button press
  void OnNo(wxCommandEvent &event);
  //! Respond to delete all button press
  void OnIgnore(wxCommandEvent &event);
  //! Respond to line selection
  void OnParSel(wxCommandEvent &event);

  DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// BerechnenData
// ----------------------------------------------------------------------------

//! Data structure for TEFBerechnenDlg
struct BerechnenData
{
  //! Function update flag
  bool Update;
  //! Pointer to view
  TEFView *EFView;
  //! Pointer to diagram window
  TEFGraphControl *EFGraph;
  //! Function name
  MGPStr FuName;
  //! Pointer to graph data
  const GraphDataObj *AkFeld;
  //! Pointer to calculated functions array
  vector<bool> *BerFus;
  //! Error information
  ParserObj::FehlerRec Fehler;
};

// ----------------------------------------------------------------------------
// TEFBerechnenDlg
// ----------------------------------------------------------------------------

//! Dialog for showing progress during function calculation.
/*!
    This dialog also executes TEFFunktionCalcObj to do the actual
    calculations in idle time.
*/
class TEFBerechnenDlg : public wxProgressDialog
{
  //! Function parser
  ParserObj &Parser;
  //! Passed in data
  BerechnenData *Data;

 public:
  //! Ctor
  TEFBerechnenDlg(wxWindow *parent,wxWindowID Id,
                  const wxString &title,ParserObj &psr);
  //! Dtor
  virtual ~TEFBerechnenDlg();
  //! Set up with new data
  void Init(BerechnenData &D);
  //! Save new graph
  void GraphSpeichern(const _TCHAR *Name = 0);
  //! Finish up after calculation of graph
  int Berechnet();

 protected:
  //! Executes the calculations
  int ShowModal();
  //! Function calculator
  TEFFunktionCalc *FunktionCalc;
  //! Data structure for function calculator
  TEFFunktionCalcData *FCD;
};

// ----------------------------------------------------------------------------
// TEFOptionenDlg
// ----------------------------------------------------------------------------

//! Dialog for general program options
class TEFOptionenDlg :public TEFHilfeDlg
{
  //! Pointer to functions array
  FunctionArrObj *Fu;
  //! Pointer to default path input line
  wxTextCtrl *TE;

 public:
  //! Ctor
  TEFOptionenDlg();
  //! Dtor
  virtual ~TEFOptionenDlg();
  //! Initialize the controls
  void InitControls();
  //! Create dialog
  void Create(wxWindow *parent,wxWindowID Id,const wxString &title);
  //! Respond to reset button press
  void OnReset(wxCommandEvent &event);
  //! Respond to OK button press
  void OnOk(wxCommandEvent &event);
  //! Respond to comment color button press
  void OnCommentFarbe(wxCommandEvent &);
  //! Respond to label color button press
  void OnLabelFarbe(wxCommandEvent &);
  //! Respond to diagram color button press
  void OnDiagColour(wxCommandEvent &);
  //! Respond to browse button press
  void OnBrowsePath(wxCommandEvent &);

 protected:

  DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// TEFAboutDlg
// ----------------------------------------------------------------------------

//! Dialog for About information
class TEFAboutDlg : public TEFHilfeDlg
{
 public:
  //! Ctor
  TEFAboutDlg () : TEFHilfeDlg() {};

 public:
  //! Send item through the internet or mail
  void OnSendItem(wxCommandEvent &event);
  //! Create dialog
  void Create(wxWindow *parent,wxWindowID Id,const wxString &title);

 protected:

   DECLARE_EVENT_TABLE()
};

#endif

