// Version 6.04  28.04.2008 Markus Greither
// Copyright (c) 1991-2008 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efgemein.h
    \brief Header for the data management classes
*/

#ifndef __EFGEMEIN__
#define __EFGEMEIN__

class EFXmlNode;

class EFXmlDocument;

// ----------------------------------------------------------------------------
// LabelObj
// ----------------------------------------------------------------------------

//! Labelling object.
/*!
    Holds label information (Font, text, position).
*/
class LabelObj
{
 public:
  //! Label text
  MGPStr Label;
  //! Label font
  FontType Font;
  //! Options
  int Flags;
  //! Position of label in diagram (value range)
  double XPos,YPos;
  //! Offset of label from XPos, YPos (in 1000/diagsize units)
  int XOffs,YOffs;
  //! Size (cached here for screen update)
  int XSize,YSize;

  //! Values for Flags
  enum FlagValues
  {
    Transparent = 1,
    Arrow       = 2,
    Frame       = 4,
    Diag2       = 8,
    ShowLabel   = 16,
    ShowX       = 32,
    ShowY       = 64
  };

  //! Ctor
  LabelObj()
  {
  }
  //! Initialize label with string, flags and position
  void Init(const _TCHAR *str,int Fl,double XP,double YP)
  {
    Label.SetStr(str);
    XPos = XP;
    YPos = YP;
    Flags = Fl;
    XOffs = 0;
    YOffs = 0;
  }
};

// ----------------------------------------------------------------------------
// LabelArrObj
// ----------------------------------------------------------------------------

//! Dynamic label array
class LabelArrObj :public ListObj<LabelObj>
{
 public:
  //! Write labels in XML format
  void XmlWrite(EFXmlNode *n,LabelArrObj *old = 0);
};

// ----------------------------------------------------------------------------
// GraphDataObj
// ----------------------------------------------------------------------------

//! Graph data object.
/*!
    Holds graph data points, colour, style.
*/
class GraphDataObj
{
 protected:
  //! Number of vertices of this graph
  unsigned int Punkte;
  //! Pointer to array with x values
//  double *X;
  vector<double> X;
  //! Pointer to array of complex y values
//  Cpx *Y;
  vector<Cpx> Y;
  //! Realloc the x and y arrays for new number of datapoints
  /*! When called, original number of points needs to be
      still available in Punkte so that old data is properly copied! */
  bool AdjustArraySize(unsigned int pkt);

 public:
  //! Sets the graph symbol option
  enum GraphSymbolFlags
  {
    GraphMitSymbol = 0,
    GraphOhneSymbol = 1
  };
  //! Other graph options
  enum GraphOptionFlags
  {
    GraphFett = 1,    // GraphOptionen-Konstanten
    PstPrcOff = 2,
    GraphDia1 = 4,
    GraphDia2 = 8,
    TextDatDatei = 8192 // ExtTextDatei-Option fuer GraphOptionen
  };
  //! Graph drawing style
  enum ZeichenArtFlags
  {
    gPunkteMarkieren = 0,
    gPunkteVerbinden = 1,
    gPunkteVerbMark  = 2,
    gPunkteSplines   = 4
  };
  enum GraphDataObjLimits
  {
    //! Defines max number of symbols per graph
    MaxFMark  = 20,
    //! Defines max. possible data points for graphs
    MaxPunkte = 2000000
  };
  //! Ctor
  GraphDataObj();
  //! Dtor, cleans up x and y arrays
  virtual ~GraphDataObj();
  //! Copy Ctor
  GraphDataObj (const GraphDataObj &Other)
  {
    VerString[0] = 0;
    FuString[0] = 0;
    FileString[0] = 0;
    Punkte = 0;
    AdjustArraySize(0);
    Other.CopyData(this,true);
  }
  //! Comparison of data
  bool HasDifferentData(const GraphDataObj &other) const
  {
    return ((Punkte != other.Punkte) ||
            (X != other.X) || (Y != other.Y));
  }
  //! Copy operator
  GraphDataObj &operator = (const GraphDataObj &Other)
  {
    Other.CopyData(this,true);
    return *this;
  }
  //! Version string for data read write (PASCAL style with size byte in [0])
  char VerString[11];
  //! Function name and external file name if used
  _TCHAR FuString[256],FileString[256];
  //! Graph options
  unsigned short GraphSymbol,GraphOptionen,ZeichenArt;
  //! Graph color
  unsigned long Farbe;
  //! Corresponding axis indexes
  unsigned char XAxisIndex[2],YAxisIndex[2];
  //! Write graph in XML format
  bool XmlWriteGraph(EFXmlNode *n,GraphDataObj *old = 0) const;
  //! Load graph in XML format
  bool XmlLoadGraph(EFXmlNode *n);
  //! Set new number of datapoints and resize x and y arrays
  void SetPunkte(unsigned int pkt);
  //! Increase number of datapoints
  void IncPunkte();
  //! Decrease number of datapoints
  void DecPunkte();
  //! Returns number of datapoints available
  const unsigned int &GetPunkte() const {return Punkte;}
  //! Accessor for given x value
  inline const double &GetX(unsigned int pkt) const
  {
    if (pkt < Punkte)
      return X[pkt];
    else
      return X[0];
  }
  //! Accessor for given y value
  inline const Cpx &GetY(unsigned int pkt) const
  {
    if (pkt < Punkte)
      return Y[pkt];
    else
      return Y[0];
  }
  //! Set x value for datapoint number pkt
  inline void SetX(unsigned int pkt,const double &value)
  {
    if (pkt < Punkte)
      X[pkt] = value;
  }
  //! Set y value for datapoint number pkt
  inline void SetY(unsigned int pkt,const Cpx &value)
  {
    if (pkt < Punkte)
      Y[pkt] = value;
  }
  //! Get size of x value array
  inline unsigned long GetXSize() const
  {
    return X.size()*sizeof(double);
  }
  //! Get size of y value array
  inline unsigned long GetYSize() const
  {
    return Y.size()*sizeof(Cpx);
  }
  //! Copy graph data to Ziel
  void CopyData(GraphDataObj *Ziel,bool CopyData) const;
  //! Build continuous memory block (for clipboard)
  void *ConvertToMemBlock(void *append = 0,int *size = 0);
  //! Unwrap continuous memory block (from clipboard)
  size_t ConvertFromMemBlock(void *tmp);
  //! Move graph data from oth here
  void MoveData(GraphDataObj &oth)
  {
    X = oth.X;
    Y = oth.Y;
    oth.X.clear();
    oth.Y.clear();
    Punkte = oth.Punkte;
    oth.Punkte = 0;
    _tcscpy(FuString,oth.FuString);
    _tcscpy(FileString,oth.FileString);
    strcpy(VerString,oth.VerString);
    GraphSymbol = oth.GraphSymbol;
    GraphOptionen = oth.GraphOptionen;
    ZeichenArt = oth.ZeichenArt;
    Farbe = oth.Farbe;
    XAxisIndex[0] = oth.XAxisIndex[0];
    YAxisIndex[0] = oth.YAxisIndex[0];
    XAxisIndex[1] = oth.XAxisIndex[1];
    YAxisIndex[1] = oth.YAxisIndex[1];
  }
};

// ----------------------------------------------------------------------------
// CopyArr
// ----------------------------------------------------------------------------

//! Array-like class with copy operator
template<class T> class CopyArr
{
  //! Data values
  T Data[2];
 public:
  //! Ctor
  CopyArr()
  {
  }
  //! Copy Ctor
  CopyArr(const CopyArr<T> &Other)
  {
    *this = Other;
  }
  //! Copy operator
  void operator = (const CopyArr<T> &Other)
  {
    Data[0] = Other.Data[0];
    Data[1] = Other.Data[1];
  }
  //! Accessor
  const T &operator [] (int ind) const
  {
    return Data[ind];
  }
  //! Accessor, writeable version
  T &operator [] (int ind)
  {
    return Data[ind];
  }
};

// ----------------------------------------------------------------------------
// MaTyp
// ----------------------------------------------------------------------------

//! Enumeration data type.
/*!
    Holds information about the kind of enumeration selected, and
    min/max/step information.
*/
class MaTyp
{
 public:
  //! Enumeration modes
  enum
  {
    ModeSum = 0,
    ModePar,
    ModeVar,
    ModeFourier,
    ModeNo,
    ModeParList
  };
  enum MaTypLimits
  {
    //! Maximum iterations for MaTyp fourier mode
    FourIt = 1000000,
    //! Maximum iterations for MaTyp summing mode
    SumIt  = 1000,
    //! Maximum iterations for MaTyp parameter mode
    ParIt  = 100
  };
 protected:
  //! Current postion for GetNext()
  unsigned int AkPos;

 public:
  //! String with parameter list, SetString because of clipboard copy
  SetString ParListe;
  //! Start, End and Step values, SetString because of clipboard copy
  double Anf,Ende,Step;
  //! Current mode
  int Mode;
  //! Label, SetString because of clipboard copy
  SetString Bezeichner;
  //! Unit for Step
  SetString Einheit;

  //! Build parameter label string for current mode
  void ParBezeichner(MGPStr &HiString,double HiVar,ZahlFormatObj &ZF);
  //! Build function label string for current mode
  void FunktionsBezeichner(MGPStr &HiString,ZahlFormatObj &ZF);
  //! Retrieves the first value (either Anf or first entry in ParListe)
  double GetFirst();
  //! Next value (either AkPos + Step or next entry in ParListe)
  bool GetNext(double &Ak);
  //! Determines whether another step is available
  bool NextStepAvail(const double &Ak);
  //! Receive the parameter list
  const _TCHAR *GetParListe() {return ParListe.c_str();};
  //! Set new parameter list
  void SetParListe(const _TCHAR *ak)
  {
    ParListe.SetStr(ak);
  }
  //! Ctor
  MaTyp()
  {
    Anf = 0;
    Ende = 0;
    Step = 0;
    AkPos = 0;
    Mode = ModeNo;
  }
};

// ----------------------------------------------------------------------------
// DatFeldObj
// ----------------------------------------------------------------------------

//! Dynamic array for graph data
class DatFeldObj :public ListObj<GraphDataObj>
{
};

// ----------------------------------------------------------------------------
// GraphManagerObj
// ----------------------------------------------------------------------------

//! Graph data management class.
/*!
    Manages graph data, with provisions for file read/write, and
    copy of graph data.
*/
class GraphManagerObj :public ZahlFormatObj
{
  enum
  {
    //! Number of colors in FarbArr
    MaxFarbe = 5
  };
  //! Version number strings
  static const char *AltDateiVersion,*DatenDateiVersion;
  //! Color array for graph colors
  static const unsigned long FarbArr[];
  //! Data for current graph, after post processing
  GraphDataObj AkFeld;
  //! Flag for post processing active
  int PstPrc;

  //! Determine wether graph data is valid
  bool ValidGraphDaten();
  //! Load another graph from Dat
  bool LoadFeld(istream &Dat);
  //! Write another graph from Dat
  void WriteFeld(ostream &Dat);
  //! Dynamic array with graph data
  DatFeldObj *DatFeld;
  //! Format string x with at least y characters, fill up with blanks
  void UniLength(MGPStr &x,unsigned int y);
  //! Skip blanks and separators until number is reached, updates Position
  bool SkipSeparators(MGPStr &MGZeile,unsigned int &Position,
                      NumListFormat &AkLDV,bool TabSeparator, bool SpaceSeparator);
  //! Updates Position with end of graph label
  void GetLabelEnd(MGPStr &MGZeile,unsigned int &Position,
                   NumListFormat &AkLDV, bool TabSeparator, bool SpaceSeparator);

public:
  //! Ctor
  GraphManagerObj()
  {
    DatFeld = new DatFeldObj;
  }
  //! Dtor
  virtual ~GraphManagerObj();
  //! Copy Ctor
  GraphManagerObj(const GraphManagerObj &Other) : ZahlFormatObj()
  {
    DatFeld = new DatFeldObj;
    *this = Other;
  }
  //! Copy operator
  GraphManagerObj &operator = (const GraphManagerObj &Other)
  {
    AkFeld = Other.AkFeld;
    PstPrc = Other.PstPrc;
    *DatFeld = *Other.DatFeld;
    SetLDV(Other.GetLDV());
    return *this;
  }
  //! Move data from oth here
  void MoveData(GraphManagerObj &Other);
  //! Return pointer to current graph
  const GraphDataObj &GetAkEint() 
  {
    if (DatFeld->GetMaxEintrag() >= 0)
      return DatFeld->GetAkEint();
    else
      return AkFeld;
  }
  //! Copy current graph to Ziel
  void GetAkFeldData(GraphDataObj *Ziel);
  //! Set current graph with data from Quelle
  void SetAkFeldData(const GraphDataObj *Quelle,bool CopyDaten);
  //! Copy graph data in Quelle into Ziel
  void CopyFeldData(const GraphDataObj *Quelle,GraphDataObj *Ziel,bool CopyDaten);
  //! Add graph in Feld at the end of DatFeld
  void AppendYFeld(const GraphDataObj *Feld,bool KeepColor = false);
  //! Initialize
  void Init(int PostProc,const NumListFormat &iLDV);
  //! Set Post Proc flag
  inline void SetPostProc(int PP) {PstPrc = PP;};
  //! Return number of current graph
  inline int GetAkPos()
  {
    return(DatFeld->AkEintNummer()+1);
  }
  //! Get number of last available graph
  inline int GetMaxPos()
  {
    return(DatFeld->GetMaxEintrag()+1);
  }
  //! Select graph Nr as current
  inline void SetAkPos(int Nr)
  {
    if ((Nr > 0) && (DatFeld->GetMaxEintrag() >= 0))
    {
      DatFeld->SetAkEint(Nr-1);
      CopyFeldData(&DatFeld->GetAkEint(),&AkFeld,true);
    }
  }
  //! Accessor for current graph
  GraphDataObj &AkGraph()
  {
    return AkFeld;
  }
  //! Delete current graph
  void LoeschAkGraph();
  //! Load next graph into AkFeld
  bool LoadYFeld(int nr = -1);
  //! Initialize graph array
  void ResetYDaten();
  //! Load graphs from Datei
  bool LoadYDatei(istream &Datei);
  //! Write graphs to Datei
  bool WriteYDatei(ostream &Datei);
  //! Load graph in XML format
  bool XmlLoadGraph(EFXmlNode *node,int index);
  //! Write graphs in XML format
  bool XmlWriteGraphs(EFXmlNode *node);
  //! Delete all graphs
  void LoeschYDaten();
  //! Checks whether any graphs are available
  bool YDatenVorhanden();
  //! Processes current graph with post processor, results in AkFeld
  void AkFeldUmrechnen(ParserObj &FuObj);
  //! Load graph from text file or Editor string
  bool LoadGraph(const _TCHAR *Name,const _TCHAR *Editor,bool Sortieren,
                 bool InDatDatei,bool ExtDatDatei,bool NeuGraph = false,
                 bool ImagDataIncluded = true,bool HeaderIncluded = false,
                 NumListFormat *LDV = 0, bool XValuesIncluded = true, int XColumn = 1,
                 bool TabSeparator = true, bool SpaceSeparator = true);
  //! Write graph to text file or Editor string
  bool WriteGraph(const _TCHAR *Name,MGPStr *Editor,bool GRD_Format,bool TAB,
                  bool Komma,bool IngNotation);
  //! Load current graph again
  bool ReloadGraph();
  //! Update current graph or all graphs with YT data
  void ModifyGraph(const GraphDataObj *YT,bool AlleGr,bool CopyDaten);
  //! Copy all graphs and returns pointer to data block
  void *CopyAllGraphs(ParserObj &FuObj,int &size);
  //! Get graph with name
  GraphDataObj *FindGraph(const _TCHAR *Name);
};

// ----------------------------------------------------------------------------
// GraphladenData
// ----------------------------------------------------------------------------

//! Data structure for TEFGraphladenDlg
struct GraphladenData
{
  //! Graph load options
  enum GraphModes
  {
    GraphSortieren = 1,
    GraphInDatei   = 2
  };
  //! Graph load option
  int Mode;
  //! Graph file name
  MGPStr FileName;
  //! Decimal separator
  MGPStr DecSeparator;
  //! List separator
  MGPStr ListSeparator;
  //! Tab separator flag
  bool TabSeparator;
  //! Space separator flag
  bool SpaceSeparator;
  //! Imaginary data flag
  bool ImaginaryDataIncluded;
  //! Header flag
  bool HeaderIncluded;
  //! X values flag
  bool XValuesIncluded;
  //! X data column
  int XValuesColumn;
};

// ----------------------------------------------------------------------------
// DarstDataObj
// ----------------------------------------------------------------------------

//! Diagramm settings class.
/*!
    Holds values for diagram settings, including Y axis mode, X axis mode
    and legend options.
*/
class DarstDataObj
{
 public:
  //! Y axis modes
  enum DarstModes
  {
    realY = 0,
    imagY,
    cabY,
    logY,
    dbY,
    canY,
    ctiY,
    ortskurY,
    einzel,
    invisible
  };

  //! X axis modes
  enum XModes
  {
    linX = 0,
    logX
  };

  //! Auto calc / legend options
  enum AutoModes
  {
    dLegendeOben = 2,
    AutoX = 256,
    AutoY = 512
  };
  //! Diagram value range
  AxisDataArr XAxis,YAxis;
  //! Display conversion
  unsigned char Conversion;
  //! Diagram number (0 or 1)
  unsigned char DiagNr;
  //! Auto and legend modes
  int Mode;
  //! Diagram title
  SetString Titel;
  //! Number of digits for axis labels
  int NumDigits;
  //! Sets same scale for X and Y axis
  bool m_SameXAndYScaling;

  //! Ctor
  DarstDataObj() {}
  //! Initialize value range with defaults
  void InitWerber();
  //! Set up diagram
  void Init(int conversion,unsigned char DN,
            int MD,const _TCHAR *TT,int ND);
  //! Calculate value ranges from graph data
  bool ErmittleWerBer(int Mode,ParserObj &Psr,
                      GraphManagerObj *YDaten);
  //! Sort value ranges (Min < Max)
  void SortAxis(AxisData &WB);
  //! Check whether value range can be displayed
  bool GetBerOK(const AxisData &WB);
  //! Scale given value range
  bool ScaleBer(AxisData &ad,double scalefactor);
  //! Convert x and y value according to current mode
  double YUmrechnen(Cpx y,double &f);
  //! Return number of active axises
  size_t GetVisYAxisCount();
};

// ----------------------------------------------------------------------------
// DarstDataArr
// ----------------------------------------------------------------------------

//! Array for the two diagrams per graph window
class DarstDataArr : public CopyArr<DarstDataObj>
{
};

// ----------------------------------------------------------------------------
// EinstData
// ----------------------------------------------------------------------------

//! Data structure for diagram options
class EinstData
{
 public:
  //! Number of data points to calculate for new graphs
  unsigned int Punkte;
  //! Helper variable Hi settings
  MaTyp Ma;
  //! Flag for debug mode
  int Debug;
  //! Flag for automatic diagram update
  int AutoUpdate;
  //! X min value for graph calculation (optional)
  double CXMin;
  //! X max value for graph calculation (optional)
  double CXMax;
  //! Flag for using CXMin / CXMax and XLog for graph calculation
  int XMode;
  //! Flag for log. stepping for CXMin / CXMax
  int XLog;
};

// ----------------------------------------------------------------------------
// LegendeType
// ----------------------------------------------------------------------------

//! Class for legend type, based on FontType + Frame option
struct LegendeType : public FontType
{
  //! Flag for frame around legend
  int Frame;
};

// ----------------------------------------------------------------------------
// BildschOptObj
// ----------------------------------------------------------------------------

//! Class for editor display options
class BildschOptObj
{
  //! Options
  int BildschOpt;
 public:
  //! Options for function editor
  enum EditorOptionen
  {
    mBildschTeil    = 1,
    mBildschBruch   = 2,
    mBildschKlammer = 4,
    mHideDiags      = 8,
    mMarkLabels    = 16
  };
  //! Ctor
  BildschOptObj() {BildschOpt = 0;}
  //! Ctor
  BildschOptObj(int Opt) {BildschOpt = Opt;}
  //! Return value
  int Get() const {return BildschOpt;}
  //! Set new value
  void Set(int O) {BildschOpt = O;}
  //! Returns function editor bruch display flag
  inline bool GetBildschBruch() const
  {
    return ((BildschOpt & mBildschBruch) > 0);
  }
  //! Set function editor bruch display flag
  inline void SetBildschBruch(bool BildschBruch)
  {
    BildschBruch ? (BildschOpt |= mBildschBruch) :
                   (BildschOpt &= ~mBildschBruch);
  }
  //! Returns function editor bracket display flag
  inline bool GetBildschKlammer() const
  {
    return ((BildschOpt & mBildschKlammer) > 0);
  }
  //! Sets function editor bracket display flag
  inline void SetBildschKlammer(bool BildschKlammer)
  {
    BildschKlammer ? (BildschOpt |= mBildschKlammer) :
                     (BildschOpt &= ~mBildschKlammer);
  }
  //! Returns function editor hide diagrams flag
  inline bool GetHideDiags() const
  {
    return ((BildschOpt & mHideDiags) > 0);
  }
  //! Sets function editor hide diagrams flag
  inline void SetHideDiags(bool BildschDiags)
  {
    BildschDiags ? (BildschOpt |= mHideDiags) :
                   (BildschOpt &= ~mHideDiags);
  }
  //! Returns function editor mark labels flag
  inline bool GetMarkLabels() const
  {
    return ((BildschOpt & mMarkLabels) > 0);
  }
  //! Sets function editor mark labels flag
  inline void SetMarkLabels(bool ml)
  {
    ml ? (BildschOpt |= mMarkLabels) :
         (BildschOpt &= ~mMarkLabels);
  }
};

// ----------------------------------------------------------------------------
// AuflRec
// ----------------------------------------------------------------------------

//! Data structure for solver data
struct AuflRec
{
  //! Result, Keep and Keep-together strings
  MGPStr ErgStr,BeibehaltenStr,ZusammenStr;
  //! Additional string definitions
  MGStrListObj ParamsList;
  //! Solver modes
  unsigned short Mode,Bereich,Anh;
};

// ----------------------------------------------------------------------------
// AuflVorlageStruct
// ----------------------------------------------------------------------------

//! Extents AuflRec for templates (adds template name)
struct AuflVorlageStruct : AuflRec
{
  //! Template name
  MGPStr Name;
};

// ----------------------------------------------------------------------------
// AuflVorlageObj
// ----------------------------------------------------------------------------

//! Dynamic array of AuflRec objects
class AuflVorlageObj : public ListObj<AuflVorlageStruct>
{
};

// ----------------------------------------------------------------------------
// EFComboData
// ----------------------------------------------------------------------------

//! Holds history information for combo boxes
class EFComboData
{
 public:
  enum EFComboDataFlags
  {
    //! Max. number of history entries for EFComboData
    MaxComboItems = 9
  };
 protected:
  //! History strings
  MGPStr strings[MaxComboItems];
  //! Number of currently available history strings
  int count;

 public:
  //! Ctor
  EFComboData() {count = 0;};
  //! Sets strings[i] to str
  void SetString(int i,const _TCHAR *str)
  {
    if (i < MaxComboItems)
    {
      strings[i].SetStr(str);
      if (i >= count)
        count = i+1;
    }
  }
  //! Return strings[i]
  const _TCHAR *GetString(int i)
  {
    if (i < count)
      return strings[i].c_str();
    else
      return 0;
  }
  //! Set string count
  int SetCount(int i)
  {
    if (i <= MaxComboItems)
      count = i;
    else
      count = MaxComboItems;
    return count;
  }
  //! Return string count
  int GetCount()
  {
    return count;
  }
};

// ----------------------------------------------------------------------------
// NaeherungDataType
// ----------------------------------------------------------------------------

//! Structure for approximation functions
struct NaeherungDataType
{

  //! Result name and parameter strings for approximation function
  MGPStr ErgStr,ParamStr;
  //! X parameter for exponential approximation
  MGPStr XStr[3];
  //! Y parameter for exponential approximation
  MGPStr YStr[3];
  //! Mode and Options
  unsigned short Mode;
  //! Flag for data from current graph
  bool m_CurGraph;
  //! Number of points from current graph
  size_t m_NumPts;
  //! Set default values
  void Clear()
  {
    ErgStr.SetStr(_TEXT(""));
    ParamStr.SetStr(_TEXT(""));
    XStr[0].SetStr(_TEXT(""));
    XStr[1].SetStr(_TEXT(""));
    XStr[2].SetStr(_TEXT(""));
    YStr[0].SetStr(_TEXT(""));
    YStr[1].SetStr(_TEXT(""));
    YStr[2].SetStr(_TEXT(""));
    Mode = 0;
    m_CurGraph = false;
    m_NumPts = 10;
  }
};

#endif
