// Version 6.05  27.04.2008 Markus Greither
// Copyright (c) 1991-2008 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efgraphc.h
    \brief Header for the diagram window class
*/

#ifndef __EFGRAPHC__
#define __EFGRAPHC__

#include "efgrelem.h"

//! Data for TEFGraphControl classes and its undo items
class TEFGraphControlData
{
 public:
  //! Undo Items
  enum UndoItems
  {
    UndoPostProcFunction = 1,
    UndoDDiagData = 2,
    UndoDiagDataBase = 4
  };   
    
  //! Copy data from GCD, evtl. update GCD with current data      
  void CopyData(int UndoItems,TEFGraphControlData *GCD, bool UpdateGCD);
  //! Post processor function
  FunctionArrObj PostProcFunktion;
  //! Diagram data
  DiagDataObj DDiagData;
};  

// ----------------------------------------------------------------------------
// DiagUndoItem
// ----------------------------------------------------------------------------

//! Undo item for TEFGraphControl
class DiagUndoItem
{
 public:
  //! The changed items flags
  int UndoElems;
  //! Original item data
  TEFGraphControlData Data;
};

// ----------------------------------------------------------------------------
// DiagUndoArr
// ----------------------------------------------------------------------------

//! Dynamic array of DiagUndoItem objects
class DiagUndoArr : public ListObj<DiagUndoItem> {};


// ----------------------------------------------------------------------------
// TEFGraphControl
// ----------------------------------------------------------------------------

//! Diagram control class
/*!
    Holds 1 or 2 diagrams and 1 legend. Is based on wxResizeableControl for
    resizing and moving.
    Has printing support.
*/
class TEFGraphControl :public TEFTabControl, protected TEFGraphControlData
{
 public:
  //! Definition of graph clipboard format name
  static const wxChar *GraphClipFormat;
  //! Definition of diagram clipboard format name
  static const wxChar *DiagClipFormat;
  //! Extents wxResizeableControl move modes for moving graph and legend, zooming
  enum MoveModes
  {
    Graph = wxResizeableControl::MaxMode,
    GraphSelect,
    Legende,
    ZoomIn,
    Label
  };

  //! Ctor, don't use
  TEFGraphControl() : TEFTabControl(),
   LegendeFont(0),
   Diag1(0,0),Diag2(0,0),
   DiagrammCursor(0),LegendeCursor(0),
   AkLabel(0)
  {
  }
  //! Ctor
  TEFGraphControl(wxWindow *AParent, int AnId,DiagDataObj *DD,
                  const NumListFormat&LDV,
                  const wxPoint &pos,const wxSize &size,
                  long style = 0,const wxString &name = wxPanelNameStr);
  //! Create the window
  void Create(wxWindow *AParent, int AnId,DiagDataObj *DD,
              const NumListFormat&LDV,
              const wxPoint &pos,const wxSize &size,
              long style = 0,const wxString &name = wxPanelNameStr);
  //! Initialize window
  void Init(DiagDataObj *DD);
  //! Dtor
  virtual ~TEFGraphControl();
  //! Returns current functions title
  const _TCHAR *GetAkFuString();
  //! Activate / Deactivate this diagram
  void SetActive(bool akt);
  //! Set selected diagram
  void SetAkDiag(unsigned char newdiag,bool recordmacro = true);
  //! Return currently selected diagram
  inline EFDiagObj &GetAkDiag()
  {
    if (AkDiag == 0)
      return Diag1;
    else
      return Diag2;
  }
  //! Get diag of given number
  inline EFDiagObj *GetDiag(unsigned char D)
  {
    return(D == 0 ? &Diag1 : &Diag2);
  }
  //! Return whether this diagram window has two diagrams
  inline bool ZweiDiags()
  {
    return ((DDiagData.DarstData)[1].Conversion != DarstDataObj::einzel);
  }
  //! Undo or redo last action
  void UndoRedo(bool redo);
  //! Add new undo item
  void AddUndo(int UndoItems);
  //! Remove last undo item
  void DelUndo()
  {
    m_UndoItems.SetAkEint(m_UndoItems.GetMaxEintrag());
    m_UndoItems.LoeschAkEint();
    if (m_UndoCount > m_UndoItems.GetMaxEintrag())
      m_UndoCount = m_UndoItems.GetMaxEintrag();
  }    
  //! Initialize graphs
  void InitGraphs(int AkGraph,wxDC &DC);
  //! Draw graphs
  void GraphenAusgabe(wxDC &DC);
  //! Draw diagram window
  virtual void Paint(wxDC& dc, bool erase, wxSize& size);
  //! Draw diagram to printer DC
  void PrintDiag(wxDC &dc,int w,int h);
  //! Set up diagram scaling
  void SetScreenParams(wxDC &DC,int w = 0,int h = 0,bool DefaultWindowXYRatio = false);
  //! Generate diagram title
  void DiagTitel(DarstDataObj &DarstData,wxString &DiagTitel);
  //! Return graph data of function with name FuStr, if available
  const GraphDataObj *FuStringVorh(const _TCHAR *FuStr);
  //! Return diagram data object
  DiagDataObj* GetDiagData()
  {
    return &DDiagData;
  }
  //! Update diagrams value ranges after changes
  bool UpdateWerBer();
  //! Mark diagram for redraw
  void Invalidate();
  //! Verify graph name
  int CheckGraphName(MGPStr &FuStr,wxWindow *parent = 0,bool dragndrop = false);
  //! Send SetAkDiag event to view
  void SendSetAkDiag();
  //! Send Modified event to view
  void SendModified();
  //! Send Diagram modified event to view
  void SendDiagModified();
  //! Send GetFunctions event to view
  void SendGetFunktions();
  //! Returns the diagram to which Label belongs
  EFDiagObj *GetLabelDiag(LabelObj &Label)
  {
    return (Label.Flags & LabelObj::Diag2 ? GetDiag(1) : GetDiag(0));
  }
  //! Set label options
  bool LabelOptions(LabelObj &Label,LabelArrObj *org = 0);
  //! Sets active label
  bool SetActiveLabel(size_t index);
  //! Returns label index at position x,y if available
  size_t FindLabel(int x, int y);
  //! Recalculate label sizes
  void CalcLabelSizes(wxDC &dc);
  //! Returns position of Label
  wxPoint GetLabelPos(LabelObj &Label,bool offset = true);
  //! Returns size of Label
  wxSize GetLabelSize(LabelObj &Label)
  {
    wxSize size(Label.XSize,Label.YSize);
    return size;
  }
  //! Draws and / or resizes label
  void LabelSizeNDraw(wxDC &dc,LabelObj &Label,int x,int y,bool draw);
  //! Draw label
  void DrawLabel(wxDC &dc,LabelObj &Label,int x, int y)
  {
    LabelSizeNDraw(dc,Label,x,y,true);
  }
   //! Move label so that they are in the visible area of the window
  bool LabelIntoSight(LabelObj &Label);
  //! Move label to x,y
  void MoveLabel(LabelObj &Label,int x, int y)
  {
    EFDiagObj *diag = GetLabelDiag(Label);
    int xsize = diag->BildBer.XMax-diag->BildBer.XMin;
    int ysize = diag->BildBer.YMax-diag->BildBer.YMin;
    Label.XOffs += int(double(x)/xsize*1000);
    Label.YOffs += int(double(y)/ysize*1000);
  }
  //! Paste new graph from data
  bool PasteGraph(wxCustomDataObject &data,bool dragndrop = false);
  //! Returns diagram window width/height ration
  float GetRatio()
  {
    return !ZweiDiags() ? 1.414 : 1/1.414;
  }
  //! Enable / disable paste event
  void CeEditPaste(wxUpdateUIEvent &ce);
  //! Handle paste event
  void OnEditPaste(wxCommandEvent &event);

 protected:
  //! Show diagram options dialog, initialize specified page
  void DarstellungsDialog(int DlgPage,int YAxisNr);
  //! Draw diagram
  void BildschirmAusgabe(wxDC &DC,bool Update,bool Printing = false,bool DefaultWindowXYRatio = false);
  //! Draw labels
  void DrawLabels(wxDC &DC);
  //! Draw markers
  void DrawMarkers(wxDC &DC);
  //! Set marker visible state
  void SetMarkersVisible(bool vis,wxDC &DC);
  //! Update diagram
  void UpdateDiag(wxDC &DC,bool Printing = false,bool DefaultWindowXYRatio = false);
  //! Returns virtual window size
  wxSize GetWindowDimensions(bool Printing = false,bool DefaultWindowXYRatio = false);
  //! Copy current graph to cmp
  void CopyGraph(wxDataObjectComposite &cmp);
  //! Returns current graph in clipboard format
  wxCustomDataObject *GetGraphData();
  //! Copy diagram window to cmp
  void CopyDiag(wxDataObjectComposite &cmp);
  //! Draw active frame (red)
  void DrawActiveFrame(wxDC &dc,bool erase,bool Printing = false);
  //! Show current marker position
  void DiagCurPos(wxDC &DC);
  //! Show graph value table dialog
  int ShowValueTable(const wxChar *label,bool Neu);
  //! Show graph options dialog
  void GraphOptionen(bool Neu,bool onlyvaluetable = false);
  //! Delete graphs
  void GraphenLoeschen(bool Alle);
  //! Draw focus rectangle
  void FocusRect(wxDC &DC);
  //! Move Legende to new position 
  void MoveLegend(const wxPoint &hp,bool finished,bool domove = true);
  //! Move Label to new position
  void MoveLabel(const wxPoint &hp,bool finished,bool domove = true);

// Event Handlers

  //! Respond to undo event
  void OnEditUndo(wxCommandEvent &);
  //! Respond to redo event
  void OnEditRedo(wxCommandEvent &);
  //! Enable / disable undo event
  void CeEditUndo(wxUpdateUIEvent& ce);
  //! Enable / disable redo event
  void CeEditRedo(wxUpdateUIEvent& ce);
  //! Enables label menu entries if labels available
  void CeLabels(wxUpdateUIEvent &ce)
  {
    ce.Enable((FindFocus() == this) && (DDiagData.Labels.GetMaxEintrag() > 0));
  }
  //! Handle delete all labels event
  void OnDelAllLabels(wxCommandEvent &);
  //! Handle move labels into sight event
  void OnLabelsIntoSight(wxCommandEvent &);
  //! Handle add label event
  void OnAddLabel(wxCommandEvent &event);
  //! Handle delete label event
  void OnDelLabel(wxCommandEvent &event);
  //! Handle label options event
  void OnLabelOptions(wxCommandEvent &event);
  //! Handle second diagram enable / disable event
  void OnDiag2(wxCommandEvent &event);
  //! Check / uncheck second diagram menu entry
  void CeDiag2(wxUpdateUIEvent &ce)
  {
    ce.Enable(FindFocus() == this);
    ce.Check(ZweiDiags());
  }
  //! Handle graph options event
  void OnGraphOptionen(wxCommandEvent &event);
  //! Handle graph value table event
  void OnGraphTable(wxCommandEvent &event);
  //! Handle graph values event
  void OnGraphWerte(wxCommandEvent &event);
  //! Handle diagram options event
  void OnDarstellung(wxCommandEvent &event);
  //! Enable / disable diagram events
  void CeHasFocus(wxUpdateUIEvent &event)
  {
    event.Enable(FindFocus() == this);
  }
  //! Handle copy graph event
  void OnCopyGraph(wxCommandEvent &event);
  //! Handle copy event
  void OnEditCopy(wxCommandEvent &event);
  //! Enable copy event
  void CeEditCopy(wxUpdateUIEvent &ce)
  {
    ce.Enable(true);
  }
  //! Disable cut event
  void CeEditCut(wxUpdateUIEvent &ce)
  {
    ce.Enable(false);
  }
  //! Handle diagram marker event
  void OnMarke(wxCommandEvent &event);
  //! Enable / disable diagram marker
  void CeMarke(wxUpdateUIEvent &ce)
  {
    ce.Enable((FindFocus() == this) && DDiagData.YDaten.YDatenVorhanden());
    ce.Check(GetAkDiag().Mark2.Markiert());
  }
  //! Handle mark minimum event
  void OnMin(wxCommandEvent &event);
  //! Handle mark maximum event
  void OnMax(wxCommandEvent &event);
  //! Handle mark left event
  void OnLinks(wxCommandEvent &event);
  //! Handle mark right event
  void OnRechts(wxCommandEvent &event);
  //! Handle mark peak event
  void OnSpitze(wxCommandEvent &event);
  //! Handle graph delete event
  void OnLoeschen(wxCommandEvent &event);
  //! Handle delete all graphs event
  void OnAlleLoeschen(wxCommandEvent &event);
  //! Enable / disable graph menu entries
  void CeGraph(wxUpdateUIEvent &ce)
  {
    ce.Enable((FindFocus() == this) && DDiagData.YDaten.YDatenVorhanden());
  }
  //! Handle load graph event
  void OnGraphLaden(wxCommandEvent &event);
  //! Handle save graph event
  void OnGraphSpeichern(wxCommandEvent &event);
  //! Handle post processor event
  void OnPostProc(wxCommandEvent &event);
  //! Check / uncheck post processor menu entry
  void CePostProc(wxUpdateUIEvent &ce)
  {
    ce.Enable(FindFocus() == this);
    ce.Check(DDiagData.PostProc != 0);
  }
  //! Handle auto x event
  void OnAutoX(wxCommandEvent &event);
  //! Increase diagram range
  void OnIncrease(wxCommandEvent &event);
  //! Decrease diagram range
  void OnDecrease(wxCommandEvent &event);
  //! Check / uncheck auto x menu entry
  void CeAutoX(wxUpdateUIEvent &ce)
  {
    ce.Enable(FindFocus() == this);
    ce.Check(((DDiagData.AutoMode) & DarstDataObj::AutoX) != 0);
  }
  //! Handle auto y event
  void OnAutoY(wxCommandEvent &event);
  //! Check / uncheck auto y menu entry
  void CeAutoY(wxUpdateUIEvent &ce)
  {
    ce.Enable(FindFocus() == this);
    ce.Check(((DDiagData.AutoMode) & DarstDataObj::AutoY) != 0);
  }
  //! Handle legend on top event
  void OnLegendeOben(wxCommandEvent &event);
  //! Check / uncheck legend on top menu entry
  void CeLegendeOben(wxUpdateUIEvent &ce)
  {
    ce.Enable(FindFocus() == this);
    ce.Check((DDiagData.AutoMode & DarstDataObj::dLegendeOben) != 0);
  }
  //! Handle fourier dialog event
  void OnFourierDialog(wxCommandEvent &event);
  //! Check / uncheck fourier dialog event
  void CeFourierDialog(wxUpdateUIEvent &ce)
  {
    ce.Enable(FindFocus() == this);
    ce.Check(DDiagData.Ma.Mode == MaTyp::ModeFourier);
  }
  //! Enable / Disable delete diagram menu entry
  void CeDelDiag(wxUpdateUIEvent &ce);
  //! Handle delete diagram event
  void OnDelDiag(wxCommandEvent &event);
  //! Enable / Disable move legend event
  void CeMoveLegend(wxUpdateUIEvent &ce)
  {
    ce.Enable((FindFocus() == this) && DDiagData.YDaten.YDatenVorhanden());
    ce.Check(m_movemode == Legende);
  }
  //! Handle move legend event
  void OnMoveLegend(wxCommandEvent &event);
  //! Enable / Disable move label event
  void CeMoveLabel(wxUpdateUIEvent &ce)
  {
    ce.Enable((FindFocus() == this) && (DDiagData.Labels.GetMaxEintrag() > 0));
    ce.Check(m_movemode == Label);
  }
  //! Handle move label event
  void OnMoveLabel(wxCommandEvent &event);
  //! Handle select label event
  void OnLabelSelect(wxCommandEvent &event);
  //! Handle left button up event
  void OnLButtonUp(wxMouseEvent &event);
  //! Handle left button down event
  void OnLButtonDown(wxMouseEvent &event);
  //! Select graph whose name is under the cursor
  void SelectGraphFromPos(wxDC &DC,wxPoint hp);
  //! Handle right button up event
  void OnContextMenu(wxMouseEvent &event);
  //! Handle mouse move event
  void OnMouseMove(wxMouseEvent &event);
  //! Handle key down event
  void OnKeyDown(wxKeyEvent &event);
  //! Handle left button double click event
  void OnLButtonDblClk(wxMouseEvent &Event);
  //! Handle set cursor event (?)
  void OnSetCursor(wxSetCursorEvent &event);
  //! Handle set focus event
  void OnSetFocus(wxFocusEvent &event);
  //! Handle kill focus event
  void OnKillFocus(wxFocusEvent &event);
  //! Handle paint event
  void OnPaint(wxPaintEvent &event);
  //! Handle size event
  void OnSize(wxSizeEvent &event);
  //! Select diaggram
  void OnMacroSelectDiagram(wxCommandEvent &event);
  //! Update graph options
  void OnMacroGraphOptions(wxCommandEvent &event);
  //! Update diagram settings
  void OnMacroDarstellung(wxCommandEvent &event);
  //! Set active graph
  void OnMacroSelectGraph(wxCommandEvent &event);
  //! Set active label
  void OnMacroSelectLabel(wxCommandEvent &event);
  //! Select diag at position hp
  int MarkDiag(const wxPoint &hp,wxDC &DC,
               bool GetSelDiag,unsigned char *NewDiag = 0);
  //! Returns true if legend on top
  bool LegendeOben()
  {
    return ((DDiagData.AutoMode & DarstDataObj::dLegendeOben) > 0);
  }
  //! Select graph, supports macro
  void SetActiveGraph(int graph,wxDC &dc);

  //! Legend font
  wxFont *LegendeFont;
  //! Function parser
  ParserDoubleObj Parser;
  //! Diagrams
  EFDiagObj Diag1,Diag2;
  //! Legend
  FunkMenueObj FunkMenue;
  //! Diagram and legend cursors
  wxCursor *DiagrammCursor,*LegendeCursor;
  //! Active flag
  bool Aktiv;
  //! Undo / redo items
  DiagUndoArr m_UndoItems;
  //! Current undo / redo item
  int m_UndoCount;
  //! Screen up to date flag
  bool Painted;
  //! Flag for cursor in menu
  bool MenuCursor;
  //! Flag for diagram in menu deleted
  bool DelDiag;
  //! Pointer to active label
  LabelObj *AkLabel;
  //! Current diagram number
  unsigned char AkDiag;
  //! Flag for same X and Y axis scaling
  bool m_SameXAndYScaling;

  DECLARE_EVENT_TABLE()
  DECLARE_DYNAMIC_CLASS(TEFGraphControl)
};

#endif

