// Version 5.26  09.07.2004 Markus Greither
// Copyright (c) 1991-2004 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efkrzn.h
    \brief Header for the elimination part of the symbolic function solver
*/

#ifndef __EFKRZN__
#define __EFKRZN__


// ----------------------------------------------------------------------------
// ShortArrKettObj
// ----------------------------------------------------------------------------

//! Dynamic array of ShortArrObjs
class ShortArrKettObj :public KettObj<ShortArrObj>
{
};

// ----------------------------------------------------------------------------
// IntArrKettObj
// ----------------------------------------------------------------------------

//! Dynamic array of IntArrObjs
class IntArrKettObj :public KettObj<IntArrObj>
{
};

// ----------------------------------------------------------------------------
// BitArrObj
// ----------------------------------------------------------------------------

//! Efficient class for arrays of type boolean
/*!
    Has members for increment of array
    (useful for varying of options).
*/
// With VECTOR_BITS == 1, solver time about 3 times as with VECTOR_BITS == 0 !!!
#define VECTOR_BITS 0

#if VECTOR_BITS
class BitArrObj : protected vector<bool>
{
  //! Maximum bit, number of set bits
  unsigned int MxBit,AnzBit;

 public:
  //! Ctor, creates data of 1 byte and no bit set
  BitArrObj()
  {
    MxBit = 0;
    AnzBit = 0;
  }
  //! Copy Ctor
  BitArrObj(const BitArrObj &other)
  {
    MxBit = 0;
    AnzBit = 0;
    *this = other;
  }
  //! Dtor
  virtual ~BitArrObj()
  {
    clear();
  }
  //! Set given bit
  void SetBit(unsigned int Bit)
  {
    if (Bit >= size())
      resize(Bit+1,false);
    if (!BitSet(Bit))
    {
      if (MxBit < Bit)
        MxBit = Bit;
      AnzBit++;
    }
    at(Bit) = true;
  }
  //! Return state of given bit
  bool BitSet(unsigned int Bit) const
  { // ermittelt, ob das Bit an pos Bit gesetzt ist
    if (Bit < size())
    {
      return at(Bit);
    }
    else
      return false;
  }
  //! Returns true if bits from Bit1 to Bit2 are all set
  bool BitsSet(unsigned int Bit1,unsigned int Bit2) const;
  //! Clear given bit (Set false)
  void ClearBit(unsigned int Bit)
  { // Setzt das Bit an pos Bit auf 0
    if (Bit >= size())
      resize(Bit+1,false);
    if (BitSet(Bit))
    {
      if ((MxBit > 0) && (MxBit == Bit))
        GetMaxBit(Bit-1);
      AnzBit--;
    }
    at(Bit) = false;
  }
  //! Toggle given bit state
  void ToggleBit(unsigned int Bit)
  {
    if (BitSet(Bit))
      ClearBit(Bit);
    else
      SetBit(Bit);
  }
  //! Increment by one (treat bit array as large integer)
  void Inc();
  //! Shift all bits to left
  void Shl();
  //! Clear all bits
  void Clear()
  {
    clear();
    MxBit = 0;
    AnzBit = 0;
  }
  //! Returns the number of highest set bit
  unsigned int MaxBit() const
  {
    return MxBit;
  }
  //! Returns the number of set bits
  unsigned int AnzBits(unsigned int Anf = 0,int End = -1)
  {
    GetAnzBits(Anf,End);
    return AnzBit;
  }
  //! Determine if the same bits are set in B2 and here
  unsigned int OnlyCommonBits(const BitArrObj &B2,unsigned int Anf = 0) const;
  //! Determine of all set bits in B2 are also set here
  unsigned int CommonBits(const BitArrObj &B2,unsigned int Anf = 0) const;
  //! Set all bits that are also set in B2
  void SetBits(const BitArrObj &B2);
  //! Delete bit at given position, shift all higher bits one down
  void DelBit(unsigned int Bit);
  //! Determine if this and B2 are identical
  bool Identical(const BitArrObj &B2) const;
  //! Copy bits from AnfBit to EndBit into B2
  void CopyBits(BitArrObj &B2,unsigned int AnfBit,unsigned int EndBit);
  //! Make string of this bit array
  void BitString(MGPStr &res);
  //! Determine the first common bit starting at Anf, return -1 if none
  int FirstCommonBit(BitArrObj &B2,unsigned int Anf = 0);

 protected:
  //! Determine maximum set bit from start position StartWert
  void GetMaxBit(unsigned int StartWert);
  //! Determine the number of set bits between Anf and End
  void GetAnzBits(unsigned int Anf,int End);
};
#else
class BitArrObj
{
  //! Pointer to data
  unsigned char *Data;
  //! Number of bytes, maximum bit, number of set bits
  unsigned int NBytes,MxBit,AnzBit;

 protected:
  //! Initialize
  void Init()
  {
    MxBit = 0;
    AnzBit = 0;
    NBytes = 1;
    Data = GetMem(NBytes);
  }
  //! Change data size
  unsigned char *GetMem(unsigned int i)
  {
    unsigned char *D1 = new unsigned char[i];
    memset(D1,0,i);
    return D1;
  }
  //! Increase data size
  void IncMem(unsigned int Bts)
  {
    unsigned char *D1 = GetMem(Bts);
    memcpy(D1,Data,NBytes);
    delete [] Data;
    Data = D1;
    NBytes = Bts;
  }
 public:
  //! Ctor, creates data of 1 byte and no bit set
  BitArrObj()
  {
    Init();
  }
  //! Copy Ctor
  BitArrObj(const BitArrObj &other)
  {
    Init();
    *this = other;
  }
  BitArrObj &operator = (const BitArrObj &other)
  {
    SetBits(other,true);
    return *this;
  }  
  //! Dtor
  virtual ~BitArrObj()
  {
    delete [] Data;
  }
  //! Set given bit
  void SetBit(unsigned int Bit)
  {
    if (!BitSet(Bit))
    {
      if (MxBit < Bit)
        MxBit = Bit;
        AnzBit++;
    }
    if ((Bit>>3) >= NBytes)
      IncMem((Bit>>3)+1); // Setzt das Bit an pos Bit
    Data[Bit>>3] |= (unsigned char)(1 << (Bit & 7));
  }
  //! Return state of given bit
  bool BitSet(unsigned int Bit) const
  { // ermittelt, ob das Bit an pos Bit gesetzt ist
    if ((Bit>>3) < NBytes)
      return ((Data[Bit>>3] & (1 << (Bit & 7))) > 0);
    else
      return false;
  }
  //! Returns true if bits from Bit1 to Bit2 are all set
  bool BitsSet(unsigned int Bit1,unsigned int Bit2) const;
  //! Clear given bit (Set false)
  void ClearBit(unsigned int Bit)
  { // Setzt das Bit an pos Bit auf 0
    if (BitSet(Bit))
    {
      if ((MxBit > 0) && (MxBit == Bit))
        GetMaxBit(Bit-1); AnzBit--;
    }
    Data[Bit>>3] &= (unsigned char)(~(1 << (Bit & 7)));
  }
  //! Toggle given bit state
  void ToggleBit(unsigned int Bit)
  {
    if (BitSet(Bit))
      ClearBit(Bit);
    else
      SetBit(Bit);
  }
  //! Increment by one (treat bit array as large integer)
  void Inc();
  //! Shift all bits to left
  void Shl();
  //! Clear all bits
  void Clear()
  {
    delete [] Data;
    Init();
  }
  //! Returns the number of highest set bit
  unsigned int MaxBit() const
  {
    return MxBit;
  }
  //! Returns the number of set bits
  unsigned int AnzBits(unsigned int Anf = 0,int End = -1)
  {
    GetAnzBits(Anf,End);
    return AnzBit;
  }
  //! Determine if the same bits are set in B2 and here
  unsigned int OnlyCommonBits(const BitArrObj &B2,unsigned int Anf = 0) const;
  //! Determine of all set bits in B2 are also set here
  unsigned int CommonBits(const BitArrObj &B2,unsigned int Anf = 0) const;
  //! Set all bits that are also set in B2
  void SetBits(const BitArrObj &B2,bool exactcopy = false);
  //! Delete bit at given position, shift all higher bits one down
  void DelBit(unsigned int Bit);
  //! Determine if this and B2 are identical
  bool Identical(const BitArrObj &B2) const;
  //! Copy bits from AnfBit to EndBit into B2
  void CopyBits(BitArrObj &B2,unsigned int AnfBit,unsigned int EndBit);
  //! Make string of this bit array
  void BitString(MGPStr &res);
  //! Determine the first common bit starting at Anf, return -1 if none
  int FirstCommonBit(BitArrObj &B2,unsigned int Anf = 0);

 protected:
  //! Determine maximum set bit from start position StartWert
  void GetMaxBit(unsigned int StartWert);
  //! Determine the number of set bits between Anf and End
  void GetAnzBits(unsigned int Anf,int End);
};
#endif
// ----------------------------------------------------------------------------
// BitArrKettObj
// ----------------------------------------------------------------------------

//! Dynamic array of BitArrObjs
class BitArrKettObj :public KettObj<BitArrObj>
{
};

// ----------------------------------------------------------------------------
// PrimZahlObj
// ----------------------------------------------------------------------------

//! Class for prime numbers
class PrimZahlObj : protected BitArrObj
{
 public:
  //! Ctor
  PrimZahlObj(long count);
  //! Return true if n is prime number
  bool IstPrim(long n)
  {
    return BitSet(n);
  }
};

// ----------------------------------------------------------------------------
// SumFakMatrixObj
// ----------------------------------------------------------------------------

//! Class for factor / summant matrix
/*!
    Build the matrix out of a given left and right value string.
    Used for subexpression elimination.
*/
class SumFakMatrixObj : public SolverTools
{
 // Matrix fuer Faktoren und Summanten
 // Summanten als Index der ShortArrs
 // Faktoren als Index des KettArrs
 public:
  //! Ctor, initializes SolverTools base class
  SumFakMatrixObj(ParserObj *P,MGPStr *Dbg) : SolverTools(P,Dbg) {}

  //! Array for factor counts.
  /*! One entry for each factor.
      Each entry has 1 integer value per Summant that holds the count of this factor in this summant.
      Multiplication and division of same factor is handled separately, like different factors.
  */
  ShortArrKettObj AnzProSum;
  //! Array for factor strings. Each factor is preceded with either * or /.
  MGStrObj Faktor;
  //! Counts of summants in RAbsStr and LAbsStr
  int AnzRSum,AnzLSum;
  //! Signs of summants (Vorz[index] holds sign for Summant number index+1 in AnzProSums)
  MGPStr Vorz;
  //! Debug-Function to display the matrix
  void MatrixOut(const _TCHAR *Filename = 0);
  //! Create the matrix from given strings
  void Create(MGPStr &LAbsStr,MGPStr &RAbsStr);
};

// ----------------------------------------------------------------------------
// FaktorVar
// ----------------------------------------------------------------------------

//! Class for Factor variation
/*
    Used to test out all available numbers of this
    factor during subexpression elimination.
*/
class FaktorVar
{
  //! Count of available summants
  int AnzSum;
  //! Max count of this factor in all summants, equals line count
  int AnzLines;
  //! Factor count for each summant for this factor
  /*! Each entry holds the factor count for the Summant of its index.
  */
  ShortArrObj FaktorVert;
  //! Index to the lines for this factor in BitArr
  ShortArrObj ZeilenIndex;
  //! Variant state
  /*! One entry per factor,
      Multiple lines for factors with count > 1.
      Each bit activates this factor in summant of the bits index.
  */
  BitArrKettObj FaktorVars;
  //! Debug output of current combination
  void KombOut(ShortArrObj &FaksProLine);

 public:
  //! Ctor
  FaktorVar() {};
  //! Returns line count
  int GetAnzLines()
  {
    return AnzLines;
  }
  //! Initialize this factor variation
  void Init(ShortArrObj &FaktorV, MGPStr &FakStr,BitArrKettObj &FakProSums,
            MGStrObj &FakStrings,ShortArrObj &FPL,int ASum);
  //! Determine next vartiant
  bool NextVar(BitArrKettObj &FakProSums,ShortArrObj &FakProLines);
  //! Initialize FaktorVars for the given Summant
  void InitFaktor(int Summant,BitArrKettObj &FakProSums,ShortArrObj &FakProLines);
};

// ----------------------------------------------------------------------------
// FaktorVarKettObj
// ----------------------------------------------------------------------------

//! Dynamic array of FaktorVar objects
class FaktorVarKettObj :public KettObj<FaktorVar>
{
};

// ----------------------------------------------------------------------------
// FaktorenVar
// ----------------------------------------------------------------------------

//! Class for factorgroup variants
/*!
    This class is used to test out all possible combinations
    of available factors during subexpression elimination.
*/
class FaktorenVar
{
  //! Assignment operator
  void operator = (const FaktorenVar &) {}
  //! Factor count for each line in FakProSums
  ShortArrObj FaksProLine;
  //! Reference to matrix for current equation
  SumFakMatrixObj &Matrix;
  //! Factor variations
  /*! One entry per summant,
      Each bit activates this factor of the bits index in the variation.
  */
  FaktorVarKettObj FaktorVars;
  //! Factor summants array
  BitArrKettObj FakProSums;
  //! Factor strings, one for each entry of FakProSums
  MGStrObj FakStrings;

 public:
  //! Ctor
  FaktorenVar(SumFakMatrixObj &Matrix);
  //! Determines number of combinations (roughly !!)
  double GetAnzKombs();
  //! Determines next variant
  bool NextVar();
  //! Check whether current variant is valid
  bool CheckVar();
  //! Verify signs
  bool CheckVorz(BitArrObj &FgKomb,BitArrKettObj &FaktorGr,ShortArrKettObj &FI,
                 MGPStr &ElimStr);
  //! Determine factor groups
  void FaktorenGruppen(BitArrKettObj &FG,ShortArrKettObj &FI);
  //! Perform the elimination
  bool Kuerzen(MGPStr &LAbsStr,MGPStr &RAbsStr,MGPStr &ElimStr);
  //! Debut output
  void MatrixOut(BitArrKettObj &AkArr);
  //! Debug function for FaktorenGruppen
  void FGOut(ShortArrKettObj &FI);
};

#endif

