// Version 6.07 12.03.2011 Markus Greither
// Copyright (c) 1991-2011 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efmacro.h
    \brief Header for the macro recorder class
*/

#include "efxmlio.h"

//! Event record
struct MacroCommandEntry
{
  //! Event literal name
  MGPStr m_Name;
  //! Event XRC identifier
  int m_XRCId;
};

//! List of event records
typedef ListObj<MacroCommandEntry> MacroCommandList;

//! Class for macro recording in XML
class MyMacroRecorder
{
  //! Path of macro file, used for submacros without path
  MGPStr m_Path;
  //! Lock status
  bool m_Locked;
  //! Stepping flag
  bool m_Stepping;
  //! XML document for loaded or recorded macros
  EFXmlDocument m_Doc;
  //! Current XML node
  EFXmlNode *m_AkNode;
  //! Flag for playback mode
  bool m_Playback;
  //! Flag for recording mode
  bool m_Recording;
  //! Repeat count for macros
  size_t m_RepeatCount;
  //! Pointer to currently active event list
  static MacroCommandList *ms_CurrentList;
  //! Static list of view events
  static MacroCommandList *ms_ViewCommandList;
  //! Static list of graphcontrol events
  static MacroCommandList *ms_GraphCCommandList;
  //! Static list of edit events
  static MacroCommandList *ms_EditCommandList;
  //! Static list of manually added events
  static MacroCommandList *ms_ManualCommandList;
  //! Original recorder instance for submacro
  static MyMacroRecorder *ms_OrgInstance;
  //! Recorder instance
  static MyMacroRecorder *ms_Instance;
  //! Return property if available
  EFXmlProperty *GetProperty(EFXmlProperty *prop,const wxChar *name);
  //! Iterator for command lists
  MacroCommandList *GetNextList(MacroCommandList *last)
  {
    if (!last)
      return ms_ViewCommandList;
    else
    if (last == ms_ViewCommandList)
      return ms_GraphCCommandList;
    else
    if (last == ms_GraphCCommandList)
      return ms_EditCommandList;
    else
    if (last == ms_EditCommandList)
      return ms_ManualCommandList;
    else
      return 0;
  }

 public:
  //! Ctor
  MyMacroRecorder();
  //! Record this event name and id, return XRCID of it
  static int GetXrcId(const char *name,int id = 0);
  //! Debug function for saving the macro command list
  void SaveMacroList(const char *fname);
  //! Delete macro lists
  void DeleteLists()
  {
    delete ms_ViewCommandList;
    delete ms_GraphCCommandList;
    delete ms_EditCommandList;
    delete ms_ManualCommandList;
  }
  //! Initialize manual command list
  void InitManualList()
  {
    ms_ManualCommandList = new MacroCommandList;
    ms_CurrentList= ms_ManualCommandList;
    GetXrcId("wxID_REVERT",wxID_REVERT);
  }
  //! Initialize command list with given name
  static void InitList(const char *name);
  //! Return property value as int
  bool GetPropertyInt(EFXmlProperty *prop,const wxChar *name,int &ret);
  //! Factory member
  static MyMacroRecorder *Get()
  {
    if (!ms_Instance)
      ms_Instance = new MyMacroRecorder;
    return ms_Instance;
  }
  //! Start macro recording
  void RecordMacro();
  //! End macro recording or playback, save macro to file name
  bool EndMacro(const char *name = 0);
  //! Add event to macro, use command lists to translate event XRC id into literal
  void AddEvent(wxEvent &event,int Id);
  //! Add event to macro
  void AddEvent(const wxChar *name,int Id);
  //! Add position properties to current node
  void AddPos(int x1,int y1,int x2,int y2);
  //! Get position properties from current node
  bool GetPos(EFXmlProperty *prop,int &x1,int &y1,int &x2,int &y2);
  //! Record text input
  void InsertString(const wxChar *str,int Id,int Pos,int Zeile);
  //! Record character input
  void InsertChar(const wxChar chr,int Id,int Pos, int Zeile);
  //! Return playback flag
  bool IsPlaying() {return m_Playback;}
  //! Return recording flag
  bool IsRecording() {return m_Recording;}
  //! Return step mode flag
  bool IsStepping() {return m_Stepping;}
  //! Return current node
  EFXmlNode *GetAkNode();
  //! Load submacro from file for playback
  bool PlaySubMacro(const char *name);
  //! Load macro from file for playback
  bool PlayMacro(const char *name);
  //! Iterator for XML nodes
  EFXmlNode *GetNextNode();
  //! Set lock mode
  bool SetLock(bool l)
  {
    bool ret = m_Locked;
    m_Locked = l;
    return ret;
  }  
};

//! Creates the instance of the macro recorder
class MyMacroRecorderCreator
{
 public:
  //! Ctor, create instance
  MyMacroRecorderCreator()
  {
    MyMacroRecorder::Get();
    MyMacroRecorder::Get()->InitManualList();
#ifdef __WXDEBUG__
    MyMacroRecorder::Get()->SaveMacroList("test.txt");
#endif
  }
  //! Dtor, delete instance
  ~MyMacroRecorderCreator()
  {
    MyMacroRecorder::Get()->DeleteLists();
    delete MyMacroRecorder::Get();
  }
};

//#define EF_DISABLE_MACRO

#ifdef EF_DISABLE_MACRO
// Code to disable theMacroRecorder

#define BEGIN_MACRO_EVENT_TABLE(theClass, baseClass, theMacroRecorder) BEGIN_EVENT_TABLE(theClass, baseClass)

#define MACROID(name) XRCID(name)
#define WXMACROID(name) name

#else

#define BEGIN_MACRO_EVENT_TABLE(theClass, baseClass, theMacroRecorder) \
MacroCommandList MacroCommandList##theClass; \
class MacroListCreator##theClass { \
 public: \
  MacroListCreator##theClass () { \
    theMacroRecorder::InitList(#theClass); \
  } \
} InstanceOfMacroListCreator##theClass; \
BEGIN_EVENT_TABLE(theClass, baseClass)

#define MACROID(name) MyMacroRecorder::GetXrcId(name)

#define WXMACROID(name) MyMacroRecorder::GetXrcId(#name,name)

#endif
