// Version 6.05  19.07.2008 Markus Greither
// Copyright (c) 1991-2008 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efmcxio.h
    \brief Header for the mcx format data storage classes
*/

#ifndef __EFMCXIO__
#define __EFMCXIO__

// ----------------------------------------------------------------------------
// EFDataIOObj
// ----------------------------------------------------------------------------

//! File I/O class.
/*!
    Provides methods for writing data to MCX-files.
*/
class EFDataIOObj
{
  //! Make brakets around the parameter string
  static void KlammernUmParamStr(MGPStr &str);
 public:
  //! Load functions from file Datei
  static void LoadFunktionen(FunctionArrObj *Funktion,int i,ParserObj &Psr,
                             istream &Datei,bool Equations = false,
                             MGPStr *PostTerm = 0);
  //! Read number from string
  double Eval(const _TCHAR *Wert,bool Kompat = false);
  //! Write label, index and double value in engineers notation into file Datei
  void IngFormIndOut(ostream &Datei,const char *Label,int index,double Wert);
  //! Write label and double value in engineers notation into file Datei
  void IngFormOut(ostream &Datei,const char *Label,double Wert);
  //! Write label, index and value Wert into file Datei
  template<class T> void ValueIndOut(ostream &Datei,const char *Label,
                                     int index,T Wert);
  //! Write label and value Wert into file Datei
  template<class T> void ValueOut(ostream &Datei,const char *Label,T Wert);

#ifdef __DMC__
  //! Write label, index and value Wert into file Datei
  void FontIndOut(ostream &Datei,const char *Label,int index,FontType Wert);
  //! Write label and value Wert into file Datei
  void FontOut(ostream &Datei,const char *Label,FontType Wert);
#endif
};

// ----------------------------------------------------------------------------
// DiagDataObj
// ----------------------------------------------------------------------------

//! Base class for DiagDataObj, without GraphManagerObj, so that copies are small
class DiagDataBaseObj : public EinstData
{
 public:
  //! Flag for post processor active
  int PostProc;
  //! Post processor function
  SetString PostProcTerm;
  //! Array of diagram labels
  LabelArrObj Labels;
  //! Legend
  LegendeType Legende;
  //! Axis font
  FontType Achse;
  //! Title font
  FontType Titel;
  //! Axis number font
  FontType AchsLabels;
  //! The two diagrams
  DarstDataArr DarstData;
  //! Automatic update mode
  int AutoMode;
  //! Number of current graph
  int AkGraph;
  //! Diagram window size
  int WinXPos,WinYPos,WinHeight,WinWidth;
  //! Legend position
  int LegX,LegY;
  //! Color
  unsigned long m_Colour;
  //! Sets same scale for X and Y axis
  int m_SameXAndYScale;

  //! Load legend information from XML file
  void XmlLoadLegend(EFXmlNode *node);
  //! Write legend information to XML file
  void XmlWriteLegend(EFXmlNode *node,DiagDataBaseObj *old);
  //! Save diagram in XML file
  void XmlSpeichern(EFXmlNode *n,DiagDataBaseObj *orig = 0);
};

//! Class for diagram window data (two diagrams + legend + labels)
class DiagDataObj : public EFDataIOObj, public DiagDataBaseObj
{
  enum
  {
    NumDiagNamen = 64
  };
  //! Names for diagram MCX file entries
  static const char *DiagNamen[];

 protected:
  //! Initialize
  void Clear();
 public:
  //! Graph data
  GraphManagerObj YDaten;

  //! Ctor, sets number format
  DiagDataObj(const NumListFormat &LDV)
  {
    Clear();
    YDaten.SetLDV(LDV);
  }
  //! Ctor
  DiagDataObj()
  {
    Clear();
  }
  //! Copy diagram data (clipboard copyable part), moves data if MoveData = true
  void Copy(DiagDataObj &DDO,bool CopyPos = true,bool MoveData = false);
  //! Save diagram to file Datei
  void Speichern(ostream &Datei);
  //! Save diagram in XML file
  void XmlSpeichern(EFXmlNode *n,DiagDataObj *orig = 0);
  //! Load diagram from XML file
  void XmlLaden(EFXmlNode *n);
  //! Set diagram value Name from SWert
  void SetData(const char *Name,const char *SWert,int Nummer,istream &Datei,
               bool Kompat,int MajorVersion);
  //! Check diagram setup
  void VerifyData();
  //! Set post processor flag
  inline void SetPostProc(int PP)
  {
    PostProc = PP;
    YDaten.SetPostProc(PP);
  }
  //! Returns post processor flag
  inline int GetPostProc()
  {
    return PostProc;
  }
  //! Sets auto mode
  void SetAuto(int A)
  {
    AutoMode = A;
  }
};

// ----------------------------------------------------------------------------
// PictureObj
// ----------------------------------------------------------------------------

//! Class for picture objects.
/*!
    Holds EnhMetaFile/Bitmap pictures.
*/
class PictureObj :public EFDataIOObj
{
  //! Name count
  enum
  {
    NumPictNamen = 7
  };
  //! Names for PictureObj MCX file entries
  static const char *PictNamen[];

 public:
  //! Picture type name
  MGPStr TypeStr;
  //! Picture size in bytes
  int Size;
  //! Data block
  char *Data;
  //! Window size and position
  int WinXPos,WinYPos,WinHeight,WinWidth;

  //! Copy Ctor
  PictureObj(const PictureObj &Other)
  {
    Size = Other.Size;
    TypeStr = Other.TypeStr;
    if (Size)
    {
      Data = new char[Size];
      memcpy(Data,Other.Data,Size);
    }
    else
      Data = 0;
  }
  //! Ctor
  PictureObj()
  {
    Size = 0;
    Data = 0;
  }
  //! Dtor
  virtual ~PictureObj()
  {
    delete[] Data;
  }
  //! Set picture value Name frome SWert
  bool SetData(const char *Name,const char *SWert,int Nummer,istream &Datei);
  //! Save picture to file Datei
  void Speichern(ostream &Datei);
  //! Save picture in XML file
  void XmlSpeichern(EFXmlNode *n);
  //! Load picture from XML file
  void XmlLaden(EFXmlNode *n);
};

// ----------------------------------------------------------------------------
// DiagDataArrObj
// ----------------------------------------------------------------------------

//! Dynamic array of DiagDataObj objects
class DiagDataArrObj :public ListObj<DiagDataObj>
{
  //! Number format
  NumListFormat AkLDV;

 protected:
  //! Initializes new entry
  DiagDataObj *InitFelder()
  {
    return new DiagDataObj(AkLDV);
  }

 public:
  //! Ctor, sets number format
  DiagDataArrObj(const NumListFormat &LDV,bool /*neueint*/ = true)
  {
    AkLDV = LDV;
  }
  //! Returns current diagram data
  DiagDataObj *AkDiagData()
  {
    return &GetAkEint();
  }
};

//! Dynamic array of DiagDataObj * objects
class DiagDataPtArrObj :public ListObj<DiagDataObj *>
{
};

// ----------------------------------------------------------------------------
// PictureArrObj
// ----------------------------------------------------------------------------

//! Dynamic array of picture objects
class PictureArrObj :public ListObj<PictureObj>
{
 public:
  //! Return current picture
  PictureObj *AkPicture()
  {
    return &GetAkEint();
  }
};

// ----------------------------------------------------------------------------
// EquationsObj
// ----------------------------------------------------------------------------

//! Class for equation windows
class EquationsObj :public EFDataIOObj
{
  enum
  {
    //! Name count
    NumEqNamen = 14
  };
  //! Names for EquationObj MCX file entries
  static const char *EqNamen[];

 public:
  //! Ctor
  EquationsObj()
  {
    BildschOpt.Set(BildschOptObj::mBildschBruch+BildschOptObj::mMarkLabels);
    ScrollPosX = ScrollPosY = 0;
    SelAnf = SelEnd = 0;
  }
  //! Equations string
  MGPStr Equations;
  //! Font for equations
  FontType Font;
  //! Window size and position
  int WinXPos,WinYPos,WinHeight,WinWidth;
  //! Function editor options
  BildschOptObj BildschOpt;
  //! Scroll and selection positions
  int ScrollPosX,ScrollPosY;
  //! Selection start and end
  unsigned int SelAnf,SelEnd;

  //! Set equations value Eq from SWert
  bool SetData(const char *Eq,const char *SWert,
               int /*Nummer*/,istream &Datei);
  //! Save equations to file Datei
  void Speichern(ostream &Datei);
  //! Save equations in XML file
  void XmlSpeichern(EFXmlNode *n);
  //! Load equations from XML file
  void XmlLaden(EFXmlNode *n);
};

// ----------------------------------------------------------------------------
// EquationsArrObj
// ----------------------------------------------------------------------------

//! Dynamic array of equation window objects
class EquationsArrObj :public ListObj<EquationsObj>
{
 public:
  //! Return current equations
  EquationsObj *AkEquations()
  {
    return &GetAkEint();
  }
};

// ----------------------------------------------------------------------------
// EFDaten
// ----------------------------------------------------------------------------

//! Main program data structure
class EFDaten :public EFDataIOObj
{
  enum
  {
    //! Name count
    NumGemNamen = 50
  };
  //! Names for EFDaten MCX file entries
  static const char *GemNamen[];

 public:
  //! Selection options for solver
  enum AuflSelektion
  {
    SOOberhalb = 1,
    SOUnterhalb= 2,
    AlleFkts   = 3,
    SOSelText  = 4
  };

  //! Other options for solver
  enum AuflOptionen
  {
    UAnh       = 1,
    UDebug     = 2,
    UKlKuerzen = 4,
    UComment   = 8,
    URules     = 16
  };
  
  //! Enums for open panels
  enum Panels
  {
    PanelVisible               = 1,
    InteractiveSolverPanelOpen = 2,
    FunctionsPanelOpen         = 4,
    SubmodulesPanelOpen        = 8,
    SingleValuePanelOpen       = 16
  };

  //! Diagrams
  DiagDataArrObj *DiagDataArr;
  //! Pictures
  PictureArrObj *PictureArr;
  //! Equations windows
  EquationsArrObj *EquationsArr;
  //! Functions
  FunctionArrObj *Funktion;
  //! Function editor options
  BildschOptObj BildschOpt;
  //! Function editor font
  FontType Editor;
  //! Solver templates
  AuflVorlageObj AuflVorlage;
  //! History for various combo boxes
  EFComboData EinzelwertHistory,ErgebnisHistory,
              BeibehaltenHistory,ZusammenHistory;
  //! Current function editor scroll position and line
  unsigned long ScrollPosY,ScrollPosX,ScrollLine;
  //! Current function editor selection
  unsigned int EditSelAnf,EditSelEnd;
  //! Approximation function dialog settings
  NaeherungDataType NaehData;
  //! Zoom and page setup options
  int Zoom,SeiteQuer;
  //! Used submodules list
  MGStrListObj m_submodules;
  //! Flags for open panels
  KettObj<int> m_OpenPanels;

  //! Ctor, sets number format
  EFDaten(const NumListFormat &LDV);
  //! Dtor
  virtual ~EFDaten();
  //! Load MCX file
  bool EinstDateiLaden(istream &Datei);
  //! Write editor font to XML file (for macro)
  void XmlWriteEditFont(EFXmlNode *chnode,const FontType &font);
  //! Load editor font from XML file (for macro)
  void XmlLoadEditFont(EFXmlNode *chnode,FontType &font);
  //! Save MCX file in XML format
  bool XmlDateiSpeichern(ostream &Datei,DiagDataPtArrObj &DiagDataPts);
  //! Load MCX file in XML format
  bool XmlDateiLaden(istream &Datei);
  //! Load MCX file in XML format (used for CDATA blocks)
  void ReadFromXmlDoc(EFXmlDocument &doc);
  //! Initialize document
  void Anfangswerte();
  //! Load functions from text file
  bool LoadTextDatei(const _TCHAR *TextDateiName);
  //! Write functions to text file
  bool WriteTextDatei(const _TCHAR *TextDateiName,bool Header);
  //! Return history for value calculation dialog
  inline EFComboData *GetEinzHistory()
  {
    return &EinzelwertHistory;
  }
  //! Return solver result history
  inline EFComboData *GetErgHistory()
  {
    return &ErgebnisHistory;
  }
  //! Return solver keep history
  inline EFComboData *GetBeiHistory()
  {
    return &BeibehaltenHistory;
  }
  //! Return solver keep-together history
  inline EFComboData *GetZusHistory()
  {
    return &ZusammenHistory;
  }
};

#endif // __EFMCXIO__
