// Version 6.06  29.11.2009 Markus Greither
// Copyright (c) 1991-2009 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efnmdlgs.h
    \brief Header for the non modal dialogs
*/

#ifndef __EFNMDLGS__
#define __EFNMDLGS__

#include "efdlgbas.h"

// ----------------------------------------------------------------------------
// LabelEntryDlg
// ----------------------------------------------------------------------------

//! Dialog for entering labels
class TEFLabelEntryDlg :public TEFHilfeDlg
{
  //! Data structure
  TEFView *View;

 public:
  //! Ctor
  TEFLabelEntryDlg(TEFView *V);
  //! Create dialog
  void Create(wxWindow *parent,wxWindowID Id,const wxString &title);
  //! Respond to insert button press
  void OnInsert(wxCommandEvent &event);
  //! Respond to set focus event by updating list of labels
  void OnActivate(wxActivateEvent &event);

 protected:
  DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// InteractiveSolverData
// ----------------------------------------------------------------------------

//! Data structure for interactive solving
struct InteractiveSolverData
{
  //! Operand mode (currently unused)
  enum Modes
  {
    UseOperand = 1
  };
  //! Operator, one out of +,-,/,*,=
  _TCHAR Operator;
  //! Expression to operate
  MGPStr Operand;
  //! Operating mode (unused)
  int Mode;
  //! Pointer to View
  TEFView *View;
  //! Pointer to Solver
  SolverObj *Solver;
  //! Pointer to functions
  FunctionArrObj *Fu;
  //! Pointer to parser
  ParserObj *Parser;
};

// ----------------------------------------------------------------------------
// InteractiveSolverThread
// ----------------------------------------------------------------------------

#ifdef EFMULTITHREAD
//! Thread class for interactive solver background calculation
class InteractiveSolverThread : public wxThread
{
  //! Function solver
  InteractiveSolverData &Data;

 public:
  //! Ctor
  InteractiveSolverThread(InteractiveSolverData &data)
   : wxThread(wxTHREAD_JOINABLE),Data(data) {}
  //! Do one solving step
  virtual ExitCode Entry()
  {
    return (ExitCode)Data.View->SolveEquations();
  }
};

// ----------------------------------------------------------------------------
// InteractiveSolverPanel
// ----------------------------------------------------------------------------

//! Panel for interactive function solving
class TEFInteractiveSolverPanel :public wxPanel
{
  //! Data structure
  InteractiveSolverData Data;
  //! Pointer to solver thread
  InteractiveSolverThread *threadid;
  //! Functions to be solved
  FunctionArrObj Fu;
  //! Name and term of current equation
  MGPStr Name,Term;

 public:
  //! Ctor
  TEFInteractiveSolverPanel(InteractiveSolverData &F);
  //! Respond to multiply button press
  void OnMultiply(wxCommandEvent &event);
  //! Respond to divide button press
  void OnDivide(wxCommandEvent &event);
  //! Respond to add button press
  void OnAdd(wxCommandEvent &event);
  //! Respond to subtract button press
  void OnSubtract(wxCommandEvent &event);
  //! Respond to set equation button press
  void OnSetEquation(wxCommandEvent &event);
  //! Respond to insert equation button press
  void OnInsertEquation(wxCommandEvent &event);
  //! Respond to optimize button press
  void OnOptimize(wxCommandEvent &event);
  //! Respond to solve button press
  void OnSolve(wxCommandEvent &event);
  //! Activate/deactivate child windows
  void OnIdle(wxIdleEvent &event);
  //! Delete active thread on close
  void OnClose(wxCloseEvent &event);
  //! Respond to cancel button press
  void OnCancel(wxCommandEvent &event);
  //! Show context help
  void OnHilfe(wxCommandEvent &);
  //! Get data member
  InteractiveSolverData &GetData() {return Data;}
  //! Start new thread
  void RunThread(bool getoperand = true);

 protected:
  DECLARE_EVENT_TABLE()
  DECLARE_CLASS(TEFInteractiveSolverPanel)
};
#endif

// ----------------------------------------------------------------------------
// EFSolverThread
// ----------------------------------------------------------------------------

#ifdef EFMULTITHREAD
//! Thread class for solver background calculation
class EFSolverThread : public wxThread
{
  //! Function solver
  SolverObj &Solver;

 public:
  //! Ctor
  EFSolverThread(SolverObj &slv) : wxThread(wxTHREAD_JOINABLE),Solver(slv) {}
  //! Do one solving step (eliminate one label)
  virtual ExitCode Entry()
  {
    return (ExitCode)Solver.ElimLabel();
  }
};
#else
typedef int EFSolverThread;
#endif

// ----------------------------------------------------------------------------
// TEFSolverDlg
// ----------------------------------------------------------------------------

//! Dialog for solver
/*!
    It uses EFSolverThread for actually doing solving.
*/
class TEFSolverDlg :public TEFEditDlg
{
  //! Solving mode flags
  int Mode;
  //! Pointer to solver
  SolverObj *Solver;
  //! Pointer to solver thread
  EFSolverThread *threadid;
  //! String for debug output
  MGPStr DbgOut;
  //! String for rules output
  MGPStr RulesOut;
  //! Function parser
  ParserObj &Parser;

 public:
  //! Ctor
  TEFSolverDlg(FunctionArrObj *F,const _TCHAR *SStr,const _TCHAR *EStr,
               const _TCHAR *ZStr,int M,ParserObj &psr);
  //! Dtor
  virtual ~TEFSolverDlg();
  //! Create dialog
  void Create(wxWindow *parent,wxWindowID Id,const wxString &title);
  //! Respond to OK button press
  void OnOk(wxCommandEvent &event);
  //! Respond to cancel button press
  void OnCancel(wxCommandEvent &event);
  //! Respond to close button press
  void OnClose(wxCloseEvent &event);
  //! Respond to resize
  void OnSize(wxSizeEvent &event);
  //! Get error message
  static MGPStr FehlerMeldung(const int &Fehler);

 protected:
  //! Put debug string into clipboard
  void CopyDbgOut();
  //! Invoke solver
  bool StartLabelThread();
  //! Start next elimination when finished
  void OnIdle(wxIdleEvent &event);
  DECLARE_EVENT_TABLE()
 };

// ----------------------------------------------------------------------------
// FunktionsData
// ----------------------------------------------------------------------------

//! Data structure for TEFFunktionsDlg
struct FunktionsData
{
  //! Mode flag
  unsigned long Mode;
  //! Pointer to view
  TEFView *EFView;
};

// ----------------------------------------------------------------------------
// TEFFunktionsPanel
// ----------------------------------------------------------------------------

//! Panel for function selection
class TEFFunktionsPanel :public wxPanel
{
  //! Reference to dialog data structure
  FunktionsData Data;

 public:
  //! Ctor
  TEFFunktionsPanel(FunktionsData &F) : Data(F) {}
  //! Initialize values
  void Initialize();
  //! Copies all labels in AkStr into Labels
  static void GetLabels(MGPStr &AkStr,MGPStrSet &Labels,bool withbrackets = true);
  //! New function type selected
  void OnLbnSelChanged(wxCommandEvent &);
  //! New function selected
  void OnLbnFunkChanged(wxCommandEvent &);
  //! Respond to Insert button press
  void OnYes(wxCommandEvent &event);
  //! Respond to cancel button press
  void OnCancel(wxCommandEvent &);
  //! Show context help
  void OnHilfe(wxCommandEvent &);
  //! Respond to text changes
  void OnTextChanged(wxCommandEvent &event);
  //! Setup functions list
  void SetFuStrs(long f);

 protected:
#ifdef __WXGTK__
  //! Add line breaks for GTK static text control
  void AddLineBreaks(wxStaticText *txt,wxString &str);
#endif

  DECLARE_EVENT_TABLE()
  DECLARE_CLASS(TEFFunktionsPanel)
};

// ----------------------------------------------------------------------------
// TEFLabelComboBox
// ----------------------------------------------------------------------------

//! Combo box that fills itself with the available labels
class TEFLabelComboBox :public wxComboBox
{
  //! Pointer to view
  TEFView *m_view;

 public:
  //! Ctor
  TEFLabelComboBox(wxWindow *parent,wxWindowID id,TEFView *v)
   : wxComboBox(parent,id),m_view(v) {}
  //! Fill available labels in on focus event
  void OnSetFocus(wxFocusEvent &event);

 protected:
  DECLARE_EVENT_TABLE()
};

// ----------------------------------------------------------------------------
// SubmodulesData
// ----------------------------------------------------------------------------

//! Data structure for TEFSubmodulesPanel
struct SubmodulesData
{
  //! Pointer to view
  TEFView *EFView;
  //! Functions array
  MGStrListObj m_submodules;
};

// ----------------------------------------------------------------------------
// TEFSubmodulesPanel
// ----------------------------------------------------------------------------

//! Dialog for selecting submodules.
class TEFSubmodulesPanel :public wxPanel
{
  //! Copy of passed in data, because non modal dialog
  SubmodulesData Data;

 public:
  //! Ctor
  TEFSubmodulesPanel(SubmodulesData &F);
  //! Initialize panel
  void Initialize();
  //! Respond to listbox selection event
  void OnSelect(wxCommandEvent &event);
  //! Respond to add button press
  void OnAdd(wxCommandEvent &event);
  //! Responds to delete button press
  void OnDelete(wxCommandEvent &event);
  //! Responds to open button press
  void OnOpen(wxCommandEvent &event);
  //! Responds to double click
  void OnDblClk(wxCommandEvent &event);
  //! Respond to cancel button press
  void OnCancel(wxCommandEvent &/*event*/)
  {
    Destroy();
  }
  //! Show context help
  void OnHilfe(wxCommandEvent &);

 protected:

  DECLARE_EVENT_TABLE()
  DECLARE_CLASS(TEFSubmodulesPanel)
};

// ----------------------------------------------------------------------------
// EinzelwertData
// ----------------------------------------------------------------------------

//! Data structure for TEFEinzelwertPanel
struct EinzelwertData
{
  //! Functions array
  FunctionArrObj Fu;
  //! Pointer to function editor
  TEFEdit *Editor;
  //! History data for input line
  EFComboData *CD;
  //! Pointer to view
  TEFView *EFView;
  //! Number format
  NumListFormat LDV;
};

// ----------------------------------------------------------------------------
// TEFEinzelwertDlg
// ----------------------------------------------------------------------------

//! Dialog for calculation of single function values.
/*!
    This dialog also has pocked calculator functionality.
*/
class TEFEinzelwertPanel :public wxPanel
{
  //! Copy of passed in data, because non modal dialog
  EinzelwertData Data;
  //! Pointer to input line
  TEFComboBox *TE;
  //! Current result
  MGPStr AkErg;
  //! Function parser
  ParserObj *Parser;
  //! Curent insertion point in TE
  long InsertionPoint;

 public:
  //! Ctor
  TEFEinzelwertPanel(EinzelwertData &F);
  //! Dtor
  virtual ~TEFEinzelwertPanel();
  //! Initialize panel
  void Initialize();
  //! Calculate function
  void Berechnen(bool forceerroroutput);
  //! Insert zeile at current position, select Anf to End
  void InsertString(const _TCHAR *zeile,int Anf,int End);
  //! Activate input line if text entered
  void EnSetFocus(wxCommandEvent &event);
  //! Respond to calculation command from editor
  void OnEditorBerechnen(wxCommandEvent &event);
  //! Responds to text button press (pocket calculator)
  void OnTextButton(wxCommandEvent &event);
  //! Responds to calculate button press
  void OnBerechnen(wxCommandEvent &event);
  //! Responds to insert button press
  void OnNo(wxCommandEvent &event);
  //! Respond to cancel button press
  void OnCancel(wxCommandEvent &/*event*/) 
  {
    Destroy();
  }
  //! Show context help
  void OnHilfe(wxCommandEvent &);
  //! Read current insertion pos in idle time
  void OnIdle(wxIdleEvent &event);
  //! Respond to show all event
  void OnAlleBerechnen(wxCommandEvent &/*event*/);
  //! Set new editor window
  void SetEditor(TEFEdit *edit) {Data.Editor = edit;}
  //! Format the result
  void FormatResult(Cpx y, double x, bool StandardFormat, 
                    const MGPStr &Namen,const MGPStr &Term,
                    MGPStr &RealStr, MGPStr &ImagStr, MGPStr &AkErg,
                    int Conversion);
  //! Convert the calculated result per Conversion
  Cpx ConvertResult(Cpx y,int Conversion);
  //! Returns selection status of local function text field
  bool GetLocalEditorActive();

 protected:

  DECLARE_EVENT_TABLE()
  DECLARE_CLASS(TEFEinzelwertPanel)
};

#endif
