// Version 5.31  08.07.2006 Markus Greither
// Copyright (c) 1991-2006 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efsolver.h
    \brief Header for the symbolic function solver
*/

#ifndef __EFSOLVER__
#define __EFSOLVER__

// ----------------------------------------------------------------------------
// MGStrObj
// ----------------------------------------------------------------------------

//! Dynamic array of MGPStr objects
class MGStrObj :public KettObj<MGPStr>
{
 public:
  //! Returns current entry
  inline MGPStr &GetAkStr()
  {
    return GetAkEint();
  }
  //! Set new value for current entry
  inline void SetAkStr(const _TCHAR *Zeile)
  {
    GetAkEint().SetStr(Zeile);
  }
};

// ----------------------------------------------------------------------------
// SolverTools
// ----------------------------------------------------------------------------

//! Tool class for elementary solver functions
class SolverTools : protected MarkerObj
{
 protected:
  //! optional Debug string
  MGPStr *DbgOut;
  //! optional Rules string
  MGPStr *RulesOut;
  //! Start time for  Debug output
  time_t firsttime;
  //! Error flag
  int Fehler;
  //! Field with all expressions
  MGStrObj AbstrFeld;
  //! Function parser
  ParserObj *Parser;
  //! Multiplies +/-, \0 is treated as +
  inline _TCHAR MultVorZei(_TCHAR a,_TCHAR b)
  {
    return ((a == '-') && (b != '-') ||
            (a != '-') && (b == '-') ? '-' : '+');
  }
  //! Change row of factors so that no * or / is upfront
  void FaksToSummant(MGPStr &AkStr);
  //! Check whether AStr is in AbstrFeld
  bool InAbstrFeld(const _TCHAR *AStr);
  //! Insert rules information in RulesOut
  virtual void RulesString(const _TCHAR *);
  //! Insert debug information in DbgOut
  virtual void DebugString(const _TCHAR *);
 public:
  //! Error definitions for solver
  enum SolverErrors
  {
    OutOfMemory = 200,
    Syntax,
    Parenthesis,
    LineToLong,
    ConstOutOfRange,
    QuadIntr,
    NotSolved,
    DivBZero,
    InvalidSigns,
    NoEquationEdit
  };
  //! Ctor
  SolverTools(ParserObj *Psr)
  {
    Parser = Psr;
    DbgOut = 0;
    Fehler = 0;
  }
  //! Ctor with debug
  SolverTools(ParserObj *Psr,MGPStr *Dbg,MGPStr *rl = 0)
  {
    Parser = Psr;
    DbgOut = Dbg;
    RulesOut = rl;
    Fehler = 0;
  }
  //! Dtor
  virtual ~SolverTools()
  {
  }
  //! Read next factor from AkStr at position p
  void NextFakt(MGPStr &AkStr,MGPStr &FakStr,unsigned int &p); // debug NEF
  //! Read summant from AkStr at p, determine sign, return true if brackets
  bool Summant(const _TCHAR *AkStr,MGPStr &SumStr,unsigned int &p,
               _TCHAR &RechZei); // debug SUM
  //! Add AkStr to AbstrFeld, f returns index
  bool AddAbstrString(const _TCHAR *AkStr,int &f); // debug ADD
  //! Formats FakStr for output (FakStr with multiple Summants), put 1 for /
  static void FormatFakStr(MGPStr &AStr,const _TCHAR *FakStr);
};

// ----------------------------------------------------------------------------
// SolverObj
// ----------------------------------------------------------------------------

//! The main function solver class
class SolverObj : public SolverTools
{
  //! Search factor (Result), factors to be kept strings
  MGPStr SuchFaktor,ErhFaktoren,ZusFaktoren;
  //! Labels that cannot be eliminated
  MGPStrSet GefLabels;
  //! Labels that cannot be eliminated and pulled out of brackets
  MGPStrSet ZusLabels;
  //! Counter for function number for constant extraction
  int AkTeil;
  //! Functions array
  FunctionArrObj *Funktion;
  //! Current function for ElimLabel()
  int AkF;
  //! Flags
  bool Teilen,Debug,Aufgeloest,NoMoreLabels,ElimLabelStatus,
       KlKrz,ErgebnisOpt;
  //! Result and value range restriction strings
  MGPStr ErgString,EinschString;
public:
  //! Solver flags
  enum SolverModes
  {
    SolverTeilen =    1,
    SolverDebug =     2,
    SolverKlKuerzen = 4,
    SolverRegelnOut = 8
  };
  //! Ctor
  SolverObj(FunctionArrObj *Fu,
            const _TCHAR *SStr,
            const _TCHAR *EStr,
            const _TCHAR *ZStr,
            int Mode,
            ParserObj *P,
            MGPStr *DO = 0,
            MGPStr *RL = 0);
  //! Dtor
  virtual ~SolverObj();
  //! Eliminate another label, return true if successful
  bool ElimLabel();
  //! Initialize
  void StartWerte();
  //! Return last elimination status
  inline bool GetElimLabelStatus()
  {
    return ElimLabelStatus;
  }
  //! Return result string
  inline const _TCHAR *GetErgString()
  {
    return ErgString.c_str();
  }
  //! Return value range restriction string
  inline const _TCHAR *GetEinschString()
  {
    return EinschString.c_str();
  }
  //! Return error information
  inline int &GetFehler()
  {
    return Fehler;
  }
  //! Return finish state
  inline bool GetAufgeloest()
  {
    return Aufgeloest;
  }
  //! Return false if no more labels for elimination available
  inline bool GetNoMoreLabels()
  {
    return NoMoreLabels;
  }
  //! Extend current function (index 1) with FakString and operator Operator (*,/,+,-)
  bool InteractiveSolve(const _TCHAR *FakString,const _TCHAR Operator);


protected:
  //! Convert function into abstract solver format
  void Abstrahieren(MGPStr &AkStr,bool NeuFeld = false); // debug ABS
  //! Convert function into abstract solver format
  void Abstrahieren(MGPStr &LAbsStr, MGPStr &RAbsStr,bool NeuFeld);
  //! Convert function into readable format
  const _TCHAR *ReAbstrahieren(MGPStr &AkStr); // debug RAB
  //! Helper function for bracket elimination
  void Faktor1(MGPStr &SumStr,MGPStr &FStr,MGPStr &KlStr,unsigned int &p1);
  //! Helper function for bracket elimination
  void Faktor2(MGPStr &SumStr,MGPStr &FStr,unsigned int &p1);
  //! Eliminate brackets
  void AusMult(MGPStr &LAbsStr,MGPStr &RAbsStr); // debug AML
  //! Optimize subexpressions with optimized bracketing
  void OptimizeSubExp(MGPStr &AkStr,bool einklammern = true);
  //! Add brackets around expression if necessary
  void Einklammern(MGPStr &AkStr);
  //! Eliminate DivStr from both strings by multiplication
  bool EntfDiv(MGPStr &LAbsStr,MGPStr &RAbsStr,const _TCHAR *DivStr); // debug DIV
  //! Eliminate labels, multiply/divide numeric constants
  void Vereinfachen(MGPStr &AkStr); // debug EIN
  //! Sum up numeric constants
  void Aufadd(MGPStr &AkStr); // debug AUF
  //! Eliminate sum expressions with different signs
  void SumKuerzen(MGPStr &AkStr); // debug SKZ
  //! Get factors of each summant
  void SumNFaktor(MGPStr &AkStr,MGStrObj *FakFeld,
                  bool addone = false);// debug SUF
  //! Eliminate factors that appear in each summant
  void Kuerzen(MGPStr &LAbsStr,MGPStr &RAbsStr); // debug KZN
  //! Symbolic addition of AkStr AddZei AddStr
  void Add(MGPStr &AkStr,_TCHAR AddZei,const _TCHAR *AddStr); // debug ADD
  //! Symbolic multiplication AkStr MultZei MStr
  void Mult(MGPStr &AkStr,_TCHAR MultZei,const _TCHAR *MStr); // debug MUL
  //! Determine if AkStr has only 1 summant
  bool IstEinSummant(const _TCHAR *AkStr,_TCHAR &RechZei);
  //! Search factor that appears in each summant in both strings
  MGStrObj *SucheFak(MGPStr &LAbsStr,MGPStr &RAbsStr); // debug SUF
  //! Eliminate SuchStr from AkStr
  bool EntfSuchStr(MGPStr &AkStr,MGPStr &SuchStr);
  //! Bracketing
  void Ausklammern(MGPStr &AkStr); // debug AUS
  //! Convert summants expression into factor by bracketing if necessary
  void MakeFaktor(MGPStr &AkStr);
  //! Delete duplicate lines in AStr
  void DelDuplicateLines(MGPStr &AStr);
  //! Sort functions (shortest first)
  void SortFunktionen(); // debug SFK
  //! Eliminate factor groups available both in LAbsStr and RAbsStr
  void FakGruppenKuerzen(MGPStr &LAbsStr,MGPStr &RAbsStr); // debug FGK
  //! Simplify brueche (Signs, bracketing ...)
  void BruchVereinfachen(MGPStr &AkLStr,MGPStr &AkRStr,bool Teilen);
  //! Optimize end result
  void ErgebnisUmformen(); // debug EU
  //! Check whether AStr is in ZusLabels
  bool InZusLabels(const MGPStr &AStr)
  {
    return ZusLabels.find(AStr) != ZusLabels.end();
  }
  //! Check whether AStr is in either ZusLabels or GefLabels
  bool InGefLabels(const MGPStr &AStr)
  {
    return GefLabels.find(AStr) != GefLabels.end();
  }
  //! Get factors in AStr and make new entry in Faktoren for each
  // erzeugt fuer jeden einen Eintrag in Faktoren
  void GetFaktoren(MGPStrSet &Faktoren,const _TCHAR *AStr);
  //! Check if AkStr has only labels that cannot be eliminated
  bool Konstanten(MGPStr &AkStr); // debug KON
  //! Remove (multiply) signs before factors
  void EntfVorzeichen(MGPStr &AkStr); // debug ENV
  //! Find divisor
  void SucheDiv(MGPStr &AkStr,MGPStr &DivStr); // debug SDV
  //! Eliminate bracket expressions
  void KlKuerzen(MGPStr &RAbsStr,MGPStr &LAbsStr); // debug KLK
  //! Test for same Factors and make exponential expression
  void MakePotenzen(MGPStr &AkStr);
  //! Insert solution for exponential function
  bool PotErg(MGPStr &ZaehlStr,MGPStr &ErgLStr,
              MGPStr &PotStr,int PotAnz);
  //! Insert solution for square function
  void QuadErg(MGPStr &BStr,MGPStr &ZaehlStr,MGPStr &ErgLStr,MGPStr &QStr);
  //! Isolate searched label on left side
  void Isolieren(MGPStr &LAbsStr,MGPStr &RAbsStr,MGPStr &SuchStr,
                 bool Teilen); // debug ISO
  //! Solve function for label SuchStr
  void FunktionAufloesen(MGPStr &LAbsStr,MGPStr &RAbsStr,MGPStr &SuchStr,
                         bool AndereLabels,bool Teilen,
                         bool NoExp = false); // debug FAL
  //! Merge functions that are solved for the same label
  bool FunktionGleichSetzen(MGPStr &LAbsStr,MGPStr &RAbsStr,MGPStr &LZielStr,
                            MGPStr &RZielStr,bool &Teilen); // debug FGS
  //! Divide ZaehlSums by NennStr and determine modulo
  int TermDivision(MGPStr &NennStr,MGStrObj &ZaehlSums,bool norest = false);
  //! Break NennStr in factors and calls TermDivision for each bracketed term
  int TeilNennStrDivision(MGPStr &NennStr,MGStrObj &ZaehlSums,
                          MGPStr &RestStr,MGPStr &ZaehlRestStr);
  //! Break ZaehlSums in factors and calls TermDivision for each bracketed term
  int TeilZaehlSumsDivision(MGPStr &NennStr,MGStrObj &ZaehlSums,
                           MGPStr &RestStr);
  //! Remove [] around factors
  void LoeschKlammern();
  //! Remove the berechnet flag from all functions
  void NichtBerechnet();
  //! Optimize Exponent and Base of exponential expression
  bool OptimizePotenz(MGPStr &AkStr,unsigned int &p,
                      unsigned int &AnfPos,int &KlZaehler);
  //! Reads fraction number from string
  bool GetBruchWert(MGPStr &AkStr,BruchZahl &NeuWert);
  //! Writes fraction number into string
  void BruchOut(BruchZahl Wert,MGPStr &AkStr);
};

#endif

