// Version 5.10  25.03.02 Markus Greither
// Copyright (c) 1991-2003 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efundo.h
    \brief Header for the undo/redo class used in the function editor
*/

#ifndef __EFUNDO__
#define __EFUNDO__

// ----------------------------------------------------------------------------
// TextPos
// ----------------------------------------------------------------------------

//! Text position for TEFEdit
struct TextPos
{
  //! Character position
  unsigned int Pos;
  //! Line position
  int Zeile;
  //! Ctor
  TextPos()
  {
    Pos = 0;
    Zeile = 0;
  }
  //! Ctor
  TextPos(unsigned int P,int PZeile)
  {
    SetPos(P,PZeile);
  }
  //! Set new position
  void SetPos(unsigned int P,int PZeile)
  {
    Pos = P;
    Zeile = PZeile;
  }
  //! Compare operator
  bool operator == (const TextPos &pos) const
  {
    return ((pos.Pos == Pos) && (pos.Zeile == Zeile));
  }
  //! Compare operator
  bool operator != (const TextPos &pos) const
  {
    return ((pos.Pos != Pos) || (pos.Zeile != Zeile));
  }
};

//! TEFEdit undo data item
struct UndoItem
{
  //! Set undo information
  void Set(TextPos A,TextPos E,TextPos I,const MGPStr &str)
  {
    Anf = A;
    End = E;
    Ins = I;
    Str = str;
  }
  //! Undo string
  MGPStr Str;
  //! Undo positions
  TextPos Anf,End,Ins;
};

// ----------------------------------------------------------------------------
// UndoArr
// ----------------------------------------------------------------------------

//! Dynamic array of UndoItem objects
class UndoArr : public ListObj<UndoItem> {};

// ----------------------------------------------------------------------------
// UndoData
// ----------------------------------------------------------------------------

//! Undo data.
/*
    Undo data holds the undo string, and the undo position.
    There are three modes:
    UndoInsert undos last text input
    UndoDelete undos last deletion
    UndoDnD undos last insert and last deletion
*/
class UndoData
{
  //! Undo/Redo information
  UndoArr UndoInfo;
 public:
  //! Clear Undo/Redo information
  void Clear()
  {
    UndoInfo.ResetKette();
    UndoInfo.NeuAkEint();
  }
  //! Ctor
  UndoData()
  {
    Clear();
  }
  //! Converts akitem into invers operation
  void InvertAkItem(TextPos E,const MGPStr &str)
  {
    UndoInfo.GetAkEint().End = E; // new end pos for delete
    TextPos tmp = UndoInfo.GetAkEint().Anf; // swop Anf and Ins
    UndoInfo.GetAkEint().Anf = UndoInfo.GetAkEint().Ins;
    UndoInfo.GetAkEint().Ins = tmp;
    UndoInfo.GetAkEint().Str = str; // new undo string for insert
  }
  //! Add new undo information
  void NewUndo(TextPos A,TextPos E,TextPos I,const MGPStr &Str);
  //! Update current insert undo
  void UpdateInsert(TextPos A,TextPos E,const _TCHAR C);
  //! Update current delete undo
  void UpdateDelete(TextPos A,TextPos E,const _TCHAR C);
  //! Update current delete undo
  void UpdateDelete(TextPos A,TextPos E,const MGPStr &str);
  //! Return whether any undo is available
  bool UndoAvailable()
  {
    return (UndoInfo.AkEintNummer() > 0);
  }
  //! Return whether any redo is available
  bool RedoAvailable()
  {
    return (UndoInfo.AkEintNummer() < UndoInfo.GetMaxEintrag());
  }
  //! Select previous item after undo
  void operator --(int)
  {
    if (UndoInfo.AkEintNummer() > 0)
      UndoInfo.SetAkEint(UndoInfo.AkEintNummer()-1);
  }
  //! Select next item after redo
  void operator ++(int)
  {
    if (UndoInfo.AkEintNummer() < UndoInfo.GetMaxEintrag())
      UndoInfo.SetAkEint(UndoInfo.AkEintNummer()+1);
  }
  //! Returns current Undo / Redo item
  UndoItem &GetAkItem()
  {
    return UndoInfo.GetAkEint();
  }
};

#endif

