// Version 5.32  27.12.2006 Markus Greither
// Copyright (c) 1991-2006 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efxmlio.h
    \brief Header for the xml data storage classes
*/

#ifndef __EFXMLIO__
#define __EFXMLIO__

//! Array of Strings
typedef ListObj<string> StringArr;

//! Represents node property(ies).
class EFXmlProperty
{
public:
    //! Ctor
    EFXmlProperty() : m_next(NULL) {}
    //! Ctor
    EFXmlProperty(const _TCHAR *name, const _TCHAR *value,
                  EFXmlProperty *next)
            : m_name(name), m_value(value), m_next(next) {}

    //! Accessor for name
    const MGPStr &GetName() const { return m_name; }
    //! Accessor for value
    const MGPStr &GetValue() const { return m_value; }
    //! Accessor for next property item
    EFXmlProperty *GetNext() const { return m_next; }

    //! Sets name
    void SetName(const MGPStr& name) { m_name = name; }
    //! Sets value
    void SetValue(const MGPStr& value) { m_value = value; }
    //! Sets next item
    void SetNext(EFXmlProperty *next) { m_next = next; }

private:
    //! Property name
    MGPStr m_name;
    //! Property value
    MGPStr m_value;
    //! Next property in chain
    EFXmlProperty *m_next;
};

//! Represents node in XML document. Node has name and may have content
class EFXmlNode
{
public:
    //! Supported XML node types
    enum EFXmlNodeType
    {
        ELEMENT_NODE       =  1,
        ATTRIBUTE_NODE     =  2,
        TEXT_NODE          =  3,
        CDATA_SECTION_NODE =  4,
        COMMENT_NODE       =  5
    };
    //! Ctor
    EFXmlNode(EFXmlNode *parent,EFXmlNodeType type,
              const _TCHAR *name, const _TCHAR *content,
              EFXmlProperty *props, EFXmlNode *next);
    //! Dtor          
    ~EFXmlNode();

    //! user-friendly creation
    EFXmlNode(EFXmlNodeType type, const _TCHAR *name,
              const _TCHAR *content = _TEXT(""));
    //! Add child node  
    void AddChild(EFXmlNode *child);
    //! Add property
    void AddProperty(const _TCHAR *name, const _TCHAR *value);
    //! Delete property
    bool DeleteProperty(const _TCHAR *name);

    //! Accessor to type
    EFXmlNodeType GetType() const { return m_type; }
    //! Accessor to name
    const MGPStr &GetName() const { return m_name; }
    //! Accessor to content
    const MGPStr &GetContent() const { return m_content; }
    //! Return CDATA content, remove reference (string needs to be deleted by caller)
    StringArr *GetCDATA()
    {
      StringArr *tmp = m_cdata;
      m_cdata = NULL;
      return tmp;
    }

    //! Accessor to parent
    EFXmlNode *GetParent() const { return m_parent; }
    //! Accessor to next node
    EFXmlNode *GetNext() const { return m_next; }
    //! Accessor to child node list
    EFXmlNode *GetChildren() const { return m_children; }

    //! Accessor to properties list
    EFXmlProperty *GetProperties() const { return m_properties; }
    //! Get property value
    bool GetPropVal(const _TCHAR *propName, MGPStr &value) const;
    //! Get property value
    MGPStr GetPropVal(const _TCHAR *propName,
                      const _TCHAR *defaultVal) const;
    //! Find property
    bool HasProp(const _TCHAR *propName) const;

    //! Set node name
    void SetName(const MGPStr& name) { m_name = name; }
    //! Set node content
    void SetContent(const MGPStr& con) { m_content = con; }
    //! Set node CDATA content
    void SetCDATA(StringArr *con) { m_cdata = con; }
    //! Set node parent
    void SetParent(EFXmlNode *parent) { m_parent = parent; }
    //! Set next node in list
    void SetNext(EFXmlNode *next) { m_next = next; }
    //! Set children list
    void SetChildren(EFXmlNode *child) { m_children = child; }

    //! Set property list
    void SetProperties(EFXmlProperty *prop) { m_properties = prop; }
    //! Add single property
    void AddProperty(EFXmlProperty *prop);

private:
    //! Node type
    EFXmlNodeType m_type;
    //! Node name
    MGPStr m_name;
    //! Node content
    MGPStr m_content;
    //! Node CDATA content
    StringArr *m_cdata;
    //! Node properties list
    EFXmlProperty *m_properties;
    //! Node parent, children list, next node
    EFXmlNode *m_parent, *m_children, *m_next, *m_lastchild;
};

//! Forward declaration
class EFDaten;

//! This class holds XML data/document as parsed by XML parser.
class EFXmlDocument
{
public:
    //! Ctor
    EFXmlDocument(EFDaten *data = 0);
    //! Dtor
    ~EFXmlDocument() { delete m_root; }

    //! Parses .xml file and loads data.
    bool Load(istream& stream,
              const MGPStr& encoding = _TEXT("UTF-8"));
    //! Saves document as .xml file.
    bool Save(ostream& stream) const;

    //! Check for valid document
    bool IsOk() const { return m_root != NULL; }

    //! Returns root node of the document.
    EFXmlNode *GetRoot() const { return m_root; }

    //! Returns version of document (may be empty).
    const MGPStr &GetVersion() const { return m_version; }
    //! Returns encoding of document (may be empty).
    const MGPStr &GetFileEncoding() const { return m_fileEncoding; }

    //! Set root node
    void SetRoot(EFXmlNode *node) { delete m_root ; m_root = node; }
    //! Set version string
    void SetVersion(const MGPStr& version) { m_version = version; }
    //! Set file encoding
    void SetFileEncoding(const MGPStr& encoding) { m_fileEncoding = encoding; }

#ifndef UNICODE
    //! Returns encoding of in-memory representation of the document
    const MGPStr &GetEncoding() const { return m_encoding; }
    //! Set encoding
    void SetEncoding(const MGPStr& enc) { m_encoding = enc; }
#endif
    //! converts Expat-produced string in UTF-8 into wxString
    static wxString CharToString(wxMBConv *conv,
                                 const char *s, size_t len = wxString::npos);
    //! Write string to output
    static void OutputString(ostream& stream, const wxString& str,
                             wxMBConv *convMem, wxMBConv *convFile);
    //! Same as above, but create entities first.
    /*! Translates '<' to "&lt;", '>' to "&gt;" and '&' to "&amp;"
    */
    static void OutputStringEnt(ostream& stream, const wxString& str,
                                wxMBConv *convMem, wxMBConv *convFile);
    //! Indents the output file
    static void OutputIndentation(ostream& stream, int indent);
    //! Write node to file
    static void OutputNode(ostream& stream, EFXmlNode *node, int indent,
                           wxMBConv *convMem, wxMBConv *convFile);

private:
    //! Version string
    MGPStr   m_version;
    //! File enconding
    MGPStr   m_fileEncoding;
#ifndef UNICODE
    //! Memory representation encoding
    MGPStr   m_encoding;
#endif
    //! Root node of document
    EFXmlNode *m_root;
    //! Reference to document
    EFDaten *m_efdaten;
};

#endif
