// Version 5.33  30.03.2007 Markus Greither
// Copyright (c) 1991-2007 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file faststr.h
    \brief Header for self defined string class
*/

// ----------------------------------------------------------------------------
// FastString
// ----------------------------------------------------------------------------

//! Efficient string class.
/*
    This class is optimized for fast performance when used for
    high volume string operations. Also deals efficiently with long strings.
*/
template <class T> class FastString
{
  //! Minimal buffer size
  enum
  {
    StringBuff = 100
  };
 private:
  //! Cached string length
  unsigned int Len;
  //! Current buffer size
  unsigned int currBuff;

 protected:
  //! Pointer to data
  T *AkStr;
  //! Initialize members
  void Init();
  //! Delete embedded string
  bool LoeschStr();

 public:
  //! Assignment Ctor
  FastString(const T *AStr)
  {
//    StringBuff = 100;
    Init();
    SetStr(AStr);
  }
  //! Assignment Ctor
  FastString(const FastString<T> &AStr)
  {
//    StringBuff = 100;
    Init();
    SetStr(AStr);
  }
  //! Ctor, creates ""
  FastString()
  {
//    StringBuff = Buff;
    Init();
  }
  //! Dtor
  virtual ~FastString()
  {
    LoeschStr();
  }
  //! Streaming operator
  FastString &operator <<(const T *AStr)
  {
    AddStr(AStr);
    return *this;
  }
  //! Streaming operator
  FastString &operator <<(const int num)
  {
    AddInt(num);
    return *this;
  }
  //! Streaming operator
  FastString &operator <<(const unsigned int num)
  {
    AddULong(num);
    return *this;
  }
  //! Streaming operator
  FastString &operator <<(const unsigned long num)
  {
    AddULong(num);
    return *this;
  }
  //! Assignment operator
  void operator =(const FastString<T> &AStr)
  {
    SetStr(AStr);
  }
  //! Assignment operator
  void operator =(const T *AStr)
  {
    SetStr(AStr);
  }
  //! Assignment operator
  void operator +=(const FastString<T> &AStr)
  {
    AddStr(AStr);
  }
  //! Assignment operator
  void operator +=(const T *AStr)
  {
    AddStr(AStr);
  }
  //! Bracket operator
  const T &operator [](size_t pos) const
  {
#ifdef __MGDEBUG__
    if (pos > Len)
      throw FastString<_TCHAR>(_TEXT("FastString: Operator [] overflow"));
#endif
    return AkStr[pos < Len ? pos : Len];
  }
  //! Append operator
  FastString &operator +(const T *AStr)
  {
    AddStr(AStr);
    return *this;
  }
  //! Compatibility with std::string, return C-String
  const T *c_str() const
  {
    return AkStr;
  }
  //! Determine line, like istream::GetLine()
  bool GetLine(FastString<T> &Zeile,int znr);
  //! Get line count
  int GetNumLines();
  //! Truncate string to length len
  bool ShortStr(unsigned int len);
  //! Insert AStr at pos
  bool InsStr(unsigned int pos,const FastString<T> &AStr)
  {
    return InsStr(pos,AStr.c_str());
  }
  //! Insert AStr at pos
  bool InsStr(unsigned int pos,const T *AStr);
  //! Insert AChar at pos
  bool InsChar(unsigned int pos,T AChar);
  //! Insert Anz characters from AStr at pos
  bool InsNChar(unsigned int pos,const T *AStr,unsigned int Anz);
  //! Appends AStr at pos (overwrite original)
  bool AddAtStr(unsigned int pos,const T *AStr);
  //! Assignment
  inline bool SetStr(const T *AStr)
  {
    if (AkStr != NULL)
      AkStr[0] = 0;
    Len = 0;
    return AddStr(AStr);
  }
  //! Assignment
  inline bool SetStr(const FastString<T> &AStr)
  {
    if (AkStr != NULL)
      AkStr[0] = 0;
    Len = 0;
    return AddStr(AStr);
  }
  //! Replace character at pos with AChar
  bool ReplChar(unsigned int pos,const T AChar);
  //! Replace all occurences of AltC by NeuC
  void ReplAllChars(const T AltC,const T NeuC);
  //! Dito, but Neu item is string
  void ReplAllChars(const T AltC,const T *Neu);
  //! Dito, but Ald item is string
  void ReplAllChars(const T *Alt,const T NeuC);
  //! Append AChar
  bool AddChar(const T AChar);
  //! Append AStr
  bool AddStr(const FastString<T> &AStr)
  {
    return AddNChar(AStr.c_str(),AStr.Length());
  }
  //! Append AStr
  bool AddStr(const T *AStr)
  {
    return AddNChar(AStr,GetLen(AStr));
  }
  //! Append integer tmp
  bool AddInt(int tmp);
  //! Append unsigned long tmp
  bool AddULong(unsigned long tmp);
  //! Append AStr
  bool AddNChar(const FastString<T> &AStr,unsigned int Anz)
  {
    return AddNChar(AStr.c_str(),Anz);
  }
  //! Append Anz characters from AStr
  bool AddNChar(const T *AStr,unsigned int Anz);
  //! Delete Anz characters starting at pos
  bool DelNChar(unsigned int pos,unsigned int Anz);
  //! Return string length
  virtual inline unsigned int Length() const
  {
    return Len;
  }
  //! Get len of character string
  size_t GetLen(const T *str);
  //! Compare character strings
  int Compare(const T *str,const T *oth,size_t len);
};

template <class T> bool FastString<T>::LoeschStr()
{
  if (AkStr != NULL)
  {
    free(AkStr);
    AkStr = NULL;
    currBuff = 0;
    Len = 0;
    return true;
  }
  return false;
}

template <class T> bool FastString<T>::AddNChar(const T *AStr,unsigned int Anz)
{
  if (AStr == NULL)
    return false;
  if (!Len && AkStr && AkStr[0])
    Len = GetLen(AkStr);
  unsigned int newSize = Anz+Len;
  if ((AkStr == NULL) || (newSize > currBuff) ||
      (newSize+2*StringBuff < currBuff))
  {
    currBuff = newSize+StringBuff;
    T *P = (T *)realloc(AkStr,(currBuff+1)*sizeof(T));
    if (P == NULL)
      return false;
    if (AkStr == NULL)
      P[0] = 0;
    AkStr = P;
  }
  unsigned long i = 0;
  while (AStr[i] && (i < Anz))
  { // maximal Anz zeichen aus AStr hinzufuegen
    AkStr[Len+i] = AStr[i];
    i++;
  }
  AkStr[Len+i] = 0;
  Len += i;
  return true;
}

template <class T> void FastString<T>::Init()
{
  AkStr = (T *)malloc(sizeof(T));
  AkStr[0] = 0;
  currBuff = 0;
  Len = 0;
}

template <class T> int FastString<T>::GetNumLines()
{
  return ::GetNumLines(AkStr);
}

template <class T> bool FastString<T>::GetLine(FastString<T> &Zeile,int znr)
{
  return ::GetLine(AkStr,Zeile,znr);
}

// loescht Anz Zeichen an Stelle pos
template <class T> bool FastString<T>::DelNChar(unsigned int Pos,
                                                unsigned int Anz)
{
  int i,f;

  if (Pos < Len)
  {
    if (Pos+Anz >= Len) // alle Zeichen hinter Pos werden geloescht
      ShortStr(Pos);
    else
    {
      f = Len-Anz;
      for (i = Pos;i <= f;i++)
        AkStr[i] = AkStr[i+Anz];
      Len -= Anz;
    }
  }
  return (AkStr != NULL);
}

template <class T> bool FastString<T>::ShortStr(unsigned int len)
{ // verkuerzt auf Laenge len
  if (len <= Len)
  {
    AkStr[len] = 0;
    Len = len;
  }
  return (AkStr != NULL);
}

template <class T> bool FastString<T>::InsStr(unsigned int pos,const T *AStr)
{ // fuegt AStr an pos ein
  FastString HiStr;

  HiStr.SetStr(AkStr);
  if (pos <= Len)
  {
    HiStr.ShortStr(pos);
    HiStr.AddStr(AStr);
    HiStr.AddStr(&AkStr[pos]);
  }
  return SetStr(HiStr);
}

template <class T> bool FastString<T>::InsChar(unsigned int pos,T AChar)
{ // fuegt AChar an pos ein
  T a[2];

  a[1] = 0;
  a[0] = AChar;
  return FastString<T>::InsStr(pos,a);
}

template <class T> bool FastString<T>::InsNChar(unsigned int pos,const T *AStr,
                                         unsigned int Anz)
{ // fuegt Anz Zeichen aus AStr an pos ein
  FastString HiStr;

  HiStr.SetStr(AStr);
  HiStr.ShortStr(Anz);
  return FastString<T>::InsStr(pos,HiStr.c_str());
}

template <class T> bool FastString<T>::AddAtStr(unsigned int pos,const T *AStr)
{ // Haengt AStr an pos an (Orginalstring hinter pos wird ueberschrieben)
  if (pos <= Len)
  {
    ShortStr(pos);
    AddStr(AStr);
  }
  return (AkStr != 0);
}

template <class T> bool FastString<T>::ReplChar(unsigned int pos,const T AChar)
{ // ersetzt das Zeichen an pos durch AChar
  if (!AChar)
    return false;
  if (pos < Len)
    AkStr[pos] = AChar;
  return (AkStr != NULL);
}

template <class T> bool FastString<T>::AddChar(const T AChar)
{ // haengt das Zeichen AChar an
  T A[2];

  A[1] = 0;
  A[0] = AChar;
  return AddStr(A);
}

template <class T> void FastString<T>::ReplAllChars(const T AltC,const T NeuC)
{
  unsigned int AkPos;

  AkPos = 0;
  while (AkPos < Len)
  {
    if (AltC == AkStr[AkPos])
      ReplChar(AkPos,NeuC);
    AkPos++;
  }
}

template <class T> void FastString<T>::ReplAllChars(const T AltC,const T *Neu)
{
  unsigned int AkPos;

  AkPos = 0;
  unsigned int len = GetLen(Neu);
  while (AkPos < Len)
  {
    if (AltC == AkStr[AkPos])
    {
      DelNChar(AkPos,1);
      InsStr(AkPos,Neu);
      AkPos += len;
    }
    else
      AkPos++;
  }
}

template <class T> void FastString<T>::ReplAllChars(const T *Alt,const T NeuC)
{
  unsigned int AkPos;

  AkPos = 0;
  unsigned int len = GetLen(Alt);
  while (AkPos < Len)
  {
    if (!_tcsncmp(Alt,&AkStr[AkPos],len))
    {
      DelNChar(AkPos,len);
      InsChar(AkPos,NeuC);
    }
    AkPos++;
  }
}

template<> size_t FastString<char>::GetLen(const char *str);

template<> int FastString<char>::Compare(const char *str,
                                          const char *oth,
                                          size_t len);

template<> bool FastString<char>::AddULong(unsigned long tmp);

template<> bool FastString<char>::AddInt(int tmp);

#ifdef UNICODE
template<> size_t FastString<wchar_t>::GetLen(const wchar_t *str);

template<> int FastString<wchar_t>::Compare(const wchar_t *str,
                                             const wchar_t *oth,
                                             size_t len);

template<> bool FastString<wchar_t>::AddULong(unsigned long tmp);

template<> bool FastString<wchar_t>::AddInt(int tmp);
#endif

// ----------------------------------------------------------------------------
// MGPStr
// ----------------------------------------------------------------------------

//! FastString based on _TCHAR
class MGPStr : public FastString<_TCHAR>
{
 public:
  //! Assignment Ctor
  MGPStr(const _TCHAR *AStr)
  {
    Init();
    SetStr(AStr);
  }
  //! Assignment Ctor
  MGPStr(const MGPStr &AStr)
  {
    Init();
    SetStr(AStr);
  }
  //! Assignment operator
  void operator =(const MGPStr &AStr)
  {
    SetStr(AStr);
  }
  //! Ctor, creates ""
  MGPStr()
  {
    Init();
  }
};

// ----------------------------------------------------------------------------
// CharString
// ----------------------------------------------------------------------------

//! FastString based on char
class CharString : public FastString<char>
{
 public:
  //! Assignment Ctor
  CharString(const char *AStr)
  {
    Init();
    SetStr(AStr);
  }
  //! Assignment Ctor
  CharString(const CharString &AStr)
  {
    Init();
    SetStr(AStr);
  }
  //! Assignment operator
  void operator =(const CharString &AStr)
  {
    SetStr(AStr);
  }
  //! Ctor, creates ""
  CharString()
  {
    Init();
  }
};

// ----------------------------------------------------------------------------
// MGPStrSet
// ----------------------------------------------------------------------------

//! Comparison function for MGPStrSet
struct lessMGPStr : binary_function<MGPStr, MGPStr, bool>
{
  //! Comparison (less than)
  bool operator() (const MGPStr& s1, const MGPStr& s2) const
  {
    return (_tcscmp(s1.c_str(),s2.c_str()) < 0);
  }
};

//! Set of MGPStrs
typedef set<MGPStr,lessMGPStr> MGPStrSet;


