// Version 5.33  30.03.2007 Markus Greither
// Copyright (c) 1991-2007 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file kettobj.h
    \brief Header for the KettObj chained list data management class
*/

// ----------------------------------------------------------------------------
// KettObj
// ----------------------------------------------------------------------------

#if 1 // <list> based implementation

#include <list>

//! Template class for dynamic lists, based on <list>
template <class T> class KettObj : protected std::list <T>
{
 protected:
  //! Embedded iterator
  typename std::list <T>::iterator iter;
  //! Position storage
  size_t Nummer;
  //! Initialize
  void AnfWert()
  {
    iter = list<T>::begin();
    Nummer = 0;
  }
  //! Create new class T data field
  virtual T *InitFelder()
  {
    return new T;
  }

 public:
  //! Ctor, initializes with one item
  KettObj(bool neueint = true)
  {
    AnfWert();
    if (neueint)
      NeuAkEint();
  }
  //! Dtor
  virtual ~KettObj()
  {
    list<T>::clear();
  }
  //! Copy ctor
  KettObj(const KettObj<T> &other) : std::list<T>()
  {
    *this = other;
  }
  //! Return reference to current item
  inline T &GetAkEint()
  {
#ifdef __MGDEBUG__
    if ((GetMaxEintrag() < 0) || ((int)Nummer > GetMaxEintrag()))
      throw FastString<char>("KettObj: GetAkEint() invalid");
#endif
    return *iter;
  }
  //! Operator return for current item
  inline T *operator->()
  {
#ifdef __MGDEBUG__
    if ((GetMaxEintrag() < 0) || ((int)Nummer > GetMaxEintrag()))
      throw FastString<char>("KettObj: operator->() invalid");
#endif
    return &(*iter);
  }
  //! Returns index of last item
  inline int GetMaxEintrag() const
  {
    return int(list<T>::size())-1;
  }
  //! Returns index of current item
  inline int AkEintNummer() const
  {
    return Nummer;
  }
  //! Activates next higher item, returns false if already the last item
  bool NextEint()
  {
    if (iter == list<T>::end())
      return false;
    else
    if (++iter != list<T>::end())
    {
      Nummer++;
      return true;
    }
    else
    {
      iter--;
      return false;
    }
  }
  //! Activate first item
  bool Begin()
  {
    iter = list<T>::begin();
    Nummer = 0;
    return (list<T>::size() > 0);
  }
  //! Activate item Nr
  bool SetAkEint(size_t Nr)
  {
    if (Nr == 0)
      return Begin();
    else
    if (Nr < list<T>::size())
    {
      advance(iter,int(Nr-Nummer));
      Nummer = Nr;
      return true;
    }
    else
      return false;
  }
  //! Create item Nr if not existing, Activate Nr
  bool CreateAkEint(int Nr);
  //! Add new item at the end
  bool NeuAkEint()
  {
    T *tmp = InitFelder();
    list<T>::push_back(*tmp);
    delete tmp;
    Nummer = list<T>::size()-1;
    iter = --list<T>::end();
    return false;
  }
  //! Insert new item at current position
  bool InsAkEint()
  {
    T *tmp = InitFelder();
    iter = list<T>::insert(iter,*tmp);
    delete tmp;
    return true;
  }
  //! Delete current item
  void LoeschAkEint()
  {
    list<T>::erase(iter);
    Begin();
  }
  //! Delete all items, creates one item if neueint = true
  void ResetKette(bool neueint = true)
  {
    list<T>::clear();
    AnfWert();
    if (neueint)
      NeuAkEint();
  }
  //! Determine whether any items are available
  bool IstLeer()
  {
    return list<T>::empty();
  }
  //! Copy Ctor
  KettObj<T> &operator = (const KettObj<T> &other)
  {
    list<T>::assign(other.begin(),other.end());
    iter = list<T>::begin();
    advance(iter,other.AkEintNummer());
    Nummer = other.Nummer;
    return *this;
  }
};

  //! Create item Nr if not existing, Activate Nr
template <class T> bool KettObj<T>::CreateAkEint(int Nr)
{
  if (Nr > GetMaxEintrag())
    while (Nr > GetMaxEintrag())
      NeuAkEint();
  else
    SetAkEint(Nr);
  return true;
}

#else

//! Template class for dynamic lists
template <class T> class KettObj
{
  //! Data item for chained list
  struct FuEintRec
  {
    FuEintRec *Next;
    T *Felder;
  };

 protected:
  //! Current item
  FuEintRec *AkEint;
  //! First item of list
  FuEintRec *ErstEint;
  //! Current index
  int Nummer;
  //! Highest available index
  int MaxEintrag;
  //! Initialize
  void AnfWert();
  //! Delete all items
  void LoeschKette();
  //! Delete class T data field
  virtual void DelFelder(T *p) {delete p;}
  //! Create new class T data field
  virtual T *InitFelder() {return new T;}

 private:
  //! Delete data item
  void DispEintrag(FuEintRec *p);
  //! Create new data item
  FuEintRec *NeuEintrag()
  {
    FuEintRec *HiEint;

    HiEint = new FuEintRec;
    if (HiEint)
    {
      HiEint->Next = NULL; // Wenn Erstellung erfolgreich, NEXT-Feld auf NULL
#if 0 // Test copy ctors
      T *org = InitFelder();
      T *tmp = new T(*org);
      HiEint->Felder = tmp;
      if (org != tmp)
      {
        FILE *file = fopen("c:\\tst.txt","a");
        fprintf(file,"%s\n",typeid(org).name());
        fclose(file);
      }
      delete org;
#else
      HiEint->Felder = InitFelder();
#endif
    }
    return HiEint;
  }


 public:
  //! Ctor, initializes with one item
  KettObj(bool neueint = true)
  {
    AnfWert();
    if (neueint)
      NeuAkEint();
  }
  //! Copy Ctor
  KettObj(const KettObj<T> &other)
  {
    AnfWert();
    NeuAkEint();
    *this = other;
  }
  //! Dtor
  virtual ~KettObj()
  {
    LoeschKette();
  }
  //! Return reference to current item
  inline T &GetAkEint()
  {
#ifdef __MGDEBUG__
    if ((GetMaxEintrag() < 0) || ((int)Nummer > GetMaxEintrag()))
      throw FastString<char>("KettObj: GetAkEint() invalid");
#endif
    return (*AkEint->Felder);
  }
  //! Operator return for current item
  inline T *operator ->()
  {
#ifdef __MGDEBUG__
    if ((GetMaxEintrag() < 0) || ((int)Nummer > GetMaxEintrag()))
      throw FastString<char>("KettObj: operator->() invalid");
#endif
    return AkEint->Felder;
  }
  //! Returns index of last item
  inline int GetMaxEintrag() const
  {
    return (MaxEintrag);
  }
  //! Returns index of current item
  inline int AkEintNummer() const
  {
    return (Nummer);
  }
  //! Activates next higher item, returns false if already the last item
  bool NextEint();
  //! Activate first item
  bool Begin() {SetAkEint(0); return (GetMaxEintrag() >= 0);}
  //! Activate item Nr
  bool SetAkEint(int Nr);
  //! Create item Nr if not existing, Activate Nr
  bool CreateAkEint(int Nr);
  //! Add new item at the end
  bool NeuAkEint();
  //! Insert new item at current position
  bool InsAkEint();
  //! Delete current item
  void LoeschAkEint();
  //! Delete all items, creates one item if neueint = true
  void ResetKette(bool neueint = true);
  //! Determine whether any items are available
  bool IstLeer()
  {
    return (ErstEint == NULL);
  }
  //! Copy operator
  KettObj<T> &operator = (const KettObj<T> &other);
};

template <class T> KettObj<T> &KettObj<T>::operator = (const KettObj<T> &other)
{
  ResetKette(false);

  FuEintRec *OthAk = other.ErstEint;
  if (OthAk)
  {
    NeuAkEint();
    GetAkEint() = *OthAk->Felder;
  }  

  while (OthAk && OthAk->Next)
  {
    OthAk = OthAk->Next;
    NeuAkEint();
    GetAkEint() = *OthAk->Felder;
  }
  // Force first index (no delta)
  SetAkEint(0);
  SetAkEint(other.Nummer);
  return *this;
}

template <class T> void KettObj<T>::AnfWert()
{
  AkEint = NULL;
  ErstEint = NULL;
  MaxEintrag = -1;
  Nummer = 0;
}

template <class T> void KettObj<T>::DispEintrag(FuEintRec *p)
{
  DelFelder(p->Felder);
  delete p;
}

template <class T> void KettObj<T>::ResetKette(bool neueint)
{
  LoeschKette();
  AnfWert();
  if (neueint)
    NeuAkEint();
}

template <class T> bool KettObj<T>::NextEint()
{
  if (AkEint && AkEint->Next)
  {
    AkEint = AkEint->Next;
    Nummer++;
    return true;
  }
  else
    return false;
}

template <class T> bool KettObj<T>::SetAkEint(int Nr)
{
  if ((Nr < Nummer) || (Nr == 0))
  {
    AkEint = ErstEint;
    Nummer = 0;
  }
  while (AkEint && AkEint->Next && (Nummer < Nr))
  {
    AkEint = AkEint->Next;
    Nummer++;
  }
  return (Nr == Nummer) && (GetMaxEintrag() >= 0);
}

template <class T> bool KettObj<T>::InsAkEint()
{
  FuEintRec *HiEint;

  if (!ErstEint)
  {
    AkEint = NeuEintrag();
    if (AkEint)
      ErstEint = AkEint;
    else
      AkEint = NULL;
  }
  else
  {
    HiEint = NeuEintrag();
    if (HiEint)
    {
      // Swop date fields (avoids changing predesessor of AkEint)
      T *tmp = HiEint->Felder;
      HiEint->Felder = AkEint->Felder;
      AkEint->Felder = tmp;
      HiEint->Next = AkEint->Next;
      AkEint->Next = HiEint;
    }
    else
      AkEint = NULL;
  }
  if (AkEint)
    MaxEintrag++;
  return (AkEint == NULL);
}

template <class T> bool KettObj<T>::NeuAkEint()
{
  FuEintRec *HiEint;

  if (!ErstEint)
  {
    AkEint = NeuEintrag();
    if (AkEint)
      ErstEint = AkEint;
    else
      AkEint = NULL;
  }
  else
  {
    while (AkEint->Next)
    {
      AkEint = AkEint->Next;
      Nummer++;
    }
    HiEint = NeuEintrag();
    if (HiEint)
    {
      AkEint->Next = HiEint;
      AkEint = HiEint;
      Nummer++;
    }
    else
      AkEint = NULL;
  }
  if (AkEint)
    MaxEintrag++;
  return (AkEint == NULL);
}

template <class T> bool KettObj<T>::CreateAkEint(int Nr)
{
  while (AkEintNummer() < Nr)
     NeuAkEint();
  return SetAkEint(Nr);
}

template <class T> void KettObj<T>::LoeschAkEint()
{
  FuEintRec *HiEint;
  FuEintRec *HiEint1;
  int i;

  HiEint = ErstEint;
  i = 0;
  while (HiEint)
  {
    if (i == Nummer-1)
    {
      HiEint1 = HiEint->Next;
      HiEint->Next = HiEint1->Next;
      DispEintrag(HiEint1);  // delete "middle" entry
      HiEint = 0;
    }
    else
    if ((i == 0) && (Nummer == 0))
    {
      HiEint = ErstEint->Next;
      DispEintrag(ErstEint);
      ErstEint = HiEint;
      HiEint = 0;
    }
    else
      HiEint = HiEint->Next;
    i++;
  }
  if (Nummer > 0)
    Nummer--;
  if (MaxEintrag >= 0)
    MaxEintrag--;
  SetAkEint(0);
}

template <class T> void KettObj<T>::LoeschKette()
{
  AkEint = ErstEint;
  while (AkEint)
  {
    ErstEint = AkEint;
    AkEint = AkEint->Next;
    DispEintrag(ErstEint);
  }
  AnfWert();
}

#endif

//! Template for dynamic lists without default element
template <class T> class ListObj : public KettObj<T>
{
 public:
  //! Ctor
  ListObj() : KettObj<T>(false) {}
  //! Clear function
  void ResetKette()
  {
    KettObj<T>::ResetKette(false);
  }
};

