// Version 6.05 20.07.2008 Markus Greither
// Copyright (c) 1991-2008 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file effnkclc.cpp
    \brief Source code for the function calculator class
*/

#ifdef __BORLANDC__

#include "efprec.h"

#pragma hdrstop
#else

#ifdef __DMC__
#include <typeinfo.h>
#endif
#include <cstdlib>
#include <cmath>
#include <complex>
#include <limits>
#include <cstring>
#include <set>
#include <functional>
#include <cctype>
#include <cfloat>

using namespace std;

#include "efdefs.h"
#include "mg.h"
#include "cpxpsr.h"
#include "diagtype.h"
#include "efgemein.h"
#include "efmcxio.h"

#endif

#include "effnkclc.h"

// ----------------------------------------------------------------------------
// TEFFunktionCalc
// ----------------------------------------------------------------------------

TEFFunktionCalc::TEFFunktionCalc(ParserObj *Psr,TEFFunktionCalcData &d)
 : Data(d)
{
  Parser = Psr;
  AkFeld = new GraphDataObj;
  Data.Fehler.ClearError();
  KoeffVar = 0;
  Parser->ResetFunktions();
  if (Data.AkFeld)
    ((Data.DiagData->DarstData)[0]).XAxis.SetAkEint(Data.AkFeld->XAxisIndex[0]);
  else  
    ((Data.DiagData->DarstData)[0]).XAxis.Begin();
  XAxis = ((Data.DiagData->DarstData)[0]).XAxis.GetAkEint();
  Punkte = Data.DiagData->Punkte;
  Debug = Data.DiagData->Debug;
  Ma = Data.DiagData->Ma;
  XLog = Data.DiagData->XLog;

  if (Punkte > MaxGaugeSteps)
    GaugeStep = Punkte/MaxGaugeSteps;
  else
    GaugeStep = 1;
  if (Data.DiagData->XMode == 1)
  {
    XAxis.Min = Data.DiagData->CXMin;
    XAxis.Max = Data.DiagData->CXMax;
  }
  else
    XLog = XAxis.GetLog();
  if ((Data.FuName.Length() > 0) && (Data.FuName[0] == '[') &&
      (Data.FuName[Data.FuName.Length()-1] == ']'))
  {
    HiString.SetStr(&Data.FuName[1]);
    HiString.ShortStr(HiString.Length()-1);
  }
  else
    HiString.SetStr(Data.FuName);
  Ma.FunktionsBezeichner(HiString,Data.DiagData->YDaten);
  if (XLog == DarstDataObj::linX)
    Step = (XAxis.Max-XAxis.Min)/(Punkte-1);
  else
    Step = exp(log(XAxis.Max/XAxis.Min)/(Punkte-1));
  if (Data.AkFeld)
    Punkte = Data.AkFeld->GetPunkte();
  HiVar = Ma.GetFirst();
  AkPar = 1; // aktueller Parameter
  OldPar = AkPar;
  if (Ma.Mode == MaTyp::ModeFourier)
  {
    FourierMode = true;
    if (Ma.Anf != 0)
      FourierStep = (int)ceil(Ma.Step/Ma.Anf);
    else
      FourierStep = 1;
    FourierObw = int((Ma.Ende-Ma.Anf)/Ma.Step);
    FGaugeStep = FourierObw;
    delete[] KoeffVar;
    KoeffVar = new Cpx[FourierObw+10];
    if (FGaugeStep > MaxGaugeSteps)
      FGaugeStep = FGaugeStep/MaxGaugeSteps;
    else
      FGaugeStep = 1;
  }
  else
  {
    KoeffVar = 0;
    FourierMode = false;
  }
  AkFeld->SetPunkte(Punkte);
  AkPt = 0;
  AkKoeff = 0;
  y = Cpx(0,0);
  InitDiffLock = false;
}

TEFFunktionCalc::~TEFFunktionCalc()
{
  delete AkFeld;
  if (KoeffVar)
    delete[] KoeffVar;
}

void TEFFunktionCalc::InitDifferentiation()
{
  InitDiffLock = true;
  if (!Data.TestModified &&
      ((Ma.Mode == MaTyp::ModeNo) || (Ma.Mode == MaTyp::ModePar) ||
       (Ma.Mode == MaTyp::ModeVar) || (Ma.Mode == MaTyp::ModeParList)))
  {
    int OrgAkPar = AkPar;
    double OrgHiVar = HiVar;
    NextStep();
    NextStep(); // Compute 2nd point for valid LastX/LastY at d()
    // Prevent that NextStep() resets InitDiffLock
    InitDiffLock = true;
    AkPt = 0;
    AkKoeff = 0;
    y = Cpx(0,0);
    HiVar = OrgHiVar;
    AkPar = OrgAkPar;
  }
}

bool TEFFunktionCalc::NextStep()
{
  // Check if new Ma parameter, if yes reset function for fist point
  if ((OldPar != AkPar) && Ma.GetNext(HiVar))
  {
    HiString.SetStr(Data.FuName);
    AkPt = 0;
    Parser->ResetFunktions();
    InitDiffLock = false;
    OldPar = AkPar;
  }

  bool Tst = (Data.Fehler.Art == ParserObj::NoError) && (AkPt < Punkte);
  if ((Data.Fehler.Art != ParserObj::NoError) ||
      (AkPt >= Punkte) && !(Ma.NextStepAvail(HiVar)))
    return false;
  if ((AkPt == 0) && !InitDiffLock)
    InitDifferentiation();
  if (FourierMode == false)
  {
    Tst &= WerteBerechnung();
    if ((Tst == false) && (Data.Fehler.Art == ParserObj::NoError))
    {
      bool tmp = Ma.NextStepAvail(HiVar);
      if (tmp) // either MaTyp::ModePar or MaTyp::ModeParList
      {
        AkPar++;
        Tst = true;
      }
    }
  }
  else
  {
    Tst &= FourierBerechnung();
    if ((Tst == false) && (Data.Fehler.Art == ParserObj::NoError))
    {
      Tst = true;
      FourierMode = false;
    }
  }
  return Tst;
}

int TEFFunktionCalc::GetGaugeValue()
{
  if (FourierMode)
    return int(AkKoeff*MaxGaugeSteps/FourierObw-1);
  else
    return int(AkPt*MaxGaugeSteps/Punkte-1);
}

int TEFFunktionCalc::GetGaugeRange()
{
  return int(MaxGaugeSteps);
}

bool TEFFunktionCalc::GetFinished()
{
  return ((Data.Fehler.Art != ParserObj::NoError) || (AkPt >= Punkte) &&
          !(Ma.NextStepAvail(HiVar)));
}

void TEFFunktionCalc::GetFunktionName(MGPStr &str,double *Par)
{
  str.SetStr(HiString);
  MGPStr Zeile;
  Ma.ParBezeichner(Zeile,Par ? *Par : HiVar,Data.DiagData->YDaten);
  str.AddStr(Zeile);
}

void TEFFunktionCalc::SetDefaultGraphOptions(const _TCHAR *HiString,
                                             int GraphSymbol,
                                             int GraphOptionen,
                                             int GraphDarst)
{
  _tcscpy(AkFeld->FuString,HiString);
  AkFeld->GraphSymbol = (unsigned short)GraphSymbol;
  AkFeld->GraphOptionen = (unsigned short)(GraphOptionen | GraphDataObj::GraphDia1 | 
                                           GraphDataObj::GraphDia2);
  AkFeld->ZeichenArt = (unsigned short)GraphDarst;
  AkFeld->FileString[0] = 0;
}

void TEFFunktionCalc::UpdateGraph()
{
  AkFeld->GraphSymbol = Data.DiagData->YDaten.AkGraph().GraphSymbol;
  AkFeld->GraphOptionen = Data.DiagData->YDaten.AkGraph().GraphOptionen;
  AkFeld->ZeichenArt = Data.DiagData->YDaten.AkGraph().ZeichenArt;
  AkFeld->Farbe = Data.DiagData->YDaten.AkGraph().Farbe;
  AkFeld->XAxisIndex[0] = Data.DiagData->YDaten.AkGraph().XAxisIndex[0];
  AkFeld->YAxisIndex[0] = Data.DiagData->YDaten.AkGraph().YAxisIndex[0];
  AkFeld->XAxisIndex[1] = Data.DiagData->YDaten.AkGraph().XAxisIndex[1];
  AkFeld->YAxisIndex[1] = Data.DiagData->YDaten.AkGraph().YAxisIndex[1];
  AkFeld->SetPunkte(AkPt);
  Data.DiagData->AkGraph = Data.DiagData->YDaten.GetAkPos();
  Data.DiagData->YDaten.ModifyGraph(AkFeld,false,true);
}

void TEFFunktionCalc::AppendNewGraph(const _TCHAR *HiString)
{
  _tcscpy(AkFeld->FuString,HiString);
  AkFeld->FileString[0] = 0;
  AkFeld->SetPunkte(AkPt);
  Data.DiagData->YDaten.AppendYFeld(AkFeld);
  Data.DiagData->AkGraph = Data.DiagData->YDaten.GetAkPos();
}

bool TEFFunktionCalc::FourierBerechnung()
{
  bool Modified = Parser->FunktionsBerechnung(KoeffVar[AkKoeff],HiVar,y,
                                              Data.Fehler,
                                              HiVar,Debug == 1,Data.BerFus);
  // Prevent multiple evaluation of modified functions
  Data.BerFus = 0;
  if (Data.TestModified)
  {
    Data.TestModified = Modified; // Flag for modified
    AkPt = Punkte+1;
    return false;
  }
  bool tst = Ma.GetNext(HiVar);
  AkKoeff++;
  if (tst && (Data.Fehler.Art == ParserObj::NoError))
    return true;
  else
    return false;
}

bool TEFFunktionCalc::WerteBerechnung()
{
  bool Modified = false;
  Cpx AkY = Cpx(0,0);
  if (!Data.AkFeld)
  { // Compute new graph
    if (XLog == DarstDataObj::linX)
      AkX = XAxis.Min+Step*AkPt;
    else
    if (AkPt == 0)
      AkX = XAxis.Min;
    else
      AkX *= Step;
  }
  else
  { // Compute available graph into new graph
    AkX = Data.AkFeld->GetX(AkPt);
    y = Data.AkFeld->GetY(AkPt);
  }
  double NeuX = AkX; // temp. X (for correction in MaTyp::ModeFourier mode)
  if (Ma.Mode == MaTyp::ModeSum)
  {
    HiVar = Ma.Anf;
    bool nxt;
    do
    {
      Modified = Parser->FunktionsBerechnung(z,AkX,y,Data.Fehler,HiVar,
                                             Debug == 1,Data.BerFus);
      AkY += z;
      nxt = Ma.GetNext(HiVar);
    }
    while (nxt && (Data.Fehler.Art == ParserObj::NoError) &&
           !Data.TestModified);
  }
  else
  if ((Ma.Ende >= Ma.Anf) && (Ma.Mode == MaTyp::ModeFourier))
  {
    int neg = (AkX < 0 ? -1 : 1); // Determine sign
    double AbsX = fabs(AkX); // get absolute of x
    double FakX = floor(AbsX*Ma.Anf); // compute full periods
    double Per = AbsX*Ma.Anf-FakX; // define period to 0-1
    double Rep = 1.0/FourierStep; // repitition rate
    while (Per >= Rep)
    { // subtract multiple repititions
      Per -= Rep;
      FakX += Rep;
    }
    long NextObw;
    // too small for multi use ->
    // all points with different angles !
    if (Per < 0.0001)
      NextObw = 0;
    else
      // computation of repition factor
      NextObw = long(1000.0/(Per*FourierStep));
    double w;
    if (NextObw > 0)
    {
      w = (FakX+1000.0/(NextObw*FourierStep))*neg; // compute NeuX*Ma.Anf
      NeuX = w/Ma.Anf; // compute corrected X

      w *= (2*M_PI); // corrected w
    }
    else
      w = AkX*2*M_PI*Ma.Anf;
    double wdel = w*FourierStep;
    double diff = 1;
    double Y = 0;
    int FourierZaehler = 0;
    double Per1,Per2,HiI,HiR;
    do
    {
//      HiR = sin(w);
      Per = w-floor(w*M_1_PI/2)*2*M_PI; // w between 0 and 2*Pi
      if (Per > M_PI)
        Per1 = 2*M_PI-Per; // mirror at Pi
      else Per1 = Per;
      if (Per1 > M_PI_2)
        Per1 = M_PI-Per1; // mirror at Pi/2
      Per2 = Per1*Per1;
      if (Per > M_PI)
        HiR = -Per1*(1+Per2*(-1.0/6+Per2*(1.0/120-Per2/5040))); // HiR = Sin(w)
      else
        HiR = Per1*(1+Per2*(-1.0/6+Per2*(1.0/120-Per2/5040))); // HiR = Sin(w)
      if ((Per > M_PI_2) && (Per < 3*M_PI_2))
        HiI = sqrt(1-HiR*HiR);
      else
        HiI = -sqrt(1-HiR*HiR);
//      HiI = -cos(w);
      long t = 0;
      if (NextObw > 0)
      do
      {
        Y += (real(KoeffVar[int(t+FourierZaehler)])*HiR-
              imag(KoeffVar[int(t+FourierZaehler)])*HiI)/(diff+t*FourierStep);
        t += NextObw; // go to next harmonic with same phase
      }
      while (t+FourierZaehler <= FourierObw);
      else
      {
        Y += (real(KoeffVar[int(FourierZaehler)])*HiR-
              imag(KoeffVar[int(FourierZaehler)])*HiI)/diff;
      }
      FourierZaehler++;
      w += wdel;
      diff += FourierStep;
    }
    while ((FourierZaehler <= FourierObw) && ((NextObw == 0) ||
           (FourierZaehler < NextObw)));
    AkY = Cpx(Y,0);
  }
  else
    Modified = Parser->FunktionsBerechnung(AkY,AkX,y,Data.Fehler,HiVar,
                                           Debug == 1,Data.BerFus);

  // Prevent that computed functions are determined multiple times
  Data.BerFus = 0;
  if (Data.TestModified)
  {
    AkPt = Punkte+1;
    Data.TestModified = Modified; // Detect that modified
  }
  // Prevent multiple evaluation of modified functions
  Data.BerFus = 0;
  if ((AkPt < Punkte) && (Data.Fehler.Art == ParserObj::NoError))
  {
    AkPt++;
    AkFeld->SetX(AkPt-1,NeuX);
    AkFeld->SetY(AkPt-1,AkY);
    return true;
  }
  else
    return false;
}
