// Version 6.06  04.12.2009 Markus Greither
// Copyright (c) 1991-2009 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efpsrplc.cpp
    \brief Source code for the function parser object public interface
*/

#ifdef __BORLANDC__

#include "efprec.h"

#pragma hdrstop
#else

#ifdef __DMC__
#include <typeinfo.h>
#endif
#include <cstdlib>
#include <cmath>
#include <complex>
#include <limits>
#include <cstring>
#include <set>
#include <functional>
#include <cctype>
#include <cfloat>

using namespace std;

#include "efdefs.h"
#include "mg.h"
#include "cpxpsr.h"

#endif

#if defined(__BORLANDC__) && defined(__WIN32__)
#include <excpt.h>
#endif

// ----------------------------------------------------------------------------
// Variables
// ----------------------------------------------------------------------------

const unsigned char ParserObj::FuAnzArr[] =
{
  0,
  FuAnz1,
  FuAnz2,
  FuAnz3,
  FuAnz4,
  FuAnz5,
  FuAnz6
};

const unsigned char ParserObj::FuAnfArr[] =
{
  0,
  0,
  FuAnz1,
  FuAnz1+FuAnz2,
  FuAnz1+FuAnz2+FuAnz3,
  FuAnz1+FuAnz2+FuAnz3+FuAnz4,
  FuAnz1+FuAnz2+FuAnz3+FuAnz4+FuAnz5
};

ParserObj::FuTypType ParserObj::FuTypArr[] =
{
{_EFT("All"),_EFT("All available funktions, Operators and Labels.")},
{_EFT("General Functions"),_EFT("General mathematical funktions.")},
{_EFT("Labels"),_EFT("All Labels of entered equations.")},
{_EFT("Complex Functions"),_EFT("Special function for complex numbers.")},
{_EFT("Constants"),_EFT("General mathematical constants.")},
{_EFT("Variables"),_EFT("Variables that get set during calculation.")},
{_EFT("Obsolete"),_EFT("Functions for compatibility with older program versions.")},
{_EFT("Operators"),_EFT("All available operators.")},
{_EFT("Conversion Functions"),_EFT("Functions for conversion of numbers in other systems.")},
{_EFT("Angular Functions"),_EFT("All available triangular functions.")}
};

const _TCHAR *ParserObj::BezString = _EFT("Returns the value of the function with the specified name.");

ParserObj::FunkDefType ParserObj::FuDefArr[] =
{
{_TEXT(""),_TEXT(""),_TEXT(""),0,0,_TEXT("")},
{_TEXT("X")     ,_TEXT("")      ,_EFT("$")        ,0,FuLaufvar,_EFT("Variable. Holds the x-values during calculation.")},
{_TEXT("F")     ,_TEXT("")      ,_EFT("$")        ,0,FuLaufvar,_EFT("Variable. Holds the x-values during calculation.")},
{_TEXT("T")     ,_TEXT("")      ,_EFT("$")        ,0,FuLaufvar,_EFT("Variable. Holds the x-values during calculation.")},
{_TEXT("W")     ,_TEXT("")      ,_EFT("$")        ,0,FuLaufvar,_EFT("Variable. Holds 2*PI*X during calculation.")},
{_TEXT("J")     ,_TEXT("")      ,_EFT("ERROR")    ,0,FuKonst,_EFT("Imaginary Number. (J*J = -1).")},
{_TEXT("Y")     ,_TEXT("")      ,_EFT("ERROR")    ,0,FuLaufvar,_EFT("Current result for the PostProc function or S (DIFF) function.")},
{_TEXT("E")     ,_TEXT("")      ,_EFT("exp(1)")   ,0,FuKonst,_EFT("Constant. E = 2.718281828.")},
{_TEXT("S")     ,_TEXT("")      ,_EFT("ERROR")    ,2,FuAllgemein,_EFT("Integrates function Arg2. Arg1 is the optional start value.")},
{_TEXT("D")     ,_TEXT("")      ,_EFT("ERROR")    ,1,FuAllgemein,_EFT("Differentiates function Arg1, using the values of Arg1 at the current and the previous x value.")},
{_TEXT("PI")    ,_TEXT("")      ,_EFT("pi()")     ,0,FuKonst,_EFT("Constant. PI = 3.141592654.")},
{_TEXT("HI")    ,_TEXT("")      ,_EFT("$")        ,0,FuLaufvar,_EFT("Helper variable. Holds the values from the Calculation Options-Window during calculation.")},
{_TEXT("IM")    ,_TEXT("")      ,_EFT("imreal")   ,1,FuKomplex,_EFT("Returns imaginary part of Arg1.")},
{_TEXT("RE")    ,_TEXT("")      ,_EFT("imaginary"),1,FuKomplex,_EFT("Returns real part of Arg1.")},
{_TEXT("DB")    ,_TEXT("")      ,_EFT("ERROR")    ,1,FuUmwandeln,_EFT("Converts Arg1 to deziBel 20*LOG(ABS(Arg1)).")},
{_TEXT("LN")    ,_TEXT("EXP")   ,_EFT("ln")       ,1,FuAllgemein,_EFT("Returns natural logarithm of Arg1.")},
{_TEXT("XC")    ,_TEXT("")      ,_EFT("ERROR")    ,1,FuKomplex,_EFT("Returns the impedance of a capacitor of value Arg1 [Farad] at frequency F.")},
{_TEXT("XL")    ,_TEXT("")      ,_EFT("ERROR")    ,1,FuKomplex,_EFT("Returns the impedance of a inductor of value Arg1 [Henry] at frequency F.")},
{_TEXT("IF")    ,_TEXT("")      ,_EFT("if")       ,3,FuAllgemein,_EFT("Condition. Is Arg1 valid, returns Arg2, else Arg3.")},
{_TEXT("CPX")   ,_TEXT("")      ,_EFT("ERROR")    ,2,FuObsolete,_EFT("Defines complex number with real part Arg1 and imaginary part Arg2.")},
{_TEXT("CDV")   ,_TEXT("")      ,_EFT("ERROR")    ,2,FuKomplex,_EFT("Returns the value of the complex divider Arg1/(Arg1+Arg2).")},
{_TEXT("CPA")   ,_TEXT("")      ,_EFT("ERROR")    ,2,FuObsolete,_EFT("Returns the parallel value of  Arg1 and Arg2 Arg1|Arg2.")},
{_TEXT("MAX")   ,_TEXT("")      ,_EFT("max")      ,2,FuAllgemein,_EFT("Returns the larger value out of Arg1 and Arg2.")},
{_TEXT("MIN")   ,_TEXT("")      ,_EFT("min")      ,2,FuAllgemein,_EFT("Returns the smaller value out of Arg1 and Arg2.")},
{_TEXT("LOG")   ,_TEXT("")      ,_EFT("log10")    ,1,FuAllgemein,_EFT("Returns the logarithm of base 10 of Arg1.")},
{_TEXT("SIN")   ,_TEXT("ASIN")  ,_EFT("sin")      ,1,FuWinkel,_EFT("Returns the sine of Arg1 (in radians).")},
{_TEXT("COS")   ,_TEXT("ACOS")  ,_EFT("cos")      ,1,FuWinkel,_EFT("Returns the cosine of Arg1 (in radians).")},
{_TEXT("TAN")   ,_TEXT("ATAN")  ,_EFT("tan")      ,1,FuWinkel,_EFT("Returns the tangens of Arg1 (in radians).")},
{_TEXT("EXP")   ,_TEXT("LN")    ,_EFT("exp")      ,1,FuAllgemein,_EFT("Returns E^Arg1 (E = 2.718281828).")},
{_TEXT("SQT")   ,_TEXT("SQR")   ,_EFT("sqrt")     ,1,FuObsolete,_EFT("Returns the square root of Arg1.")},
{_TEXT("SQR")   ,_TEXT("SQRT")  ,_EFT("ERROR")    ,1,FuAllgemein,_EFT("Returns the square of Arg1.")},
{_TEXT("ABS")   ,_TEXT("")      ,_EFT("abs")      ,1,FuAllgemein,_EFT("Returns the absolute Arg1 (also for complex arguments).")},
{_TEXT("FAC")   ,_TEXT("")      ,_EFT("fact")     ,1,FuAllgemein,_EFT("Returns the Fakultaet of Arg1.")},
{_TEXT("SGN")   ,_TEXT("")      ,_EFT("ERROR")    ,1,FuAllgemein,_EFT("Returns 1 if Arg1 positiv, -1 if Arg1 negativ or 0 if Arg1 = 0.")},
{_TEXT("ODD")   ,_TEXT("")      ,_EFT("odd")      ,1,FuAllgemein,_EFT("Returns 1 if Arg1 odd of 0 if Arg1 even.")},
{_TEXT("INT")   ,_TEXT("")      ,_EFT("trunc")    ,1,FuAllgemein,_EFT("Returns the  Arg1.")},
{_TEXT("CXC")   ,_TEXT("")      ,_EFT("ERROR")    ,1,FuObsolete,_EFT("Identical with XC.")},
{_TEXT("CXL")   ,_TEXT("")      ,_EFT("ERROR")    ,1,FuObsolete,_EFT("Identical with XL.")},
{_TEXT("VAL")   ,_TEXT("")      ,_EFT("ERROR")    ,1,FuObsolete,_EFT("Returns the value of the function with number Arg1.")},
{_TEXT("CEX")   ,_TEXT("")      ,_EFT("imexp")    ,1,FuObsolete,_EFT("Identical with EXP.")},
{_TEXT("CAB")   ,_TEXT("")      ,_EFT("imabs")    ,1,FuObsolete,_EFT("Identical with ABS.")},
{_TEXT("CTI")   ,_TEXT("")      ,_EFT("ERROR")    ,1,FuObsolete,_EFT("Returns the delay time of the complex argument Arg1.")},
{_TEXT("CAN")   ,_TEXT("")      ,_EFT("ERROR")    ,1,FuObsolete,_EFT("Identical ANG.")},
{_TEXT("RAD")   ,_TEXT("")      ,_EFT("radians")  ,1,FuUmwandeln,_EFT("Converts the value of Arg1 in angles into radians Arg1*2*PI/360.")},
{_TEXT("DEG")   ,_TEXT("")      ,_EFT("degrees")  ,1,FuUmwandeln,_EFT("Converts the value of Arg1 im radians into angles Arg1*360/2/PI.")},
{_TEXT("XCF")   ,_TEXT("")      ,_EFT("ERROR")    ,2,FuKomplex,_EFT("Like XC, Arg2 is frequency F.")},
{_TEXT("XLF")   ,_TEXT("")      ,_EFT("ERROR")    ,2,FuKomplex,_EFT("Like XL, Arg2 is frequency F.")},
{_TEXT("ANG")   ,_TEXT("")      ,_EFT("ERROR")    ,1,FuKomplex,_EFT("Returns the angle of complex argument Arg1.")},
{_TEXT("AVG")   ,_TEXT("")      ,_EFT("ERROR")    ,2,FuAllgemein,_EFT("Calculates the average for argument Arg2. Arg1 is the optional start value.")},
{_TEXT("ARG")   ,_TEXT("")      ,_EFT("ERROR")    ,1,FuKomplex,_EFT("Returns the angle (radians) of complex argument Arg1.")},
{_TEXT("ASIN")  ,_TEXT("SIN")   ,_EFT("arcsin")   ,1,FuWinkel,_EFT("Returns the Arcus sine of Arg1 (in radians).")},
{_TEXT("ACOS")  ,_TEXT("COS")   ,_EFT("arccos")   ,1,FuWinkel,_EFT("Returns the Arcus cosine of Arg1 (in radians).")},
{_TEXT("ATAN")  ,_TEXT("TAN")   ,_EFT("arctan")   ,1,FuWinkel,_EFT("Returns the Arcus Tangens of Arg1 (in radians).")},
{_TEXT("SINH")  ,_TEXT("ARSINH"),_EFT("sinhyp")   ,1,FuWinkel,_EFT("Returns the Sinus Hyperbolicus of Arg1.")},
{_TEXT("COSH")  ,_TEXT("ARCOSH"),_EFT("coshyp")   ,1,FuWinkel,_EFT("Returns the Cosinus Hyperbolicus of Arg1.")},
{_TEXT("TANH")  ,_TEXT("ARTANH"),_EFT("tanhyp")   ,1,FuWinkel,_EFT("Returns the Tangens Hyperbolicus of Arg1.")},
{_TEXT("SQRT")  ,_TEXT("SQR")   ,_EFT("sqrt")     ,1,FuAllgemein,_EFT("Returns the square root of Arg1.")},
{_TEXT("DIFF")  ,_TEXT("")      ,_EFT("ERROR")    ,2,FuObsolete,_EFT("Defines Arg1 as start value of the differential equation Arg2 in the form Arg2 = y' = f(y,x). Identical with S.")},
{_TEXT("ARSINH"),_TEXT("SINH")  ,_EFT("arcsinhyp"),1,FuWinkel,_EFT("Returns the Area Sinus Hyperbolicus of Arg1.")},
{_TEXT("ARCOSH"),_TEXT("COSH")  ,_EFT("arccoshyp"),1,FuWinkel,_EFT("Returns the Area Cosinus Hyperbolicus of Arg1.")},
{_TEXT("ARTANH"),_TEXT("TANH")  ,_EFT("arctanhyp"),1,FuWinkel,_EFT("Returns the Area Tangens Hyperbolicus of Arg1.")},
{_TEXT("POLDEG"),_TEXT("")      ,_EFT("ERROR")    ,1,FuUmwandeln,_EFT("Converts the value of Arg1 in polar coordinates. Real part:absolute, Imaginary part:angle in degrees.")},
{_TEXT("POLRAD"),_TEXT("")      ,_EFT("ERROR")    ,1,FuUmwandeln,_EFT("Converts the value of Arg1 in polar coordinates. Real part:absolute, Imaginary part:angle in radians.")},
{_TEXT("KRTDEG"),_TEXT("")      ,_EFT("ERROR")    ,1,FuUmwandeln,_EFT("Converts the value of Arg1 in the format of POLDEG back into karthesian coordinates.")},
{_TEXT("KRTRAD"),_TEXT("")      ,_EFT("ERROR")    ,1,FuUmwandeln,_EFT("Converts the value of Arg1 in the format of POLRAD back into karthesian coordinates.")},
{_TEXT("+")     ,_TEXT("")      ,_EFT("+")        ,0,FuOperator,_EFT("Addition.")},
{_TEXT("-")     ,_TEXT("")      ,_EFT("-")        ,0,FuOperator,_EFT("Subtraction.")},
{_TEXT("*")     ,_TEXT("")      ,_EFT("*")        ,0,FuOperator,_EFT("Multiplication.")},
{_TEXT("/")     ,_TEXT("")      ,_EFT("/")        ,0,FuOperator,_EFT("Division.")},
{_TEXT("\\")    ,_TEXT("")      ,_EFT("/")        ,0,FuOperator,_EFT("Division. Fraction stroke.")},
{_TEXT("|")     ,_TEXT("")      ,_EFT("ERROR")    ,0,FuOperator,_EFT("Parallel operator. A|B = (A*B)/(A+B).")},
{_TEXT("^")     ,_TEXT("")      ,_EFT("^")        ,0,FuOperator,_EFT("Exponential operator. A^B = EXP(B*LN(A)).")},
{_TEXT("(")     ,_TEXT("")      ,_EFT("(")        ,0,FuOperator,_EFT("Bracket open.")},
{_TEXT(")")     ,_TEXT("")      ,_EFT(")")        ,0,FuOperator,_EFT("Bracket closed.")},
{_TEXT("=")     ,_TEXT("")      ,_EFT("=")        ,0,FuOperator,_EFT("Separates function label from term. Or equal operator (Use only within brackets !).")},
{_TEXT("<>")    ,_TEXT("")      ,_EFT("<>")       ,0,FuOperator,_EFT("Not equal operator. If Arg1 <> Arg2 returns 1, else 0.")},
{_TEXT("<=")    ,_TEXT("")      ,_EFT("<=")       ,0,FuOperator,_EFT("Smaller or equal operator. If Arg1 <= Arg2 returns 1, else 0 (Use only within brackets !).")},
{_TEXT(">=")    ,_TEXT("")      ,_EFT(">=")       ,0,FuOperator,_EFT("Larger or equal Operator. If Arg1 >= Arg2 returns 1, else 0 (Use only within brackets !).")},
{_TEXT("<")     ,_TEXT("")      ,_EFT("<")        ,0,FuOperator,_EFT("Smaller operator. If Arg1 < Arg2 returns 1, else 0.")},
{_TEXT(">")     ,_TEXT("")      ,_EFT(">")        ,0,FuOperator,_EFT("Larger operator. If Arg1 > Arg2 returns 1, else 0.")}
};

// ----------------------------------------------------------------------------
// ParserObj
// ----------------------------------------------------------------------------

bool ParserObj::SucheZeichen(const _TCHAR *Zeile,unsigned int &p,_TCHAR c)
{ // Zeichen suchen, Ende bei c
  int KlZaehler = 0;

  while (Zeile[p] && ((KlZaehler != 0) || (Zeile[p] != c)))
  {
    switch (Zeile[p])
    {
      case '(': // Klammern zaehlen
        KlZaehler++;
        break;
      case ')': // Klammern zaehlen
        KlZaehler--;
        if (KlZaehler < 0)
        {
          p++;
          return true;
        }
        break;
    }
    NextFunkZeichen(Zeile,p);
  }
  return KlZaehler != 0;
}

void ParserObj::NextFunkZeichen(const _TCHAR *Zeile,unsigned int &p)
{
  switch (Zeile[p])
  {
    case '[': // Labels ueberlesen
      p++;
      while (Zeile[p] && (Zeile[p] != ']'))
        p++;
      break;
    case '"': // Kommentare ueberlesen
      p++;
      while (Zeile[p] && (Zeile[p] != '"'))
        p++;
      break;
  }
  if (Zeile[p])
    p++;
}

int ParserObj::IstGleichPos(const _TCHAR *Zeile,bool *Fehler)
{ // sucht Position des rechtesten = Zeichens und gibt index zurueck
  unsigned int p = 0;
  int AkInd = -1;
  bool tmp = false;

  do
  {
    tmp |= SucheZeichen(Zeile,p,'=');
    if (Zeile[p] == '=')
    {
      AkInd = p;
      p++;
    }
  }
  while (!tmp && (AkInd != -1) && (p < _tcslen(Zeile)));
  if (Fehler)
    *Fehler = tmp;
  return AkInd;
}

bool ParserObj::PruefeKlammern(FunctionArrObj *Funktion)
{
  int p;
  int KlZaehler,LKlZaehler;
  _TCHAR a;
  bool Kommentar = false;

  p = 0;
  KlZaehler = 0;
  LKlZaehler = 0;
  while (((*Funktion)->Term[p] != 0) && (LKlZaehler<2) &&
   (KlZaehler >= 0) && (LKlZaehler >= 0))
  {
    a = (*Funktion)->Term[p];
    if (!Kommentar)
    {
      switch (a)
      {
        case '(' : KlZaehler++; break;
        case ')' : KlZaehler--; break;
        case '[' : LKlZaehler++; break;
        case ']' : LKlZaehler--; break;
        case '"' : Kommentar = true; break;
      }
    }
    else
    if (a == '"')
      Kommentar = false;
    if ((KlZaehler < 0) || (LKlZaehler < 0))
      return false;
    p++;
  }
  return ((KlZaehler == 0) && (LKlZaehler == 0));
}

unsigned char ParserObj::FuNummer(unsigned int &p,const _TCHAR *Zeile)
{ // sucht den Index der Funktion in Zeile an Stelle p
  const _TCHAR *Term;
  unsigned char i,n,HiN;
  int AltP;

  AltP = p;
  Term = &Zeile[p];
  n = 0;
  HiN = InvalidFkt;
  do
  {
    n++; // naechste Funktions-Gruppe (mit n Buchstaben)
    i = 1;
    while ((i <= FuAnzArr[n]) &&
           (MGStrnicmp(FuDefArr[i+FuAnfArr[n]].FuName,Term,n) != 0))
      i++;
    if (i <= FuAnzArr[n])
      HiN = (unsigned char)(i+FuAnfArr[n]);
    p++;
  }
  while ((n != MaxFuLength) && (p < _tcslen(Zeile)) &&
         (IST_BUCHS(Zeile[p])));
  if (HiN != InvalidFkt)
    p = AltP+_tcslen(FuDefArr[HiN].FuName);
  else
    p = AltP;
  return(HiN);
}

unsigned int ParserObj::LeerZNKomment(const _TCHAR *AkStr,int &p)
{ // ueberliest Leerzeichen und Kommentare in "" in AkStr ab Position p
  int AnfLastKomment = 0; // "offener" letzter Kommentar
  while (AkStr[p] == ' ')
    p++; // Leerzeichen vor Kommentar
  while (AkStr[p] == '"')
  { // wenn Kommentar gefunden
    AnfLastKomment = p;
    do
    {
      p++;
    }
    while (AkStr[p] && (AkStr[p] != '"'));
    if (AkStr[p] == '"')
    {
      p++; // Kommentar-Ende ueberlesen
      AnfLastKomment = 0; // Kein "offener" letzter Kommentar
    }
    while (AkStr[p] == ' ')
      p++; // Leerzeichen nach Kommentar
  }
  // Gibt die Anfangsposition des letzten "offenen" Kommentars zurueck
  return AnfLastKomment;
}

int ParserObj::GetParamsAnfPos(const _TCHAR *Zeile)
{ // sucht den Anfang der Parameterliste in Zeile
  unsigned int p = 0;
  int KlZ = 0;
  while ((p < _tcslen(Zeile)) && ((Zeile[p] != '(') || KlZ))
  {
    if (Zeile[p] == '[')
      KlZ++;
    else
    if (Zeile[p] == ']')
      KlZ--;
    p++;
  }
  if (p < _tcslen(Zeile))
    return p;
  else
    return 0;
}

bool ParserObj::GetParamsNamen(const _TCHAR *ParamStr,ParamsObj *Par)
{ // Namen aus der Parameterliste auswerten
  unsigned int p = 0,f;

  if (!ParamStr)
    ParamStr = (*Funktion)->ParamStr.c_str();
  if (!Par)
    Par = (*Funktion)->Params;
  Par->Name.clear();
  do
  {
    if (!_tcsncmp(&ParamStr[p],GetLDV().ListSeparator.c_str(),
                 GetLDV().ListSeparator.Length()))
      p += GetLDV().ListSeparator.Length();
    else
      p++; // ( oder Trennzeichen ueberlesen
    LeerZNKomment(ParamStr,p);
    if ((ParamStr[p] == '[') || (ParamStr[p] == '\''))
    { // Parameter lesen
      f = p;
      if (!NextAkWert(ParamStr,p) ||
          (ParamStr[f] == '[') && (ParamStr[p-1] != ']') ||
          (ParamStr[f] == '\'') && (ParamStr[p-1] != '\''))
      {
        Par->Name.clear();
        if (Par == (*Funktion)->Params)
          (*Funktion)->DelParams();
        return false;
      }
      Par->Name.resize(Par->Name.size()+1);
      MGPStr &tmp = Par->Name[Par->Name.size()-1];
      tmp.SetStr(_TEXT(""));
      tmp.AddNChar(&ParamStr[f],p-f);
      LeerZNKomment(ParamStr,p);
    }
    else
    { // Parameter ohne eckige Klammern
      f = p;
      SucheZeichen(ParamStr,p,GetLDV().ListSeparator[0]);
      if ((p > 0) && (p == _tcslen(ParamStr)) &&
          (ParamStr[p-1] == ')'))
        p--;
      Par->Name.resize(Par->Name.size()+1);
      MGPStr &tmp = Par->Name[Par->Name.size()-1];
      tmp.SetStr(_TEXT("["));
      tmp.AddNChar(&ParamStr[f],p-f);
      tmp.AddStr(_TEXT("]"));
    }
  }
  while ((p < _tcslen(ParamStr)) &&
         (ParamStr[p] != ')') &&
         !_tcsncmp(&ParamStr[p],GetLDV().ListSeparator.c_str(),
                  GetLDV().ListSeparator.Length()));
  if (ParamStr[p] != ')')
    return false;
  else
    p++;
  return true;
}

int ParserObj::BezeichnerFuNummer(const _TCHAR *AkStr)
{ // ermittelt die LETZTE Funktion mit der angegebenen Bezeichnung
  const _TCHAR *OldTerm = AkTerm; // originalen AkTerm retten
  unsigned int p = 0; // Bezeichnung ab 1. Zeichen durchsuchen
  MGPStr Tmp;
  Tmp.SetStr(AkStr);
  if (Tmp[0] != '[')
  { // Klammern hinzufuegen wenn nicht vorhanden
    Tmp.InsChar(0,'[');
    Tmp.AddChar(']');
  }
  AkTerm = Tmp.c_str(); // Temporaer auf gesuchtes Label aendern
  int tst = HoleLabelFuNummer(p);
  AkTerm = OldTerm; // originalen AkTerm zuruecksetzen
  return tst;
}

// ermittelt die Nummer der LETZTEN (UNTERSTEN) Funktion,
// die die Bezeichnung des Labels in AkTerm an Position p traegt
int ParserObj::HoleLabelFuNummer(unsigned int &p)
{
  const _TCHAR *LabStr;
  unsigned int AkFuNummer,AkLen;
  int tmp = 0; // Fehler wenn tmp 0 bleibt -> Label nicht gefunden !

  p++; // [ ueberlesen
  LabStr = &AkTerm[p];
  AkLen = 0;
  while (AkTerm[p] && (AkTerm[p] != ']'))
  { // Ende des Labels suchen
    p++;
    AkLen++;
  }
  p++;
  AkFuNummer = Funktion->AkEintNummer();
  Funktion->Begin();
  while (Funktion->NextEint())
  { // alle Funktions-Bezeichnungen durchsuchen
    MGPStr &name = (*Funktion)->Namen;
    if ((_tcsncmp(name.c_str(),LabStr,AkLen) == 0) &&
        (AkLen == name.Length()) ||
        (name[0] == '[') &&
        (_tcsncmp(&name.c_str()[1],LabStr,AkLen) == 0) &&
        (AkLen+2 == name.Length()))
      tmp = Funktion->AkEintNummer();
  }
  Funktion->SetAkEint(AkFuNummer);
  return(tmp);
}

void ParserObj::GetFehler()
{
 FehlerRec Fehl;
 switch (AkErr)
 {
   case DOMAIN: Fehl.Art = Domain; break;
   case OVERFLOW: Fehl.Art = Overflow; break;
   case SING: Fehl.Art = Singularity; break;
   case TLOSS: Fehl.Art = WrongValue; break;
   default: Fehl.Art = UnknownError; break;
 }
 Fehl.Pos = p;
 Fehl.FuNummer = Funktion->AkEintNummer();
 AkErr = 0;
 if (Fehl.Art != NoError)
   throw Fehl;
}

// liest den naechsten Ausdruck in AkStr ab Position p, Operatorausdruecke
// und mehrfache Funktionsaufrufe werden komplett abgearbeitet
_TCHAR ParserObj::AkWert(MGPStr &AkStr,unsigned int &p,bool DelComment,
                       bool ReadOps,bool ResolveOps)
{
  unsigned char Num,Num2;
  unsigned int p1,p2;
  _TCHAR Typ;

  p1 = p;
  Typ = AkWertLesen(AkStr.c_str(),p,Num);
  if (Typ != 0)
  { // Wenn Ausdruck gefunden
    unsigned int akp = p;
    LeerZNKomment(AkStr.c_str(),p); // Leerzeichen/Kommentare ueberlesen
    if ((p > akp) && DelComment)
    {
      AkStr.DelNChar(akp,p-akp);
      p = akp;
    }
    Num2 = Num;
    // Label mit Parameterliste
    while ((Num2 < InvalidFkt) && (FuDefArr[Num2].AnzParam > 0) ||
           (Typ == LabelMarker) && (AkStr[p] == '('))
    {
      // wenn 1. Ausdruck Funktion, evtl. weitere Funktionsaufrufe
      // oder Parameter lesen (sin cos 5)
      AkWertLesen(AkStr.c_str(),p,Num2);
      unsigned int akp = p;
      LeerZNKomment(AkStr.c_str(),p); // Leerzeichen/Kommentare ueberlesen
      if ((p > akp) && DelComment)
      {
        AkStr.DelNChar(akp,p-akp);
        p = akp;
      }
    }
    p2 = p;
    if (ReadOps && Operator(AkStr,p,DelComment,ResolveOps))
    {
      if (AkStr[p2] == '|')
      {
        if (!ParallelAufloesen(AkStr,p1,p2,p))
          return 0;
        else
          return _TCHAR('L');
      }
      else
      if (AkStr[p2] == '^')
      {
        if (!PotenzAufloesen(AkStr,p1,p2,p,DelComment,ResolveOps))
          return PotMarker;
        else
          return _TCHAR('P');
      }
      else
        return PotMarker;
    }
  }
  return Typ;
}

bool ParserObj::NextAkWert(const _TCHAR *AkStr,unsigned int &p)
{ // liest den naechsten Ausdruck ohne dessen Typ auszuwerten
  unsigned char Num;
  return (AkWertLesen(AkStr,p,Num) != 0);
}

// liest den naechsten Ausdruck aus AkStr ab Position p, gibt in Num den
// Funktionsindex zurueck wenn Funktion
_TCHAR ParserObj::AkWertLesen(const _TCHAR *AkStr,unsigned int &p,unsigned char &Num)
{
  int KlZaehler;

  Num = InvalidFkt;
  if ((AkStr[p] == '+') || (AkStr[p] == '-'))
    p++; // Vorzeichen ueberlesen
  if ((IST_ZAHL(AkStr[p])) ||
      !_tcsncmp(&AkStr[p],GetLDV().DecSeparator.c_str(),
               GetLDV().DecSeparator.Length()))
  { // wenn Zahl oder Dezimal-Trennzeichen
    int PrsFehler = ZahlFormatObj::NoError;
    Zahl(AkStr,p,PrsFehler);
    if (PrsFehler != ZahlFormatObj::NoError)
      return 0;
    else
      return ZahlMarker;
  }
  else
  switch (AkStr[p])
  {
    case '[' : // Bezeichner
    case '\'' : // Namen eines anderen Graphen
      if (AkStr[p] == '[')
        do
          p++;
        while (AkStr[p] && (AkStr[p] != ']'));
      else
        do
          p++;
        while (AkStr[p] && (AkStr[p] != '\''));
      if (!AkStr[p])
        return 0; // Fehler wenn ] nicht gefunden
      p++; // ] ueberlesen
      return LabelMarker;
    case '(' : // Klammerausdruck oder Parameterliste der Funktion
      KlZaehler = 1;
      p++;
      do // Ende dieses Klammerausdruckes ermitteln
      {
        if (AkStr[p] == '(')
          KlZaehler++;
        else
        if (AkStr[p] == ')')
          KlZaehler--;
        p++;
      }
      while (AkStr[p] && (KlZaehler != 0));
      // Fehler wenn am Ende des Strings und Klammerebenen nicht geschlossen
      if (!AkStr[p] && (KlZaehler != 0))
        return 0;
      else
        return _TCHAR('K');
    default:
      Num = FuNummer(p,AkStr);
      if (Num == InvalidFkt)
        return 0;
      return FuMarker;
  }
}

// Funktionen zum Berechnen einer weiteren Funktion neu initialisieren
void ParserObj::ResetFunktions()
{
  int AltFunk = Funktion->AkEintNummer();
  for (int Ak = 0;Ak <= Funktion->GetMaxEintrag();Ak++)
  {
    Funktion->SetAkEint(Ak);
    // Fr jede Funktion Startwert fr Div-Gleichungen zuruecksetzen
    (*Funktion)->Status &= ~(CpxObj::DiffErg);
  }
  Funktion->SetAkEint(AltFunk);
}

void ParserObj::InitFunktions(FunctionArrObj *Funk)
{
  Funktion = Funk;
  free(KonstFeld);
  KonstFeld = 0;
  AkZahl = 1; // Startwert fr KonstFeld-Eintrge
  int AltFunk = Funktion->AkEintNummer();
  for (int Ak = 0;Ak <= Funktion->GetMaxEintrag();Ak++)
  {
    Funktion->SetAkEint(Ak);
    // Fr jede Funktion Startwert fr Div-Gleichungen +
    // LaufVariable initialisieren
    (*Funktion)->Status &= ~(CpxObj::Berechnet+
                            CpxObj::LaufVariable+
                            CpxObj::DiffErg);
  }
  Funktion->SetAkEint(AltFunk);
}

bool ParserObj::Operator(MGPStr &AkStr,unsigned int &p,bool DelComment,
                         bool ResolveOps)
{ // prueft ob in AkStr an Position p ein Operator ist
  bool Op = false;
  while ((p < AkStr.Length()) && ((AkStr[p]  == '^') || (AkStr[p] == '|')))
  { // wenn Operator
    while ((p < AkStr.Length()) && (AkStr[p+1] == '|'))
      p++; // mehrfache || fuer Parallel ueberlesen
    Op = true;
    p++;
    AkWert(AkStr,p,DelComment,true,ResolveOps);
  }
  return Op;
}

// ersetzt den Parallel-Ausdruck in AkStr durch die Loesung
// (A || B -> A*B/(A+B))
bool ParserObj::ParallelAufloesen(MGPStr &AkStr,unsigned int PAnf,
                                  unsigned int POp,unsigned int &p)
{
  MGPStr Zeile;
  int len = 0;

  while (AkStr[POp+1+len] == '|') // evtl. mehrfache | entfernen
    len++;
  if (len > 0)
  {
    AkStr.DelNChar(POp+1,len);
    p -= len;
  }
  Zeile.AddNChar(AkStr,PAnf);
  Zeile.AddStr(_TEXT("("));
  Zeile.AddNChar(&AkStr[PAnf],POp-PAnf);
  Zeile.AddStr(_TEXT("*"));
  Zeile.AddNChar(&AkStr[POp+1],p-POp-1);
  Zeile.AddStr(_TEXT(")\\("));
  Zeile.AddNChar(&AkStr[PAnf],POp-PAnf);
  Zeile.AddStr(_TEXT("+"));
  Zeile.AddNChar(&AkStr[POp+1],p-POp-1);
  Zeile.AddStr(_TEXT(")"));
  Zeile.AddStr(&AkStr[p]);
  AkStr.SetStr(Zeile);
  return true;
}

// ersetzt den Potenz-Ausdruck in AkStr durch die Loesung
// (A ^ B -> A*A*...*A (B As))
// fasst mehrere Potenzen zusammen A^B^C -> A^(B*C)
// loest Potenzen fuer Exponent <= 3 auf A^3 -> A*A*A
// entfernt Klammern um Basis wenn Potenz (A^B)^C -> A^B^C -> A^(B*C)
bool ParserObj::PotenzAufloesen(MGPStr &AkStr,unsigned int PAnf,
                                       unsigned int POp,unsigned int &p,
                                       bool DelComment,bool ResolveOps)
{
  MGStrListObj BasisList;
  bool MehrereBasis = false;
  if ((AkStr[PAnf] == '(') && (AkStr[POp-1] == ')'))
  { // Wenn Basis eingeklammert aber Potenzausdruck, Klammern entfernen
    MGPStr tmp;
    tmp.AddNChar(&AkStr[PAnf+1],POp-PAnf-2);
    unsigned int p3 = 0;
    do // alle Basis-Faktoren lesen
    {
      unsigned int orgp = p3;
      if (p3)
        p3++;
      AkWert(tmp,p3,DelComment,true,ResolveOps);
      BasisList.NeuAkEint();
      BasisList.GetAkStr().AddNChar(&tmp[orgp],p3-orgp);
    }
    while ((p3 < tmp.Length()) &&
           ((tmp[p3] == '*') || (tmp[p3] == '/') || (tmp[p3] == '\\')));
    MehrereBasis = (p3 == tmp.Length()); // wenn mehrere Summen, abbrechen
  }
  if (!MehrereBasis)
  { // wenn nicht mehrere Summanten in Klammer, ganze Basis in 1 Eintrag
    BasisList.ResetKette();
    BasisList.NeuAkEint();
    BasisList.GetAkStr().AddNChar(&AkStr[PAnf],POp-PAnf);
  }
  MGPStr PotStr;
  int pc = 0;
  do // Alle Potenzen lesen
  {
    pc++;
    unsigned int p1 = POp+1;
    if (PotStr.Length() > 0)
      PotStr.AddStr(_TEXT("*"));
    AkWert(AkStr,p1,DelComment,false);
    PotStr.AddNChar(&AkStr[POp+1],p1-(POp+1));
    POp = p1;
  }
  while (AkStr[POp] == '^');
  MGPStr AuflStr;
  AuflStr.AddNChar(AkStr,PAnf);
  BasisList.Begin();
  if (BasisList.GetMaxEintrag() > 0)
    AuflStr.AddChar('(');
  if (BasisList.GetMaxEintrag() >= 0)
    do // alle Basis-Faktoren bearbeiten
    {
      int anfpos = (BasisList.AkEintNummer() > 0 ? 1 : 0);
      unsigned int p3 = anfpos;
      AkWert(BasisList.GetAkStr(),p3,DelComment,false);
      int nPOp = p3;
      MGPStr AkPotStr;
      int count = 0;
      while (BasisList.GetAkStr()[p3] == '^')
      { // alle Potenzen dieser Basis lesen
        count++;
        p3++;
        int akp = p3;
        AkWert(BasisList.GetAkStr(),p3,DelComment,false);
        if (AkPotStr.Length() > 0)
          AkPotStr.AddChar('*');
        AkPotStr.AddNChar(&BasisList.GetAkStr()[akp],p3-akp);
      }
      if (AkPotStr.Length() > 0)
        AkPotStr.AddChar('*');
      AkPotStr.AddStr(PotStr);
      if (anfpos)
        AuflStr.AddChar(BasisList.GetAkStr()[0]); // evtl. Rechenzeichen
      AuflStr.AddNChar(&BasisList.GetAkStr()[anfpos],nPOp-anfpos); // Basis
      AuflStr.AddChar('^');
      if (pc+count > 1)
        AuflStr.AddChar('(');
      AuflStr.AddStr(AkPotStr); // Exponent hinzufuegen
      if (pc+count > 1)
        AuflStr.AddChar(')');
    }
    while (BasisList.NextEint());
  if (BasisList.GetMaxEintrag() > 0)
    AuflStr.AddChar(')');
  int np = AuflStr.Length();
  AuflStr.AddStr(&AkStr[p]);
  p = np;
  AkStr.SetStr(AuflStr);
  return (BasisList.GetMaxEintrag() > 0);
}

// ----------------------------------------------------------------------------
// ParserBaseObj
// ----------------------------------------------------------------------------


