// Version 6.06  29.11.2009 Markus Greither
// Copyright (c) 1991-2009 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file efxmlio.cpp
    \brief Source code for the xml data storage classes
*/


#include "efprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#ifdef __DMC__
#include <typeinfo.h>
#endif
#include <wx/wx.h>

#include <cstring>
#include <set>
#include <functional>
#if defined(__MINGW32__)
#include <wx/msw/winundef.h>
#endif
#include <cstdlib>
#include <cstdio>
#include <fstream>
#include <ctime>
#include <complex>
#include <limits>
#include <cfloat>
#include <string>
#include <sstream>

using namespace std;

#include "efdefs.h"
#include "mg.h"
#include "cpxpsr.h"
#include "diagtype.h"
#include "efgemein.h"
#include "efmcxio.h"
#include "efresid.h"

#endif

#include "efxmlio.h"
#include "xmlparse.h"
#include "base64.h"

//! Data structure for current node information passed to expat handlers
struct EFXmlParsingContext
{
    //! Memory encoding for strings
    wxMBConv  *conv;
    //! Root node of document
    EFXmlNode *root;
    //! Current node
    EFXmlNode *node;
    //! Last created text node
    EFXmlNode *lastAsText;
    //! Encoding name
    MGPStr     encoding;
    //! Version name
    MGPStr     version;
    //! Data storage for CDATA
    StringArr *isCdata;
    //! EFDaten instance
    EFDaten   *m_efdaten;

    //! Ctor
    EFXmlParsingContext()
    {
        conv = 0;
        root = 0;
        node = 0;
        lastAsText = 0;
        isCdata = 0;
    };
};

//-----------------------------------------------------------------------------
//  EFXmlNode
//-----------------------------------------------------------------------------

EFXmlNode::EFXmlNode(EFXmlNode *parent,EFXmlNodeType type,
                     const _TCHAR *name, const _TCHAR *content,
                     EFXmlProperty *props, EFXmlNode *next)
    : m_type(type), m_name(name), m_content(content), m_cdata(NULL),
      m_properties(props), m_parent(parent),
      m_children(NULL), m_next(next), m_lastchild(NULL)
{
    if (m_parent)
        m_parent->AddChild(this);
}

EFXmlNode::EFXmlNode(EFXmlNodeType type, const _TCHAR *name,
                     const _TCHAR *content)
    : m_type(type), m_name(name), m_content(content), m_cdata(NULL),
      m_properties(NULL), m_parent(NULL),
      m_children(NULL), m_next(NULL), m_lastchild(NULL)
{}

EFXmlNode::~EFXmlNode()
{
    EFXmlNode *c, *c2;
    for (c = m_children; c; c = c2)
    {
        c2 = c->m_next;
        delete c;
    }

    EFXmlProperty *p, *p2;
    for (p = m_properties; p; p = p2)
    {
        p2 = p->GetNext();
        delete p;
    }
    delete m_cdata;
}

bool EFXmlNode::HasProp(const _TCHAR *propName) const
{
    EFXmlProperty *prop = GetProperties();

    while (prop)
    {
        if (!_tcscmp(prop->GetName().c_str(),propName)) 
            return TRUE;
        prop = prop->GetNext();
    }

    return FALSE;
}

bool EFXmlNode::GetPropVal(const _TCHAR *propName, MGPStr &value) const
{
    EFXmlProperty *prop = GetProperties();

    while (prop)
    {
        if (!_tcscmp(prop->GetName().c_str(),propName))
        {
            value = prop->GetValue();
            return TRUE;
        }
        prop = prop->GetNext();
    }

    return FALSE;
}

MGPStr EFXmlNode::GetPropVal(const _TCHAR *propName,
                             const _TCHAR *defaultVal) const
{
    MGPStr tmp;
    if (GetPropVal(propName, tmp))
        return tmp;
    else
        return defaultVal;
}

void EFXmlNode::AddChild(EFXmlNode *child)
{
    if (m_children == NULL)
    {
        m_children = child;
        m_lastchild = child;
    }
    else
    {
        m_lastchild->m_next = child;
        m_lastchild = child;
    }
    child->m_next = NULL;
    child->m_parent = this;
}

void EFXmlNode::AddProperty(const _TCHAR *name, const _TCHAR *value)
{
    AddProperty(new EFXmlProperty(name, value, NULL));
}

void EFXmlNode::AddProperty(EFXmlProperty *prop)
{
    if (m_properties == NULL)
        m_properties = prop;
    else
    {
        EFXmlProperty *p = m_properties;
        while (p->GetNext()) 
            p = p->GetNext();
        p->SetNext(prop);
    }
}

bool EFXmlNode::DeleteProperty(const _TCHAR *name)
{
    EFXmlProperty *prop;

    if (m_properties == NULL)
        return FALSE;

    else 
    if (!_tcscmp(m_properties->GetName().c_str(),name))
    {
        prop = m_properties;
        m_properties = prop->GetNext();
        prop->SetNext(NULL);
        delete prop;
        return TRUE;
    }

    else
    {
        EFXmlProperty *p = m_properties;
        while (p->GetNext())
        {
            if (!_tcscmp(p->GetNext()->GetName().c_str(),name))
            {
                prop = p->GetNext();
                p->SetNext(prop->GetNext());
                prop->SetNext(NULL);
                delete prop;
                return TRUE;
            }
            p = p->GetNext();
        }
        return FALSE;
    }
}

//-----------------------------------------------------------------------------
//  EFXmlDocument
//-----------------------------------------------------------------------------

EFXmlDocument::EFXmlDocument(EFDaten *data)
    : m_version(_TEXT("1.0")), m_fileEncoding(_TEXT("utf-8")), m_root(NULL),
      m_efdaten(data)
{
#if !wxUSE_UNICODE
    m_encoding = _TEXT("utf-8");
#endif
}

//-----------------------------------------------------------------------------
//  EFXmlDocument loading routines
//-----------------------------------------------------------------------------

//! converts Expat-produced string in UTF-8 into wxString
wxString EFXmlDocument::CharToString(wxMBConv *conv,
                                     const char *s, size_t len)
{
#if wxUSE_UNICODE
    (void)conv;
    return wxString(s,wxConvUTF8, len);
#else
    if ( conv )
    {
        size_t nLen = (len != wxSTRING_MAXLEN) ? len :
                          nLen = wxConvUTF8.MB2WC((wchar_t*) NULL, s, 0);

        wchar_t *buf = new wchar_t[nLen+1];
        wxConvUTF8.MB2WC(buf, s, nLen);
        buf[nLen] = 0;
        // Move control chars from 0x7F... to 0x00...
        // FIXME: should be in ConvertForEF(), but needs to be done _before_
        // conversion into multibyte string for non-unicode version
        // Only used in ANSI mode
        for (size_t i = 0;i < nLen;i++)
          if ((buf[i] > 0x7F) && (buf[i] <= 0x9F))
            buf[i] -= 0x7F;

        wxString tmp(buf, *conv, len);
        delete[] buf;
        return tmp;
    }
    else
        return wxString(s, len);
#endif
}

//! Handler for new XmlNode
static void StartElementHnd(void *userData, const char *name, const char **atts)
{
    EFXmlParsingContext *ctx = (EFXmlParsingContext*)userData;
    EFXmlNode *node = new EFXmlNode(EFXmlNode::ELEMENT_NODE,
                                    EFXmlDocument::CharToString(ctx->conv, name).c_str());
    const char **a = atts;
    while (*a)
    {
        node->AddProperty(EFXmlDocument::CharToString(ctx->conv, a[0]).c_str(),
                          EFXmlDocument::CharToString(ctx->conv, a[1]).c_str());
        a += 2;
    }
    if (ctx->root == NULL)
        ctx->root = node;
    else
        ctx->node->AddChild(node);
    ctx->node = node;
    ctx->lastAsText = NULL;
}

//! End of new node
static void EndElementHnd(void *userData, const char* WXUNUSED(name))
{
    EFXmlParsingContext *ctx = (EFXmlParsingContext*)userData;
    ctx->node = ctx->node->GetParent();
    ctx->lastAsText = NULL;
}

//! Handler for new CDATA item
static void StartCdataHnd(void *userData)
{
    EFXmlParsingContext *ctx = (EFXmlParsingContext*)userData;
    ctx->isCdata = new StringArr;
    ctx->lastAsText = NULL;
}

//! End of CDATA item
static void EndCdataHnd(void *userData)
{
    EFXmlParsingContext *ctx = (EFXmlParsingContext*)userData;
    EFXmlNode *tmp = new EFXmlNode(EFXmlNode::CDATA_SECTION_NODE,_TEXT("CDATA"));
    tmp->SetCDATA(ctx->isCdata);

    if (ctx->m_efdaten)
    {
      // Write CDATA directly to EFDaten, using a temporary EFXmlDocument
      EFXmlNode *parent = ctx->node;
      while (parent)
      {
        EFXmlNode *t = new EFXmlNode(NULL,parent->GetType(),
                                     parent->GetName().c_str(),
                                     parent->GetContent().c_str(),
                                     NULL,NULL);
        t->AddChild(tmp);
        tmp = t;
        parent = parent->GetParent();
      }
      EFXmlDocument doc;
      doc.SetRoot(tmp);
  
      ctx->m_efdaten->ReadFromXmlDoc(doc);
    }
    else // for macros, add cdata to tree
      ctx->node->AddChild(tmp);

    ctx->isCdata = 0;
    ctx->lastAsText = NULL;
}

//! Handler for text item
static void TextHnd(void *userData, const char *s, int len)
{
    EFXmlParsingContext *ctx = (EFXmlParsingContext*)userData;
    char *buf = 0;
    if (!ctx->isCdata)
    {
        buf = new char[len + 1];

        buf[len] = '\0';
        memcpy(buf, s, (size_t)len);
    }

    if (ctx->lastAsText)
    {
        MGPStr tmp(ctx->lastAsText->GetContent());
        tmp.AddStr(EFXmlDocument::CharToString(ctx->conv, buf).c_str());
        ctx->lastAsText->SetContent(tmp);
    }
    else
    {
        if (ctx->isCdata)
        {
            ctx->isCdata->NeuAkEint();
            ctx->isCdata->GetAkEint().append(s,len);
        }
        else
        {
            bool whiteOnly = TRUE;
            for (char *c = buf; *c != '\0'; c++)
                if (*c != ' ' && *c != '\t' && *c != '\n' && *c != '\r')
                {
                    whiteOnly = FALSE;
                    break;
                }
            if (!whiteOnly)
            {
                ctx->lastAsText = new EFXmlNode(EFXmlNode::TEXT_NODE,
                                                _TEXT("text"),
                                                EFXmlDocument::CharToString(ctx->conv, buf));
                ctx->node->AddChild(ctx->lastAsText);
            }
        }
    }

    delete[] buf;
}

//! Handler for comment item
static void CommentHnd(void *userData, const char *data)
{
    EFXmlParsingContext *ctx = (EFXmlParsingContext*)userData;

    if (ctx->node)
    {
        ctx->node->AddChild(new EFXmlNode(EFXmlNode::COMMENT_NODE,
                                          _TEXT("comment"),
                                          EFXmlDocument::CharToString(ctx->conv, data)));
    }
    ctx->lastAsText = NULL;
}

//! Default Handler
static void DefaultHnd(void *userData, const char *s, int len)
{
    // XML header:
    if (len > 6 && memcmp(s, "<?xml ", 6) == 0)
    {
        EFXmlParsingContext *ctx = (EFXmlParsingContext*)userData;

        wxString buf = EFXmlDocument::CharToString(ctx->conv, s, (size_t)len);
        int pos;
        pos = buf.Find(_TEXT("encoding="));
        if (pos != wxNOT_FOUND)
            ctx->encoding = buf.Mid(pos + 10).BeforeFirst(buf[(size_t)pos+9]).EF_C_STR();
        pos = buf.Find(_TEXT("version="));
        if (pos != wxNOT_FOUND)
            ctx->version = buf.Mid(pos + 9).BeforeFirst(buf[(size_t)pos+8]).EF_C_STR();
    }
}

//! Handler for unknown enconding
static int UnknownEncodingHnd(void * WXUNUSED(encodingHandlerData),
                              const XML_Char *name, XML_Encoding *info)
{
    // We must build conversion table for expat. The easiest way to do so
    // is to let wxCSConv convert as string containing all characters to
    // wide character representation:
    wxCSConv conv(wxString(name, wxConvLibc));
    char mbBuf[2];
    wchar_t wcBuf[10];
    size_t i;

    mbBuf[1] = 0;
    info->map[0] = 0;
    for (i = 0; i < 255; i++)
    {
        mbBuf[0] = (char)(i+1);
        if (conv.MB2WC(wcBuf, mbBuf, 2) == (size_t)-1)
        {
            // invalid/undefined byte in the encoding:
            info->map[i+1] = -1;
        }
        info->map[i+1] = (int)wcBuf[0];
    }

    info->data = NULL;
    info->convert = NULL;
    info->release = NULL;

    return 1;
}

bool EFXmlDocument::Load(istream& stream, const MGPStr& encoding)
{
#if wxUSE_UNICODE
    (void)encoding;
#else
    m_encoding = encoding;
#endif

    const size_t BUFSIZE = 0xFFFF;
    char *buf = new char[BUFSIZE];
    EFXmlParsingContext ctx;
    bool done;
    XML_Parser parser = XML_ParserCreate(NULL);

    ctx.encoding = _TEXT("utf-8"); // default in absence of encoding=""
    ctx.m_efdaten = m_efdaten;
    
#if !wxUSE_UNICODE
    ctx.conv = new wxCSConv(encoding.c_str());
#endif

    XML_SetUserData(parser, (void*)&ctx);
    XML_SetElementHandler(parser, StartElementHnd, EndElementHnd);
    XML_SetCdataSectionHandler(parser, StartCdataHnd, EndCdataHnd);
    XML_SetCharacterDataHandler(parser, TextHnd);
    XML_SetCommentHandler(parser, CommentHnd);
    XML_SetDefaultHandler(parser, DefaultHnd);
    XML_SetUnknownEncodingHandler(parser, UnknownEncodingHnd, NULL);

    do
    {
        stream.read(buf, BUFSIZE);
        size_t len = stream.gcount();
        done = (len < BUFSIZE);
        if (!XML_Parse(parser, buf, len, done))
        {
            wxLogError(_("XML parsing error: '%s' at line %d"),
                       XML_ErrorString(XML_GetErrorCode(parser)),
                       XML_GetCurrentLineNumber(parser));
            return FALSE;
        }
    }
    while (!done);
        
    delete [] buf;

    SetVersion(ctx.version);
    SetFileEncoding(ctx.encoding);
    SetRoot(ctx.root);

    XML_ParserFree(parser);
#if !wxUSE_UNICODE
    if ( ctx.conv )
        delete ctx.conv;
#endif
    return TRUE;
}

//-----------------------------------------------------------------------------
//  EFXmlDocument saving routines
//-----------------------------------------------------------------------------

//! Write string to output
void EFXmlDocument::OutputString(ostream& stream, const wxString& str,
                                 wxMBConv *convMem, wxMBConv *convFile)
{
    if (str.IsEmpty())
        return;
#if wxUSE_UNICODE
    (void)convMem;
    const wxWX2MBbuf buf(str.mb_str(*(convFile ? convFile : &wxConvUTF8)));
    stream.write((const char*)buf, strlen((const char*)buf));
#else
    if ( convFile == NULL )
        stream.write(str.c_str(), str.Len());
    else
    {
        wxString str2(str.wc_str(*convMem), *convFile);
        stream.write(str2.mb_str(), str2.Len());
    }
#endif
}

//! Same as above, but create entities first.
/*! Translates '<' to "&lt;", '>' to "&gt;" and '&' to "&amp;"
*/
void EFXmlDocument::OutputStringEnt(ostream& stream, const wxString& str,
                                    wxMBConv *convMem, wxMBConv *convFile)
{
    wxString buf;
    size_t i, last, len;
    wxChar c;

    len = str.Len();
    last = 0;
    for (i = 0; i < len; i++)
    {
        c = str.GetChar(i);
        if (c == _TEXT('<') || c == _TEXT('>') ||
            (c == _TEXT('&') &&
             (str.Mid(i+1, 4) != _TEXT("amp;")) && (str[i+1] != '#')))
        {
            OutputString(stream, str.Mid(last, i - last),
                                        convMem, convFile);
            switch (c)
            {
                case _TEXT('<'):
                    OutputString(stream, _TEXT("&lt;"), NULL, NULL);
                    break;
                case _TEXT('>'):
                    OutputString(stream, _TEXT("&gt;"), NULL, NULL);
                    break;
                case _TEXT('&'):
                    OutputString(stream, _TEXT("&amp;"), NULL, NULL);
                    break;
                default: break;
            }
            last = i + 1;
        }
    }
    OutputString(stream, str.Mid(last, i - last), convMem, convFile);
}

//! Indents the output file
void EFXmlDocument::OutputIndentation(ostream& stream, int indent)
{
    wxString str = _TEXT("\n");
    for (int i = 0; i < indent; i++)
        str << _TEXT(' ') << _TEXT(' ');
    OutputString(stream, str, NULL, NULL);
}

//! Write node to file
void EFXmlDocument::OutputNode(ostream& stream, EFXmlNode *node, int indent,
                               wxMBConv *convMem, wxMBConv *convFile)
{
    EFXmlNode *n, *prev;
    EFXmlProperty *prop;

    switch (node->GetType())
    {
        case EFXmlNode::TEXT_NODE:
            OutputStringEnt(stream, node->GetContent().c_str(),
                            convMem, convFile);
            break;

        case EFXmlNode::CDATA_SECTION_NODE:
        {
            OutputString(stream, _TEXT("<![CDATA["), NULL, NULL);
            StringArr *tmp = node->GetCDATA();
            stream.write(tmp->GetAkEint().c_str(),tmp->GetAkEint().length());
            delete tmp;
            OutputString(stream, _TEXT("]]>"), NULL, NULL);
            break;
        }

        case EFXmlNode::ELEMENT_NODE:
            OutputString(stream, _TEXT("<"), NULL, NULL);
            OutputString(stream, node->GetName().c_str(), NULL, NULL);

            prop = node->GetProperties();
            while (prop)
            {
                OutputString(stream, _TEXT(" ") +
                                     wxString(prop->GetName().c_str()) +
                                     _TEXT("=\"") +
                                     wxString(prop->GetValue().c_str()) +
                                     _TEXT("\""),
                             NULL, NULL);
                // FIXME - what if prop contains '"'?
                prop = prop->GetNext();
            }

            if (node->GetChildren())
            {
                OutputString(stream, _TEXT(">"), NULL, NULL);
                prev = NULL;
                n = node->GetChildren();
                while (n)
                {
                    if (n && n->GetType() != EFXmlNode::TEXT_NODE)
                        OutputIndentation(stream, indent + 1);
                    OutputNode(stream, n, indent + 1, convMem, convFile);
                    prev = n;
                    n = n->GetNext();
                }
                if (prev && prev->GetType() != EFXmlNode::TEXT_NODE)
                    OutputIndentation(stream, indent);
                OutputString(stream, _TEXT("</"), NULL, NULL);
                OutputString(stream, node->GetName().c_str(), NULL, NULL);
                OutputString(stream, _TEXT(">"), NULL, NULL);
            }
            else
                OutputString(stream, _TEXT("/>"), NULL, NULL);
            break;

        default:
            wxFAIL_MSG(_TEXT("unsupported node type"));
    }
}

bool EFXmlDocument::Save(ostream& stream) const
{
    if ( !IsOk() )
        return FALSE;

    wxString s;

    wxMBConv *convMem = NULL, *convFile;
#if wxUSE_UNICODE
    convFile = new wxCSConv(GetFileEncoding().c_str());
#else
    convFile = NULL;
    if ( _tcscmp(GetFileEncoding().c_str(),GetEncoding().c_str()) )
    {
        convFile = new wxCSConv(GetFileEncoding().c_str());
        convMem = new wxCSConv(GetEncoding().c_str());
    }
#endif

    s.Printf(_TEXT("<?xml version=\"%s\" encoding=\"%s\"?>\n"),
             GetVersion().c_str(), GetFileEncoding().c_str());
    OutputString(stream, s, NULL, NULL);

    OutputNode(stream, GetRoot(), 0, convMem, convFile);
    OutputString(stream, _TEXT("\n"), NULL, NULL);

    if ( convFile )
        delete convFile;
    if ( convMem )
        delete convMem;

    return TRUE;
}

// ----------------------------------------------------------------------------
// EFNodeCreator
// ----------------------------------------------------------------------------

//! Class for EFXmlNode creation functions
class EFNodeCreator
{
 public:
  //! Build CDATA item node
  static void CDATAItemOut(const MGPStr &name,StringArr *value,EFXmlNode *Parent)
  {
    EFXmlNode *snode = new EFXmlNode(Parent,EFXmlNode::ELEMENT_NODE,
                                     name.c_str(),_TEXT(""),0,0);
    EFXmlNode *tmp = new EFXmlNode(snode,
                                   EFXmlNode::CDATA_SECTION_NODE,
                                   _TEXT(""),_TEXT(""),0,0);
    tmp->SetCDATA(value);
  }

#if wxUSE_UNICODE
  //! Create text node for non-unicode strings
  static void ItemOut(const MGPStr &name,const char *value,
                      EFXmlNode *Parent,
                      EFXmlNode::EFXmlNodeType type =
                      EFXmlNode::TEXT_NODE)
  {
    wxString tmp = EFXmlDocument::CharToString(0,value,strlen(value));
    EFXmlNode *snode = new EFXmlNode(Parent,EFXmlNode::ELEMENT_NODE,
                                     name.c_str(),_TEXT(""),0,0);
    new EFXmlNode(snode,type,_TEXT(""),tmp.c_str(),0,0);
  }
#endif

#ifndef __DMC__
  //! Create node for item
  template<class T> static void ItemOut(const MGPStr &name,const T value,
                                        EFXmlNode *Parent,
                                        EFXmlNode::EFXmlNodeType type =
                                          EFXmlNode::TEXT_NODE)
  {
    MGPStr tmp;
    tmp << value;
    EFXmlNode *snode = new EFXmlNode(Parent,EFXmlNode::ELEMENT_NODE,
                                     name.c_str(),_TEXT(""),0,0);
    new EFXmlNode(snode,type,_TEXT(""),tmp.c_str(),0,0);
  }
#endif

  // Create node for boolean item
  static void ItemOut(const MGPStr &name,const bool value,EFXmlNode *Parent,
                      EFXmlNode::EFXmlNodeType type = EFXmlNode::TEXT_NODE)
  {
    MGPStr tmp;
    tmp << (int)value;
    EFXmlNode *snode = new EFXmlNode(Parent,EFXmlNode::ELEMENT_NODE,
                                     name.c_str(),_TEXT(""),0,0);
    new EFXmlNode(snode,type,_TEXT(""),tmp.c_str(),0,0);
  }

  //! Create node for FontType item
  static void ItemOut(const MGPStr &name,const FontType value,EFXmlNode *Parent,
                      EFXmlNode::EFXmlNodeType type = EFXmlNode::TEXT_NODE)
  {
    EFXmlNode *snode = new EFXmlNode(Parent,EFXmlNode::ELEMENT_NODE,
                                     name.c_str(),_TEXT(""),0,0);
    ItemOut(_TEXT("Font"),value.Font.c_str(),snode,type);
    ItemOut(_TEXT("Size"),value.Size,snode,type);
    ItemOut(_TEXT("Style"),value.Style,snode,type);
    ItemOut(_TEXT("Color"),value.Color,snode,type);
  }

#ifdef __DMC__
  //! Create node for int item
  static void ItemOut(const MGPStr &name,int value,EFXmlNode *Parent,
                      EFXmlNode::EFXmlNodeType type = EFXmlNode::TEXT_NODE)
  {
    MGPStr tmp;
    tmp << value;
    ItemOut(name,tmp.c_str(),Parent,type);
  }

  //! Create node for unsigned long item
  static void ItemOut(const MGPStr &name,unsigned long value,EFXmlNode *Parent,
                      EFXmlNode::EFXmlNodeType type = EFXmlNode::TEXT_NODE)
  {
    MGPStr tmp;
    tmp << value;
    ItemOut(name,tmp.c_str(),Parent,type);
  }

  //! Create node for unsigned int item
  static void ItemOut(const MGPStr &name,unsigned int value,EFXmlNode *Parent,
                      EFXmlNode::EFXmlNodeType type = EFXmlNode::TEXT_NODE)
  {
    MGPStr tmp;
    tmp << value;
    ItemOut(name,tmp.c_str(),Parent,type);
  }
#endif

  //! Create node for double item
  static void ItemOut(const MGPStr &name,double value,EFXmlNode *Parent,
                      EFXmlNode::EFXmlNodeType type = EFXmlNode::TEXT_NODE)
  {
    ZahlFormatObj ZFO;
    MGPStr tmp;
    ZFO.Ausgabe(tmp,value,ZahlFormatObj::IngForm);
    ItemOut(name,tmp.c_str(),Parent,type);
  }

  //! Differential ItemOut
  template<class T> static void DiffItemOut(const wxChar *txt,const T &data,EFXmlNode *node,const T *orig)
  {
    if (!orig || (*orig != data))
      ItemOut(txt,data,node);
  }

  //! Differential ItemOut for text
  static void DiffItemOut(const wxChar *txt,const wxChar *data,EFXmlNode *node,const wxChar *orig)
  {
    if (!orig || wxStrcmp(orig,data))
      ItemOut(txt,data,node);
  }

  //! Create node for array item
  template<class T> static void ArrayOut(const MGPStr &name,const T &value,
                                         int index,EFXmlNode *Parent,
                                         EFXmlNode::EFXmlNodeType type =
                                           EFXmlNode::TEXT_NODE)
  {
    MGPStr tmp(name);
    tmp << index;
    ItemOut(tmp.c_str(),value,Parent,type);
  }
};

// ----------------------------------------------------------------------------
// EFNodeReader
// ----------------------------------------------------------------------------

//! Class for reading information from EFXmlNode
class EFNodeReader
{
 public:
  //! Get CDATA from node
  static StringArr *GetCDATA(EFXmlNode *node)
  {
    EFXmlNode *txt = node->GetChildren();
    StringArr *tmp;
    if (txt)
      tmp = txt->GetCDATA();
    else
      tmp = node->GetCDATA(); // Dummy value if not found
    return tmp;
  }

  //! Get text value of node
  static const MGPStr &GetValue(EFXmlNode *node)
  {
    EFXmlNode *txt = node->GetChildren();
    if (txt)
      return txt->GetContent();
    else
      return node->GetContent(); // Dummy value if not found
  }

  //! Read font from current node
  static void ReadFont(FontType &value,EFXmlNode *Parent);
};

void EFNodeReader::ReadFont(FontType &value,EFXmlNode *Parent)
{
  EFXmlNode *child = Parent->GetChildren();
  while (child)
  {
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Font")))
      value.Font = EFNodeReader::GetValue(child).c_str();
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Size")))
      value.Size = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Style")))
      value.Style = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Color")))
      value.Color = _ttoi(EFNodeReader::GetValue(child).c_str());
    child = child->GetNext();
  }
}

// ----------------------------------------------------------------------------
// EFXmlDataConverter
// ----------------------------------------------------------------------------

//! Class for data conversion from/to Xml files
class EFXmlDataConverter
{
 public:
  //! Get numeric index from end of given str
  static int GetIndex(const MGPStr &str);

  //!Convert string for output into XML
  static void ConvertForXml(MGPStr &str);

  //! Convert string in XML format into standart format
  static void ConvertForEF(MGPStr &str);

  //! Converts binary data into HEX string, allocates memory for result
  static void DataToString(const void *Data,int Size,MGPStr &erg);

  //! Converts Data in HEX-Format into binary, allocates memory for result
  static char *StringToData(void *Data,int Size,const char *erg);

  //! Get total length of all strings in array, converted from Base64
  static size_t GetDataLength(StringArr *tmp);

  //! Decode Base64 data that was loaded in junks
  static void DecodeDataBase64(StringArr *tmp,void *Data);

  //! Recombine junks of data into one string
  static void StringFromStringArr(StringArr *tmp,string &str);
};

int EFXmlDataConverter::GetIndex(const MGPStr &str)
{
  size_t i = str.Length();
  while ((i > 0) && _istdigit(str[i-1]))
    i--;
  int res = _ttoi(&str.c_str()[i]);
  return res;
}

void EFXmlDataConverter::ConvertForXml(MGPStr &str)
{
  for (size_t i = 0;i < str.Length();i++)
  {
#if !wxUSE_UNICODE
    unsigned char tmp = (unsigned char)str[i];
#else
    _TCHAR tmp = str[i];
#endif
    // replace &, so that text like &#23 does not become screwed
    if (tmp == '&')
    {
      str.InsStr(i+1,_TEXT("amp;"));
      i += 4;
    }
    else
    if (tmp < 32) // Control characters are moved in 0x7F... range
    {
      MGPStr tmps(_TEXT("&#"));
      tmps.AddInt(tmp+0x7F);
      tmps.AddStr(_TEXT(";"));
      str.DelNChar(i,1);
      str.InsStr(i,tmps);
      i += tmps.Length()-1;
    }
  }
}

//! Convert string in XML format into standart format
void EFXmlDataConverter::ConvertForEF(MGPStr &str)
{
  for (size_t i = 0;i < str.Length();i++)
  {
#if !wxUSE_UNICODE
    unsigned char tmp = (unsigned char)str[i];
#else
    _TCHAR tmp = str[i];
#endif
    if ((tmp > 0x7F) && (tmp <= 0x9F))
      str.ReplChar(i,tmp-0x7F);
  }
}

//! Converts binary data into HEX string, allocates memory for result
void EFXmlDataConverter::DataToString(const void *Data,int Size,MGPStr &erg)
{
  _TCHAR *d = new _TCHAR[Size*2+1];
  for (int i = 0;i < Size;i++)
  {
    unsigned char D = ((unsigned char *)Data)[i];
    unsigned char High = D >> 4;
    if (High > 9)
      d[i*2] = 'A'+High-10;
    else
      d[i*2] = '0'+High;
    unsigned char Low = D & 0x0F;
    if (Low > 9)
      d[i*2+1] = 'A'+Low-10;
    else
      d[i*2+1] = '0'+Low;
  }
  d[Size*2] = 0;
  erg = d;
  delete[] d;
}

//! Converts Data in HEX-Format into binary, allocates memory for result
char *EFXmlDataConverter::StringToData(void *Data,int Size,const char *erg)
{
  if (!Data)
    Data = new char[Size/2];
  for (int i = 0;i < Size;i++)
  {
    unsigned char D;
    if (erg[i] >= 'A')
      D = (erg[i]-'A'+10) << 4;
    else
      D = (erg[i]-'0') << 4;
    if (erg[i+1] >= 'A')
      D += (erg[i+1]-'A'+10);
    else
      D += (erg[i+1]-'0');
    ((char *)Data)[i/2] = D;
    i++;
  }
  return (char *)Data;
}

//! Get total length of all strings in array, converted from Base64
size_t EFXmlDataConverter::GetDataLength(StringArr *tmp)
{
  size_t len = 0;
  tmp->Begin();
  if (tmp->GetMaxEintrag() >= 0)
    do
      len += tmp->GetAkEint().length();
    while (tmp->NextEint());
  len = len*3/4;
  return len;
}

//! Decode Base64 data that was loaded in junks
void EFXmlDataConverter::DecodeDataBase64(StringArr *tmp,void *Data)
{
  size_t anfpos = 0;
  size_t akpt = 0;
  string result;
  tmp->Begin();
  if (tmp->GetMaxEintrag() >= 0)
    do
    {
      size_t rest = (tmp->GetAkEint().length()-anfpos) % 4;
      size_t missing = (rest ? 4-rest : 0);
      string *akstr = &tmp->GetAkEint();
      if (missing && tmp->NextEint())
        akstr->append(tmp->GetAkEint(),0,missing);
      result = "";
      Base64::decode(&akstr->c_str()[anfpos],akstr->length()-anfpos,result);
      anfpos = missing;
      memcpy((void *)&((char *)Data)[akpt],result.c_str(),result.length());
      akpt += result.length();
      tmp->Begin();
      tmp->LoeschAkEint();
      tmp->Begin();
    }
    while (tmp->GetMaxEintrag() >= 0);
}

//! Recombine junks of data into one string
void EFXmlDataConverter::StringFromStringArr(StringArr *tmp,string &str)
{
  tmp->Begin();
  if (tmp->GetMaxEintrag() >= 0)
    do
      str.append(tmp->GetAkEint());
    while (tmp->NextEint());
}

// ----------------------------------------------------------------------------
// EFDaten
// ----------------------------------------------------------------------------

bool EFDaten::XmlDateiSpeichern(ostream &stream,DiagDataPtArrObj &DiagDataPts)
{
  EFXmlDocument doc;
#if !wxUSE_UNICODE
  doc.SetEncoding(wxLocale::GetSystemEncodingName().c_str());
#endif
  EFXmlNode *node;
  time_t timer = time(NULL);
  struct tm *TS = localtime(&timer);
  wxString s;
  s.Printf(_TEXT("%02hi.%02hi.%4i %02hi:%02hi:%02hi"),
           TS->tm_mday,TS->tm_mon+1,TS->tm_year+1900,TS->tm_hour,
           TS->tm_min,TS->tm_sec);
  node = new EFXmlNode(0,EFXmlNode::ELEMENT_NODE,
                       _TEXT("Easy_Funktion_Document"),
                       _TEXT(""),
                       new EFXmlProperty(_TEXT("version"),_TEXT(VERSIONSTRING),
                       new EFXmlProperty(_TEXT("date"),s.c_str(),0)),0);
  doc.SetRoot(node);
  EFXmlNode *chnode;
  s = _TEXT("");
  s << Funktion->GetMaxEintrag();
  chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                         _TEXT("Functions"),_TEXT(""),
                         new EFXmlProperty(_TEXT("count"),s.c_str(),0),0);
  int AltFunk = Funktion->AkEintNummer();
  for (int f = 1;f <= Funktion->GetMaxEintrag();f++)
  { // write all functions
    Funktion->SetAkEint(f);
    MGPStr tmp = (*Funktion)->Namen.c_str();
    tmp.AddStr((*Funktion)->ParamStr.c_str());
    if ((tmp.Length() > 0) && ((*Funktion)->Term.Length() > 0))
      tmp.AddStr(_TEXT(" = "));
    tmp.AddStr((*Funktion)->Term.c_str());
    EFXmlDataConverter::ConvertForXml(tmp);
    EFNodeCreator::ArrayOut(_TEXT("Function"),tmp.c_str(),f,chnode);
  }
  Funktion->SetAkEint(AltFunk);
  chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                         _TEXT("ResultHistory"),_TEXT(""),0,0);
  for (int f = 0;f < ErgebnisHistory.GetCount();f++)
    EFNodeCreator::ArrayOut(_TEXT("Item"),ErgebnisHistory.GetString(f),f,chnode);
  chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                         _TEXT("KeepHistory"),_TEXT(""),0,0);
  for (int f = 0;f < BeibehaltenHistory.GetCount();f++)
    EFNodeCreator::ArrayOut(_TEXT("Item"),BeibehaltenHistory.GetString(f),f,chnode);
  chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                         _TEXT("KeeptogetherHistory"),_TEXT(""),0,0);
  for (int f = 0;f < ZusammenHistory.GetCount();f++)
    EFNodeCreator::ArrayOut(_TEXT("Item"),ZusammenHistory.GetString(f),f,chnode);
  chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                         _TEXT("SinglevalueHistory"),_TEXT(""),0,0);
  for (int f = 0;f < EinzelwertHistory.GetCount();f++)
    EFNodeCreator::ArrayOut(_TEXT("Item"),EinzelwertHistory.GetString(f),f,chnode);
  AuflVorlage.Begin();
  chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                         _TEXT("Solve"),_TEXT(""),0,0);
  do
  {
    MGPStr tmp(_TEXT("Item"));
    tmp << AuflVorlage.AkEintNummer();
    EFXmlNode *cn = new EFXmlNode(chnode,EFXmlNode::ELEMENT_NODE,
                                  tmp.c_str(),_TEXT(""),0,0);
    if (AuflVorlage.AkEintNummer() > 0)
      EFNodeCreator::ItemOut(_TEXT("Name"),AuflVorlage->Name.c_str(),cn);
    if (AuflVorlage->ParamsList.GetMaxEintrag() >= 0)
    {
      MGPStr tmp;
      tmp << AuflVorlage->ParamsList.GetMaxEintrag()+1;
      EFXmlNode *n = new EFXmlNode(cn,EFXmlNode::ELEMENT_NODE,
                                   _TEXT("ParamList"),_TEXT(""),
                                   new EFXmlProperty(_TEXT("count"),
                                                     tmp.c_str(),0),0);
      AuflVorlage->ParamsList.Begin();
      do
        EFNodeCreator::ArrayOut(_TEXT("Param"),AuflVorlage->ParamsList.GetAkEint().c_str(),
                 AuflVorlage->ParamsList.AkEintNummer()+1,n);
      while (AuflVorlage->ParamsList.NextEint());
    }
    EFNodeCreator::ItemOut(_TEXT("ErgStr"),AuflVorlage->ErgStr.c_str(),cn);
    EFNodeCreator::ItemOut(_TEXT("KeepStr"),AuflVorlage->BeibehaltenStr.c_str(),cn);
    EFNodeCreator::ItemOut(_TEXT("KeeptogetherStr"),AuflVorlage->ZusammenStr.c_str(),cn);
    EFNodeCreator::ItemOut(_TEXT("Mode"),AuflVorlage->Mode,cn);
    EFNodeCreator::ItemOut(_TEXT("Range"),AuflVorlage->Bereich,cn);
    EFNodeCreator::ItemOut(_TEXT("Append"),AuflVorlage->Anh,cn);
  }
  while (AuflVorlage.NextEint());
  chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                         _TEXT("Editor"),_TEXT(""),0,0);
  EFNodeCreator::ItemOut(_TEXT("FractionDisplay"),(int)BildschOpt.GetBildschBruch(),chnode);
  EFNodeCreator::ItemOut(_TEXT("Brackets"),(int)BildschOpt.GetBildschKlammer(),chnode);
  EFNodeCreator::ItemOut(_TEXT("HideDiags"),(int)BildschOpt.GetHideDiags(),chnode);
  EFNodeCreator::ItemOut(_TEXT("MarkLabels"),(int)BildschOpt.GetMarkLabels(),chnode);
  EFNodeCreator::ItemOut(_TEXT("Zoom"),Zoom,chnode);
  EFNodeCreator::ItemOut(_TEXT("PageLandscape"),SeiteQuer,chnode);
  XmlWriteEditFont(chnode,Editor);
  EFNodeCreator::ItemOut(_TEXT("ScrollLine"),ScrollLine,chnode);
  EFNodeCreator::ItemOut(_TEXT("ScrollPosX"),ScrollPosX,chnode);
  EFNodeCreator::ItemOut(_TEXT("ScrollPosY"),ScrollPosY,chnode);
  EFNodeCreator::ItemOut(_TEXT("SelStart"),EditSelAnf,chnode);
  EFNodeCreator::ItemOut(_TEXT("SelEnd"),EditSelEnd,chnode);
  chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                         _TEXT("ApproximationData"),_TEXT(""),0,0);
  for (int f = 0;f < 3;f++)
  {
    EFNodeCreator::ArrayOut(_TEXT("XStr"),NaehData.XStr[f].c_str(),f,chnode);
    EFNodeCreator::ArrayOut(_TEXT("YStr"),NaehData.YStr[f].c_str(),f,chnode);
  }
  EFNodeCreator::ItemOut(_TEXT("ErgStr"),NaehData.ErgStr.c_str(),chnode);
  EFNodeCreator::ItemOut(_TEXT("ParamStr"),NaehData.ParamStr.c_str(),chnode);
  EFNodeCreator::ItemOut(_TEXT("CurGraph"),(int)NaehData.m_CurGraph,chnode);
  EFNodeCreator::ItemOut(_TEXT("NumPts"),(int)NaehData.m_NumPts,chnode);
  EFNodeCreator::ItemOut(_TEXT("Mode"),NaehData.Mode,chnode);
  DiagDataPts.Begin();
  if (DiagDataPts.GetMaxEintrag() >= 0)
    do
    {
      MGPStr tmp(_TEXT("Diagram"));
      tmp << DiagDataPts.AkEintNummer()+1;
      chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                             tmp.c_str(),_TEXT(""),0,0);
      DiagDataPts.GetAkEint()->XmlSpeichern(chnode);
    }
    while (DiagDataPts.NextEint());
  PictureArr->Begin();
  if (PictureArr->GetMaxEintrag() >= 0)
    do
    {
      if (PictureArr->AkPicture()->Data != 0)
      {
        MGPStr tmp(_TEXT("Picture"));
        tmp << PictureArr->AkEintNummer()+1;
        chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                               tmp.c_str(),_TEXT(""),0,0);
        PictureArr->AkPicture()->XmlSpeichern(chnode);
      }
    }
    while (PictureArr->NextEint());
  EquationsArr->Begin();
  if (EquationsArr->GetMaxEintrag() >= 0)
    do
    {
      MGPStr tmp(_TEXT("Equations"));
      tmp << EquationsArr->AkEintNummer()+1;
      chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                             tmp.c_str(),_TEXT(""),0,0);
      EquationsArr->AkEquations()->XmlSpeichern(chnode);
    }
    while (EquationsArr->NextEint());
  m_submodules.Begin();
  if (m_submodules.GetMaxEintrag() >= 0)
  {
    MGPStr tmp(_TEXT("Submodules"));
    chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                           tmp.c_str(),_TEXT(""),0,0);
    do
    {
      EFNodeCreator::ArrayOut(_TEXT("Module"),m_submodules.GetAkStr().c_str(),
                                              m_submodules.AkEintNummer()+1,chnode);
    }
    while (m_submodules.NextEint());
  }
  m_OpenPanels.Begin();
  if (m_OpenPanels.GetMaxEintrag() >= 0)
  {
    MGPStr tmp(_TEXT("OpenPanels"));
    chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                           tmp.c_str(),_TEXT(""),0,0);
    do
    {
      EFNodeCreator::ArrayOut(_TEXT("Panel"),m_OpenPanels.GetAkEint(),
                                             m_OpenPanels.AkEintNummer()+1,chnode);
    }
    while (m_OpenPanels.NextEint());
  }
  doc.Save(stream);
  return true;
}

void EFDaten::XmlWriteEditFont(EFXmlNode *chnode,const FontType &font)
{
  EFNodeCreator::ItemOut(_TEXT("Font"),font,chnode);
}

void EFDaten::XmlLoadEditFont(EFXmlNode *node,FontType &font)
{
  EFXmlNode *chnode = node->GetChildren();
  EFNodeReader::ReadFont(font,chnode);
}

bool EFDaten::XmlDateiLaden(istream &stream)
{
  EFXmlDocument doc(this);
#if !wxUSE_UNICODE
  doc.Load(stream,wxLocale::GetSystemEncodingName().c_str());
#else
  doc.Load(stream);
#endif
  ReadFromXmlDoc(doc);
  return true;
}

void EFDaten::ReadFromXmlDoc(EFXmlDocument &doc)
{
  EFXmlNode *node = doc.GetRoot();
  if (!node || _tcscmp(node->GetName().c_str(),_TEXT("Easy_Funktion_Document")))
    return;
  EFXmlNode *child = node->GetChildren();
  while (child)
  {
    EFXmlNode *cc = child->GetChildren();
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Functions")))
      while (cc)
      {
        Funktion->CreateAkEint(EFXmlDataConverter::GetIndex(cc->GetName()));
        MGPStr tmp(EFNodeReader::GetValue(cc).c_str());
        EFXmlDataConverter::ConvertForEF(tmp);
        (*Funktion)->Term.SetStr(tmp);
        cc = cc->GetNext();
      }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("ResultHistory")))
      while (cc)
      {
        ErgebnisHistory.SetString(EFXmlDataConverter::GetIndex(cc->GetName()),
                                  EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("KeepHistory")))
      while (cc)
      {
        BeibehaltenHistory.SetString(EFXmlDataConverter::GetIndex(cc->GetName()),
                                     EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("KeeptogetherHistory")))
      while (cc)
      {
        BeibehaltenHistory.SetString(EFXmlDataConverter::GetIndex(cc->GetName()),
                                     EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("SinglevalueHistory")))
      while (cc)
      {
        EinzelwertHistory.SetString(EFXmlDataConverter::GetIndex(cc->GetName()),
                                    EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Solve")))
    {
      while (cc)
      {
        AuflVorlage.CreateAkEint(EFXmlDataConverter::GetIndex(cc->GetName()));
        EFXmlNode *ccc = cc->GetChildren();
        while (ccc)
        {
          if (!_tcscmp(ccc->GetName().c_str(),_TEXT("Name")))
            AuflVorlage->Name = EFNodeReader::GetValue(ccc);
          else
          if (!_tcscmp(ccc->GetName().c_str(),_TEXT("ErgStr")))
            AuflVorlage->ErgStr = EFNodeReader::GetValue(ccc);
          else
          if (!_tcscmp(ccc->GetName().c_str(),_TEXT("KeepStr")))
            AuflVorlage->BeibehaltenStr = EFNodeReader::GetValue(ccc);
          else
          if (!_tcscmp(ccc->GetName().c_str(),_TEXT("KeeptogetherStr")))
            AuflVorlage->ZusammenStr = EFNodeReader::GetValue(ccc);
          else
          if (!_tcscmp(ccc->GetName().c_str(),_TEXT("Mode")))
            AuflVorlage->Mode = _ttoi(EFNodeReader::GetValue(ccc).c_str());
          else
          if (!_tcscmp(ccc->GetName().c_str(),_TEXT("Range")))
            AuflVorlage->Bereich = _ttoi(EFNodeReader::GetValue(ccc).c_str());
          else
          if (!_tcscmp(ccc->GetName().c_str(),_TEXT("Append")))
            AuflVorlage->Anh = _ttoi(EFNodeReader::GetValue(ccc).c_str());
          else
          if (!_tcscmp(ccc->GetName().c_str(),_TEXT("ParamList")))
          {
            EFXmlNode *cccc = ccc->GetChildren();
            while (cccc)
            {
              AuflVorlage->ParamsList.CreateAkEint(EFXmlDataConverter::GetIndex(cccc->GetName())-1);
              AuflVorlage->ParamsList->SetStr(EFNodeReader::GetValue(cccc));
              cccc = cccc->GetNext();
            }
          }
          ccc = ccc->GetNext();
        }
        cc = cc->GetNext();
      }
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Editor")))
    {
      while (cc)
      {
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("FractionDisplay")))
          BildschOpt.SetBildschBruch(_ttoi(EFNodeReader::GetValue(cc).c_str()));
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Brackets")))
          BildschOpt.SetBildschKlammer(_ttoi(EFNodeReader::GetValue(cc).c_str()));
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("HideDiags")))
          BildschOpt.SetHideDiags(_ttoi(EFNodeReader::GetValue(cc).c_str()));
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("MarkLabels")))
          BildschOpt.SetMarkLabels(_ttoi(EFNodeReader::GetValue(cc).c_str()));
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Zoom")))
          Zoom = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("PageLandscape")))
          SeiteQuer = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Font")))
          EFNodeReader::ReadFont(Editor,cc);
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("ScrollLine")))
          ScrollLine = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("ScrollPosX")))
          ScrollPosX = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("ScrollPoxY")))
          ScrollPosY = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("SelStart")))
          EditSelAnf = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("SelEnd")))
          EditSelEnd = _ttoi(EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("ApproximationData")))
    {
      while (cc)
      {
        if (!wxStrncmp(cc->GetName().c_str(),_TEXT("XStr"),4))
          NaehData.XStr[EFXmlDataConverter::GetIndex(cc->GetName())] = EFNodeReader::GetValue(cc).c_str();
        else
        if (!wxStrncmp(cc->GetName().c_str(),_TEXT("YStr"),4))
          NaehData.YStr[EFXmlDataConverter::GetIndex(cc->GetName())] = EFNodeReader::GetValue(cc).c_str();
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("ErgStr")))
          NaehData.ErgStr = EFNodeReader::GetValue(cc).c_str();
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("ParamStr")))
          NaehData.ParamStr = EFNodeReader::GetValue(cc).c_str();
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Options")) ||
            !_tcscmp(cc->GetName().c_str(),_TEXT("CurGraph")))
          NaehData.m_CurGraph = _ttoi(EFNodeReader::GetValue(cc).c_str()) > 0;
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("NumPts")))
          NaehData.m_NumPts = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Mode")))
          NaehData.Mode = _ttoi(EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }
    }
    else
    if (!wxStrncmp(child->GetName().c_str(),_TEXT("Diagram"),7))
    {
      DiagDataArr->CreateAkEint(EFXmlDataConverter::GetIndex(child->GetName().c_str())-1);
      (*DiagDataArr)->XmlLaden(child);
    }
    else
    if (!wxStrncmp(child->GetName().c_str(),_TEXT("Picture"),7))
    {
      PictureArr->CreateAkEint(EFXmlDataConverter::GetIndex(child->GetName().c_str())-1);
      (*PictureArr)->XmlLaden(child);
    }
    else
    if (!wxStrncmp(child->GetName().c_str(),_TEXT("Equations"),9))
    {
      EquationsArr->CreateAkEint(EFXmlDataConverter::GetIndex(child->GetName().c_str())-1);
      (*EquationsArr)->XmlLaden(child);
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Submodules")))
      while (cc)
      {
        m_submodules.CreateAkEint(EFXmlDataConverter::GetIndex(cc->GetName())-1);
        m_submodules.SetAkStr(EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("OpenPanels")))
      while (cc)
      {
        m_OpenPanels.CreateAkEint(EFXmlDataConverter::GetIndex(cc->GetName().c_str())-1);
        m_OpenPanels.GetAkEint() = _ttoi(EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }

    child = child->GetNext();
  }
}

// ----------------------------------------------------------------------------
// PictureObj
// ----------------------------------------------------------------------------

void PictureObj::XmlSpeichern(EFXmlNode *node)
{
  EFNodeCreator::ItemOut(_TEXT("WinXPos"),WinXPos,node);
  EFNodeCreator::ItemOut(_TEXT("WinYPos"),WinYPos,node);
  EFNodeCreator::ItemOut(_TEXT("WinWidth"),WinWidth,node);
  EFNodeCreator::ItemOut(_TEXT("WinHeight"),WinHeight,node);
  EFNodeCreator::ItemOut(_TEXT("Type"),TypeStr.c_str(),node);
  EFNodeCreator::ItemOut(_TEXT("Size"),Size,node);
  StringArr *result = new StringArr;
  result->NeuAkEint();
  Base64::encode((const char *)Data,Size,result->GetAkEint());
  EFNodeCreator::CDATAItemOut(_TEXT("DataBase64"),result,node);
  delete[] Data;
  Data = 0;
  Size = 0;
}

void PictureObj::XmlLaden(EFXmlNode *node)
{
  EFXmlNode *child = node->GetChildren();
  while (child)
  {
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinXPos")))
      WinXPos = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinYPos")))
      WinYPos = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinWidth")))
      WinWidth = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinHeight")))
      WinHeight = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Type")))
      TypeStr = EFNodeReader::GetValue(child).c_str();
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Size")))
      Size = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Data")))
    {
      StringArr *tmp1 = EFNodeReader::GetCDATA(child);
      if (tmp1)
      {
        string tmp;
        EFXmlDataConverter::StringFromStringArr(tmp1,tmp);
        delete tmp1;
        Data = 0;
        Data = EFXmlDataConverter::StringToData(Data,tmp.length(),tmp.c_str());
      }
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("DataBase64")))
    {
      StringArr *tmp = EFNodeReader::GetCDATA(child);
      if (tmp)
      {
        Data = new char[EFXmlDataConverter::GetDataLength(tmp)];
        EFXmlDataConverter::DecodeDataBase64(tmp,Data);
        delete tmp;
      }
    }
    child = child->GetNext();
  }
}

// ----------------------------------------------------------------------------
// LabelArrObj
// ----------------------------------------------------------------------------

void LabelArrObj::XmlWrite(EFXmlNode *node,LabelArrObj *old)
{
  Begin();
  if (old)
  {
    old->Begin();
    if (GetMaxEintrag() != old->GetMaxEintrag())
      old = 0;
  }
  if (GetMaxEintrag() >= 0)
    do
    {
      MGPStr tmp(_TEXT("Label"));
      tmp << AkEintNummer()+1;
      EFXmlNode *chnode = new EFXmlNode(0,EFXmlNode::ELEMENT_NODE,
                                        tmp.c_str(),_TEXT(""),0,0);
      if (old)
        old->NextEint();
      MGPStr str(GetAkEint().Label);
      EFXmlDataConverter::ConvertForXml(str);
      MGPStr *strold = old ? new MGPStr((*old)->Label) : 0;
      if (strold)
         EFXmlDataConverter::ConvertForXml(*strold);
      EFNodeCreator::DiffItemOut(_TEXT("Label"),str.c_str(),chnode,old ? strold->c_str() : 0);
      delete strold;
      EFNodeCreator::DiffItemOut(_TEXT("Font"),GetAkEint().Font,chnode,old ? &(*old)->Font : 0);
      EFNodeCreator::DiffItemOut(_TEXT("Flags"),GetAkEint().Flags,chnode,old ? &(*old)->Flags : 0);
      EFNodeCreator::DiffItemOut(_TEXT("XPos"),GetAkEint().XPos,chnode,old ? &(*old)->XPos : 0);
      EFNodeCreator::DiffItemOut(_TEXT("YPos"),GetAkEint().YPos,chnode,old ? &(*old)->YPos : 0);
      EFNodeCreator::DiffItemOut(_TEXT("XOffs"),GetAkEint().XOffs,chnode,old ? &(*old)->XOffs : 0);
      EFNodeCreator::DiffItemOut(_TEXT("YOffs"),GetAkEint().YOffs,chnode,old ? &(*old)->YOffs : 0);
      if (chnode->GetChildren())
        node->AddChild(chnode);
      else
        delete chnode;
    }
    while (NextEint());
}

// ----------------------------------------------------------------------------
// DiagDataBaseObj
// ----------------------------------------------------------------------------

void DiagDataBaseObj::XmlSpeichern(EFXmlNode *node,DiagDataBaseObj *old)
{
  int f;

  EFXmlNode *chnode;
  EFNodeCreator::DiffItemOut(_TEXT("Points"),Punkte,node,old ? &old->Punkte : 0);
  EFNodeCreator::DiffItemOut(_TEXT("AkGraph"),AkGraph,node,old ? &old->AkGraph : 0);
  EFNodeCreator::DiffItemOut(_TEXT("Debug"),Debug,node,old ? &old->Debug : 0);
  EFNodeCreator::DiffItemOut(_TEXT("AutoUpdate"),AutoUpdate,node,old ? &old->AutoUpdate : 0);
  EFNodeCreator::DiffItemOut(_TEXT("PostProc"),PostProc,node,old ? &old->PostProc : 0);
  EFNodeCreator::DiffItemOut(_TEXT("PostProcTerm"),PostProcTerm.c_str(),node,old ? old->PostProcTerm.c_str() : 0);
  EFNodeCreator::DiffItemOut(_TEXT("AutoMode"),AutoMode,node,old ? &old->AutoMode : 0);
  for (f = 0;f < 2;f++)
  {
    MGPStr tmp(_TEXT("Plot"));
    tmp << f;
    chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                           tmp.c_str(),_TEXT(""),0,0);
    DarstDataObj &DD = DarstData[f];
    DarstDataObj *DDO = (old ? &old->DarstData[f] : 0);
    EFNodeCreator::DiffItemOut(_TEXT("Mode"),DD.Mode,chnode,old ? &DDO->Mode : 0);
    EFNodeCreator::DiffItemOut(_TEXT("Title"),DD.Titel.c_str(),chnode,old ? DDO->Titel.c_str() : 0);
    EFNodeCreator::DiffItemOut(_TEXT("Conversion"),DD.Conversion,chnode,old ? &DDO->Conversion : 0);
    EFNodeCreator::DiffItemOut(_TEXT("NumDigits"),DD.NumDigits,chnode,old ? &DDO->NumDigits : 0);
    EFNodeCreator::DiffItemOut(_TEXT("SameXAndYScale"),DD.m_SameXAndYScaling,chnode,old ? &DDO->m_SameXAndYScaling : 0);
    for (size_t i = 0;i < 2;i++)
    {
      AxisDataArr &ada(i ? DD.YAxis : DD.XAxis);
      AxisDataArr *oada = (old ? (i ? &DDO->YAxis : &DDO->XAxis) : 0);
      ada.Begin();
      if (oada)
        oada->Begin();
      do
      {
        MGPStr tmp(i ? _TEXT("YAxis") :  _TEXT("XAxis"));
        tmp << ada.AkEintNummer();
        EFXmlNode *cnode1 = new EFXmlNode(0,EFXmlNode::ELEMENT_NODE,
                                          tmp.c_str(),_TEXT(""),0,0);
        AxisData &ad = ada.GetAkEint();
        AxisData *oad = (old && (ada.GetMaxEintrag() == oada->GetMaxEintrag()) ? &oada->GetAkEint() : 0);
        EFNodeCreator::DiffItemOut(_TEXT("Unit"),ad.Label.c_str(),cnode1,old ? oad->Label.c_str() : 0);
        EFNodeCreator::DiffItemOut(_TEXT("Min"),ad.Min,cnode1,old ? &oad->Min : 0);
        EFNodeCreator::DiffItemOut(_TEXT("Max"),ad.Max,cnode1,old ? &oad->Max : 0);
        EFNodeCreator::DiffItemOut(_TEXT("Options"),ad.GetOptions(),cnode1,old ? &oad->GetOptions() : 0);
        if (old)
          oada->NextEint();
        if (cnode1->GetChildren())
          chnode->AddChild(cnode1);
        else
          delete cnode1;
      }
      while (i && ada.NextEint());
    }
  }
  chnode = new EFXmlNode(node,EFXmlNode::ELEMENT_NODE,
                         _TEXT("Hi"),_TEXT(""),0,0);
  EFNodeCreator::DiffItemOut(_TEXT("Label"),Ma.Bezeichner.c_str(),chnode,old ? old->Ma.Bezeichner.c_str() : 0);
  EFNodeCreator::DiffItemOut(_TEXT("Unit"),Ma.Einheit.c_str(),chnode,old ? old->Ma.Einheit.c_str() : 0);
  EFNodeCreator::DiffItemOut(_TEXT("Start"),Ma.Anf,chnode,old ? &old->Ma.Anf : 0);
  EFNodeCreator::DiffItemOut(_TEXT("End"),Ma.Ende,chnode,old ? &old->Ma.Ende : 0);
  EFNodeCreator::DiffItemOut(_TEXT("Step"),Ma.Step,chnode,old ? &old->Ma.Step : 0);
  EFNodeCreator::DiffItemOut(_TEXT("ParList"),Ma.GetParListe(),chnode,old ? old->Ma.GetParListe() : 0);
  EFNodeCreator::DiffItemOut(_TEXT("Mode"),Ma.Mode,chnode,old ? &old->Ma.Mode : 0);
  EFNodeCreator::DiffItemOut(_TEXT("CXMin"),CXMin,node,old ? &old->CXMin : 0);
  EFNodeCreator::DiffItemOut(_TEXT("CXMax"),CXMax,node,old ? &old->CXMax : 0);
  EFNodeCreator::DiffItemOut(_TEXT("XMode"),XMode,node,old ? &old->XMode : 0);
  EFNodeCreator::DiffItemOut(_TEXT("XLog"),XLog,node,old ? &old->XLog : 0);
  EFNodeCreator::DiffItemOut(_TEXT("FontAxis"),Achse,node,old ? &old->Achse : 0);
  EFNodeCreator::DiffItemOut(_TEXT("FontAxisLabels"),AchsLabels,node,old ? &old->AchsLabels : 0);
  EFNodeCreator::DiffItemOut(_TEXT("FontTitle"),Titel,node,old ? &old->Titel : 0);
  EFNodeCreator::DiffItemOut(_TEXT("WinXPos"),WinXPos,node,old ? &old->WinXPos : 0);
  EFNodeCreator::DiffItemOut(_TEXT("WinYPos"),WinYPos,node,old ? &old->WinYPos : 0);
  EFNodeCreator::DiffItemOut(_TEXT("WinWidth"),WinWidth,node,old ? &old->WinWidth : 0);
  EFNodeCreator::DiffItemOut(_TEXT("WinHeight"),WinHeight,node,old ? &old->WinHeight : 0);
  EFNodeCreator::DiffItemOut(_TEXT("Colour"),m_Colour,node,old ? &old->m_Colour : 0);
}

void DiagDataBaseObj::XmlWriteLegend(EFXmlNode *node,DiagDataBaseObj *old)
{
  EFNodeCreator::DiffItemOut(_TEXT("FontLegend"),*(FontType *)&Legende,node,old ? (FontType *)&old->Legende : 0);
  EFNodeCreator::DiffItemOut(_TEXT("LegendFrame"),Legende.Frame,node,old ? &old->Legende.Frame : 0);
  EFNodeCreator::DiffItemOut(_TEXT("LegX"),LegX,node,old ? &old->LegX : 0);
  EFNodeCreator::DiffItemOut(_TEXT("LegY"),LegY,node,old ? &old->LegY : 0);
}

void DiagDataBaseObj::XmlLoadLegend(EFXmlNode *node)
{
  EFXmlNode *child = node->GetChildren();
  while (child)
  {
    if (!_tcscmp(child->GetName().c_str(),_TEXT("FontLegend")))
      EFNodeReader::ReadFont(Legende,child);
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("LegendFrame")))
      Legende.Frame = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("LegX")))
      LegX = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("LegY")))
      LegY = _ttoi(EFNodeReader::GetValue(child).c_str());
    child = child->GetNext();
  }
}

// ----------------------------------------------------------------------------
// DiagDataObj
// ----------------------------------------------------------------------------

void DiagDataObj::XmlSpeichern(EFXmlNode *node,DiagDataObj *old)
{
  DiagDataBaseObj::XmlSpeichern(node,old);
  DiagDataBaseObj::XmlWriteLegend(node,old);
  YDaten.XmlWriteGraphs(node);
  Labels.XmlWrite(node);
}

void DiagDataObj::XmlLaden(EFXmlNode *node)
{
  EFXmlNode *child = node->GetChildren();
  while (child)
  {
    EFXmlNode *cc = child->GetChildren();
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Points")))
      Punkte = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("AkGraph")))
      AkGraph = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Debug")))
      Debug = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("AutoUpdate")))
      AutoUpdate = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("PostProc")))
      PostProc = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("PostProcTerm")))
      PostProcTerm = EFNodeReader::GetValue(child).c_str();
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("AutoMode")))
      AutoMode = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("AutoMode")))
      AutoMode = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Hi")))
    {
      while (cc)
      {
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Label")))
          Ma.Bezeichner = EFNodeReader::GetValue(cc).c_str();
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Unit")))
          Ma.Einheit = EFNodeReader::GetValue(cc).c_str();
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Start")))
          Ma.Anf = Eval(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("End")))
          Ma.Ende = Eval(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Step")))
          Ma.Step = Eval(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("ParList")))
          Ma.SetParListe(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Mode")))
          Ma.Mode = _ttoi(EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("CXMin")))
      CXMin = Eval(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("CXMax")))
      CXMax = Eval(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("XMode")))
      XMode = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("XLog")))
      XLog = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("FontLegend")))
      EFNodeReader::ReadFont(Legende,child);
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("LegendFrame")))
      Legende.Frame = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("FontAxis")))
      EFNodeReader::ReadFont(Achse,child);
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("FontAxisLabels")))
      EFNodeReader::ReadFont(AchsLabels,child);
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("FontTitle")))
      EFNodeReader::ReadFont(Titel,child);
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinXPos")))
      WinXPos = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinYPos")))
      WinYPos = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinWidth")))
      WinWidth = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinHeight")))
      WinHeight = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Colour")))
      m_Colour = _tcstoul(EFNodeReader::GetValue(child).c_str(),0,10);
    else
    if (!wxStrncmp(child->GetName().c_str(),_TEXT("Graph"),5))
      YDaten.XmlLoadGraph(child,EFXmlDataConverter::GetIndex(child->GetName().c_str()));
    else
    if (!wxStrncmp(child->GetName().c_str(),_TEXT("Label"),5))
    {
      Labels.CreateAkEint(EFXmlDataConverter::GetIndex(child->GetName().c_str())-1);
      while (cc)
      {
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Label")))
        {
          MGPStr str(EFNodeReader::GetValue(cc).c_str());
          EFXmlDataConverter::ConvertForEF(str);
          Labels->Label = str;
        }
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Font")))
          EFNodeReader::ReadFont(Labels->Font,cc);
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Flags")))
          Labels->Flags = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("XPos")))
          Labels->XPos = Eval(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("YPos")))
          Labels->YPos = Eval(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("XOffs")))
          Labels->XOffs = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("YOffs")))
          Labels->YOffs = _ttoi(EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }
    }
    else
    if (!wxStrncmp(child->GetName().c_str(),_TEXT("Plot"),4))
    {
      int index = EFXmlDataConverter::GetIndex(child->GetName().c_str());
      while (cc)
      {
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Conversion")))
          DarstData[index].Conversion = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("DisplayMode")))
        {
          DarstData[index].Conversion = _ttoi(EFNodeReader::GetValue(cc).c_str());
          DarstData[index].YAxis.Begin();
          DarstData[index].YAxis->SetLog(DarstData[index].Conversion == DarstDataObj::logY);
        }
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("XMode")))
        {
          DarstData[index].XAxis.Begin();
          DarstData[index].XAxis->SetLog(_ttoi(EFNodeReader::GetValue(cc).c_str()));
        }
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Mode")))
          DarstData[index].Mode = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("Title")))
          DarstData[index].Titel = EFNodeReader::GetValue(cc).c_str();
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("XUnit")))
        {
          DarstData[index].XAxis.Begin();
          DarstData[index].XAxis->Label = EFNodeReader::GetValue(cc).c_str();
        }
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("XMin")))
        {
          DarstData[index].XAxis.Begin();
          DarstData[index].XAxis->Min = Eval(EFNodeReader::GetValue(cc).c_str());
        }
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("XMax")))
        {
          DarstData[index].XAxis.Begin();
          DarstData[index].XAxis->Max = Eval(EFNodeReader::GetValue(cc).c_str());
        }
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("YUnit")))
        {
          DarstData[index].YAxis.Begin();
          DarstData[index].YAxis->Label = EFNodeReader::GetValue(cc).c_str();
        }
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("YMin")))
        {
          DarstData[index].YAxis.Begin();
          DarstData[index].YAxis->Min = Eval(EFNodeReader::GetValue(cc).c_str());
        }
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("YMax")))
        {
          DarstData[index].YAxis.Begin();
          DarstData[index].YAxis->Max = Eval(EFNodeReader::GetValue(cc).c_str());
        }
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("NumDigits")))
          DarstData[index].NumDigits = _ttoi(EFNodeReader::GetValue(cc).c_str());
        else
        if (!wxStrncmp(cc->GetName().c_str(),_TEXT("XAxis"),5) ||
            !wxStrncmp(cc->GetName().c_str(),_TEXT("YAxis"),5))
        {
          int ind = EFXmlDataConverter::GetIndex(cc->GetName().c_str());
          AxisDataArr &ad = (cc->GetName()[0] == 'X' ? DarstData[index].XAxis :
                                                       DarstData[index].YAxis);
          ad.SetAkEint(ind);
          EFXmlNode *ccc = cc->GetChildren();
          while (ccc)
          {
            if (!_tcscmp(ccc->GetName().c_str(),_TEXT("Unit")))
              ad->Label = EFNodeReader::GetValue(ccc).c_str();
            else
            if (!_tcscmp(ccc->GetName().c_str(),_TEXT("Min")))
              ad->Min = Eval(EFNodeReader::GetValue(ccc).c_str());
            else
            if (!_tcscmp(ccc->GetName().c_str(),_TEXT("Max")))
              ad->Max = Eval(EFNodeReader::GetValue(ccc).c_str());
            else
            if (!_tcscmp(ccc->GetName().c_str(),_TEXT("Options")))
              ad->SetOptions(_ttoi(EFNodeReader::GetValue(ccc).c_str()));
            ccc = ccc->GetNext();
          }
        }
        else
        if (!_tcscmp(cc->GetName().c_str(),_TEXT("SameXAndYScale")))
          DarstData[index].m_SameXAndYScaling = _ttoi(EFNodeReader::GetValue(cc).c_str());
        cc = cc->GetNext();
      }
    }
    child = child->GetNext();
  }
  XmlLoadLegend(node);
}

// ----------------------------------------------------------------------------
// EquationsObj
// ----------------------------------------------------------------------------

void EquationsObj::XmlSpeichern(EFXmlNode *node)
{
  EFNodeCreator::ItemOut(_TEXT("WinXPos"),WinXPos,node);
  EFNodeCreator::ItemOut(_TEXT("WinYPos"),WinYPos,node);
  EFNodeCreator::ItemOut(_TEXT("WinWidth"),WinWidth,node);
  EFNodeCreator::ItemOut(_TEXT("WinHeight"),WinHeight,node);
  EFNodeCreator::ItemOut(_TEXT("Font"),Font,node);
  EFNodeCreator::ItemOut(_TEXT("Options"),BildschOpt.Get(),node);
  EFNodeCreator::ItemOut(_TEXT("ScrollPosX"),ScrollPosX,node);
  EFNodeCreator::ItemOut(_TEXT("ScrollPosY"),ScrollPosY,node);
  EFNodeCreator::ItemOut(_TEXT("SelStart"),SelAnf,node);
  EFNodeCreator::ItemOut(_TEXT("SelEnd"),SelEnd,node);
  EFNodeCreator::ItemOut(_TEXT("Lines"),GetNumLines(Equations.c_str()),node);
  EFXmlDataConverter::ConvertForXml(Equations);
  EFNodeCreator::ItemOut(_TEXT("Equations"),Equations.c_str(),node);
  EFXmlDataConverter::ConvertForEF(Equations);
}

void EquationsObj::XmlLaden(EFXmlNode *node)
{
  EFXmlNode *child = node->GetChildren();
  while (child)
  {
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinXPos")))
      WinXPos = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinYPos")))
      WinYPos = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinWidth")))
      WinWidth = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("WinHeight")))
      WinHeight = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Font")))
      EFNodeReader::ReadFont(Font,child);
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Options")))
      BildschOpt.Set(_ttoi(EFNodeReader::GetValue(child).c_str()));
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("ScrollPosX")))
      ScrollPosX = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("ScrollPosY")))
      ScrollPosY = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("SelStart")))
      SelAnf = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("SelEnd")))
      SelEnd = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Equations")))
    {
      MGPStr tmp(EFNodeReader::GetValue(child).c_str());
      EFXmlDataConverter::ConvertForEF(tmp);
      Equations = tmp;
    }
    child = child->GetNext();
  }
}

// ----------------------------------------------------------------------------
// GraphDataObj
// ----------------------------------------------------------------------------

bool GraphDataObj::XmlLoadGraph(EFXmlNode *node)
{
  EFXmlNode *child = node->GetChildren();
  while (child)
  {
    if (!_tcscmp(child->GetName().c_str(),_TEXT("FileName")))
    {
      _tcscpy(FileString,EFNodeReader::GetValue(child).c_str());
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Points")))
    {
      SetPunkte(_tcstoul(EFNodeReader::GetValue(child).c_str(),0,10));
    }
    else
    if (!wxStrncmp(child->GetName().c_str(),_TEXT("YAxisIndex"),10) ||
        !wxStrncmp(child->GetName().c_str(),_TEXT("XAxisIndex"),10))
    {
      int index = EFXmlDataConverter::GetIndex(child->GetName().c_str());
      if (!wxStrncmp(child->GetName().c_str(),_TEXT("YAxisIndex"),10))
        YAxisIndex[index] = _ttoi(EFNodeReader::GetValue(child).c_str());
      else
        XAxisIndex[index] = _ttoi(EFNodeReader::GetValue(child).c_str());
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Name")))
    {
      if (!FileString[0] ||
          !FuString[0])
        _tcscpy(FuString,EFNodeReader::GetValue(child).c_str());
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("XValues")))
    {
      StringArr *tmp1 = EFNodeReader::GetCDATA(child);
      if (tmp1)
      {
        string tmp;
        EFXmlDataConverter::StringFromStringArr(tmp1,tmp);
        delete tmp1;
        size_t len =tmp.length()/sizeof(GetX(0))/2;
        SetPunkte(len);
        EFXmlDataConverter::StringToData((char *)&GetX(0),tmp.length(),
                                         tmp.c_str());
      }
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("XValuesBase64")))
    {
      StringArr *tmp = EFNodeReader::GetCDATA(child);
      if (tmp)
      {
        SetPunkte(EFXmlDataConverter::GetDataLength(tmp)/sizeof(GetX(0)));
        EFXmlDataConverter::DecodeDataBase64(tmp,(void *)&GetX(0));
        delete tmp;
      }
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("YValues")))
    {
      StringArr *tmp1 = EFNodeReader::GetCDATA(child);
      if (tmp1)
      {
        string tmp;
        EFXmlDataConverter::StringFromStringArr(tmp1,tmp);
        delete tmp1;
        size_t len = tmp.length()/sizeof(GetY(0))/2;
        SetPunkte(len);
        EFXmlDataConverter::StringToData((char *)&GetY(0),tmp.length(),
                     tmp.c_str());
      }
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("YValuesBase64")))
    {
      StringArr *tmp = EFNodeReader::GetCDATA(child);
      if (tmp)
      {
        SetPunkte(EFXmlDataConverter::GetDataLength(tmp)/sizeof(GetY(0)));
        EFXmlDataConverter::DecodeDataBase64(tmp,(void *)&GetY(0));
        delete tmp;
      }
    }
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Symbol")))
      GraphSymbol = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Options")))
      GraphOptionen = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Style")))
      ZeichenArt = _ttoi(EFNodeReader::GetValue(child).c_str());
    else
    if (!_tcscmp(child->GetName().c_str(),_TEXT("Color")))
      Farbe = _ttol(EFNodeReader::GetValue(child).c_str());
    child = child->GetNext();
  }
  return true;
}

bool GraphDataObj::XmlWriteGraph(EFXmlNode *node,GraphDataObj *old) const
{
  EFNodeCreator::DiffItemOut(_TEXT("Points"),GetPunkte(),node,old ? &old->GetPunkte() : 0);
  EFNodeCreator::DiffItemOut(_TEXT("Name"),FuString,node,old ? old->FuString : 0);
  EFNodeCreator::DiffItemOut(_TEXT("YAxisIndex0"),YAxisIndex[0],node,old ? &old->YAxisIndex[0] : 0);
  EFNodeCreator::DiffItemOut(_TEXT("YAxisIndex1"),YAxisIndex[1],node,old ? &old->YAxisIndex[1] : 0);
  if (FileString[0])
    EFNodeCreator::DiffItemOut(_TEXT("FileName"),FileString,node,old ? old->FileString : 0);
  else
  if (!old || HasDifferentData(*old))
  {
    StringArr *result = new StringArr;
    result->NeuAkEint();
    Base64::encode((const char *)&GetX(0),
                   GetPunkte()*
                     sizeof(GetX(0)),
                   result->GetAkEint());
    EFNodeCreator::CDATAItemOut(_TEXT("XValuesBase64"),result,node);
    result = new StringArr;
    result->NeuAkEint();
    Base64::encode((const char *)&GetY(0),
                   GetPunkte()*
                     sizeof(GetY(0)),
                   result->GetAkEint());
    EFNodeCreator::CDATAItemOut(_TEXT("YValuesBase64"),result,node);
  }
  EFNodeCreator::DiffItemOut(_TEXT("Symbol"),GraphSymbol,node,old ? &old->GraphSymbol : 0);
  EFNodeCreator::DiffItemOut(_TEXT("Options"),GraphOptionen,node,old ? &old->GraphOptionen : 0);
  EFNodeCreator::DiffItemOut(_TEXT("Style"),ZeichenArt,node,old ? &old->ZeichenArt : 0);
  EFNodeCreator::DiffItemOut(_TEXT("Color"),Farbe,node,old ? &old->Farbe : 0);
  return true;
}

// ----------------------------------------------------------------------------
// GraphManagerObj
// ----------------------------------------------------------------------------

bool GraphManagerObj::XmlLoadGraph(EFXmlNode *node,int index)
{
  DatFeld->CreateAkEint(index-1);
  (*DatFeld)->XmlLoadGraph(node);
  if (_tcslen((*DatFeld)->FileString) &&
      !LoadGraph((*DatFeld)->FileString,0,false,false,true))
  {
    (*DatFeld)->SetPunkte(2);
    (*DatFeld)->SetX(0,0.0);
    (*DatFeld)->SetY(0,Cpx(0.0,0.0));
    (*DatFeld)->SetX(1,0.0);
    (*DatFeld)->SetY(1,Cpx(0.0,0.0));
    _tcscpy((*DatFeld)->FuString,_("FILE NOT FOUND"));
  }
  return true;
}

bool GraphManagerObj::XmlWriteGraphs(EFXmlNode *node)
{
  DatFeld->Begin();
  if (DatFeld->GetMaxEintrag() >= 0)
    do
    {
      MGPStr tmp(_TEXT("Graph"));
      tmp << DatFeld->AkEintNummer()+1;
      EFXmlNode *cc = new EFXmlNode(0,EFXmlNode::ELEMENT_NODE,
                                    tmp.c_str(),_TEXT(""),0,0);
      (*DatFeld)->XmlWriteGraph(cc);
      if (cc->GetChildren())
        node->AddChild(cc);
      else
        delete cc;
    }
    while (DatFeld->NextEint());
  return true;
}

