// Version 6.06 10.01.2010 Markus Greither
// Copyright (c) 1991-2010 Markus Greither

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/*! \file tstsolv.cpp
    \brief Source code for functional and regression test program
*/

#ifdef __DMC__
#include <typeinfo.h>
#endif
#include <malloc.h>
#include <fstream>
#include <iostream>
#include <efdefs.h>
#include <complex>
#include <limits>
#include <set>
#include <functional>
#if defined(__MINGW32__)
//#include <wx/msw/winundef.h>
#endif
#include <string.h>
#ifdef __BORLANDC__
#include <excpt.h>
#endif
#include <time.h>

using namespace std;

#define __MGDEBUG__
#include "mg.h"
#include "cpxpsr.h"
#include "diagtype.h"
#include "efgemein.h"
#include "efmcxio.h"
#include "efsolver.h"
#include "efkrzn.h"


// Derivative of Solver with debug output on console
class TstSolverObj : public SolverObj
{
 public:
  TstSolverObj(FunctionArrObj *Fu,const _TCHAR *SStr,const _TCHAR *EStr,
            const _TCHAR *ZStr,int Mode,ParserObj *P,MGPStr *DO = 0)
   : SolverObj(Fu,SStr,EStr,ZStr,Mode,P,DO)
  {
  }
  virtual void DebugString(const _TCHAR *str);
};

void TstSolverObj::DebugString(const _TCHAR *str)
{
  if (DbgOut)
    cout << str << _TEXT(" ");
};

// Function parser test code
bool TestParser(ostream &str)
{
  bool okres = true;
  str << _TEXT("Function Parser Test") << endl << endl;
  // Setup data object
  EFDaten Daten(NumListFormat(_TEXT(";"),_TEXT(","),0));
  // Load test functions
  Daten.LoadTextDatei(_TEXT("tstsolv/tstpars.txt"));
  // Setup function parser object
  ParserCpxObj P(NumListFormat(_TEXT(";"),_TEXT(","),0));
  // Initialize the functions
  P.InitFunktions(Daten.Funktion);
  
  // Go through all found functions
  Daten.Funktion->Begin();
  while (Daten.Funktion->NextEint())
  {
    // Look if there are parameters for the current function
    int tmp = P.GetParamsAnfPos((*Daten.Funktion)->Namen.c_str());
    // Get parameter list manually, as not performed by LoadTextDatei
    if (tmp)
    {
      (*Daten.Funktion)->ParamStr.SetStr(&(*Daten.Funktion)->Namen[tmp]);
      (*Daten.Funktion)->Namen.ShortStr(tmp);
    }
    // Check if this is a comment line, if yes ignore it
    MGPStr &Zeile = (*Daten.Funktion)->Namen;
    if ((Zeile[0] <= ' ') && (Zeile[0] != 0))
    {
      str << Zeile << endl;
      continue;
    }
    // Check for the existence of a double quote, as this indicates the start point for the expected result
    const _TCHAR *tst = _tcsstr((*Daten.Funktion)->Term.c_str(),_TEXT("\""));
    if (tst)
    {
      // Skip double quote
      tst++;
      double Res1;
      Cpx Res2,y;
      unsigned int p = 0;
      ParserObj::FehlerRec Fehler;
      // Read the expected result value
      Res1 = P.Zahl(tst,p,Fehler.Art);
      // Perform the execution of current term
      P.FunktionsBerechnung(Res2,0,y,Fehler,0,true,0);
      // Check if result fits expected result
      if ((Res1 != abs(Res2)) || (Fehler.Art != ZahlFormatObj::NoError))
      {
        str << "Error! Error code: " << Fehler.Art << " Result: " <<
                abs(Res2) << " " << (*Daten.Funktion)->Term << endl;
        okres = false;
      }
      else
        str << "OK: " << (*Daten.Funktion)->Term << " = " << abs(Res1) << endl;
    }
  }
  return okres;
}

// Function solver test code
bool TestSolver(ostream &str,bool output = true)
{
  bool okres = true;
  if (output)
    str << endl << _TEXT("Function Solver Test") << endl << endl;
  // Setup data object
  EFDaten Daten(NumListFormat(_TEXT(";"),_TEXT(","),0));
  // Load functions to be solved
  Daten.LoadTextDatei(_TEXT("tstsolv/tstsolv.txt"));
  // Setup parser object
  ParserDoubleObj P(NumListFormat(_TEXT(";"),_TEXT(","),0));
  MGPStr DebugOut;
  
  // Go through all functions
  Daten.Funktion->Begin();
  while (Daten.Funktion->NextEint())
  {
    MGPStr &Zeile = (*Daten.Funktion)->Namen;
    // Look for comment lines and ignore them
    if ((Zeile[0] <= ' ') && (Zeile[0] != 0))
    {
      if (output)
        str << Zeile << endl;
      continue;
    }
    // Check for the existence of a double quote, as this indicates the start point for the expected result
    const _TCHAR *tst = _tcsstr((*Daten.Funktion)->Term.c_str(),_TEXT("\""));
    if (tst)
    {
      // Get the expected results string
      MGPStr ResStr(&tst[1]);
      // Find equal sign in results string
      int p = P.IstGleichPos(ResStr.c_str());
      int p1 = p;
      // Find end of searched result label
      p1--;
      while (p1 && (ResStr[p1] == ' '))
        p1--;
      // Find the beginning or result term
      p++;
      while (ResStr[p] == ' ')
        p++;
      // Get searched result label
      MGPStr SuchStr;
      SuchStr.AddNChar(ResStr,p1+1);
      // Create temporary functions object
      FunctionArrObj Fu;
      Fu.NeuAkEint();
      Fu->Term.SetStr((*Daten.Funktion)->Term);
      Fu->Namen.SetStr((*Daten.Funktion)->Namen);
      // Setup solver and perform function solving
      TstSolverObj Solver(&Fu,SuchStr.c_str(),_TEXT("A;B;C;D;E"),_TEXT(""),SolverObj::SolverKlKuerzen,&P);
      Solver.StartWerte();
      while (Solver.ElimLabel());
      Fu.SetAkEint(1);
      if ((Solver.GetFehler() != 0) && output)
      {
        str << "Error occured ! " << Solver.GetFehler() << ResStr << endl;
        okres = false;
      }
      else
      if (_tcscmp(&ResStr[p],Fu->Term.c_str()) && output)
      {
        str << "Result different from expected ! " << ResStr << endl;
        okres = false;
      }
      else
//        continue;
        if (output) 
          cout << "OK: " << (*Daten.Funktion)->Namen << " = " << (*Daten.Funktion)->Term << endl;
//      if (output)
//        str << Fu->Namen << " = " << Fu->Term << endl;
    }
//    else
//      cout << Zeile << " nicht bearbeitet." << endl;
  }
  return okres;
}

// Test code for BitArrObj
void TestBitArrObj()
{
 BitArrObj tst;
 tst.SetBit(2);
 MGPStr str;
 tst.BitString(str);
 cout << "Startwert " << str.c_str() << endl;
 tst.ClearBit(2);
 str.SetStr(_TEXT(""));
 tst.BitString(str);
 cout << "ClearBit(2) " << str.c_str() << endl;
 cout << endl;
 tst.SetBit(2);
 tst.SetBit(4);
 tst.SetBit(1);
 str.SetStr(_TEXT(""));
 tst.BitString(str);
 cout << "SetBit(1,2,4) " << str.c_str() << endl;
 tst.Inc();
 str.SetStr(_TEXT(""));
 tst.BitString(str);
 cout << "Inc() " << str.c_str() << endl;
 tst.Inc();
 str.SetStr(_TEXT(""));
 tst.BitString(str);
 cout << "Inc() " << str.c_str() << endl;
 tst.SetBit(0);
 tst.SetBit(1);
 tst.SetBit(2);
 str.SetStr(_TEXT(""));
 tst.BitString(str);
 cout << "SetBit(0,1,2) " << str.c_str() << endl;
 tst.Inc();
 str.SetStr(_TEXT(""));
 tst.BitString(str);
 cout << "Inc() " << str.c_str() << endl;
 tst.Shl();
 str.SetStr(_TEXT(""));
 tst.BitString(str);
 cout << "Shl() " << str.c_str() << endl;
 tst.Clear();
 str.SetStr(_TEXT(""));
 tst.BitString(str);
 cout << "Clear() " << str.c_str() << endl;
 tst.Inc();
 str.SetStr(_TEXT(""));
 tst.BitString(str);
 cout << "Inc() " << str.c_str() << endl;

 cout << endl;
}

// Testcode ZahlArrObj
void TestZahlArrObj()
{
  ZahlArrObj<int> tst;
  cout << "Set first value\n";
  tst.SetValue(0,5);
  cout << "Set next value\n";
  tst.SetValue(1,6);
  cout << "Set third value\n";
  tst.SetValue(2,6);
  cout << "Inc third value\n";
  tst.IncValue(2);
  cout << "Copy\n";
  ZahlArrObj<int> tmp(tst);
  cout << "AnzInts\n";
  tmp.AnzInts();
  cout << "Output copy\n";
  for (size_t i = 0;i < tmp.AnzInts();i++)
    cout << tmp[i] << endl;
}

int main(int argc,char **argv)
{

#if defined(__DMC__) || defined(__MINGW32__)
  ofstream str("tstsolv/tst.txt",ios::out);
#else
  ostream &str(cout);
#endif
  bool okres = true;
  okres &= TestParser(str);
  time_t before = time(NULL);
  BitArrObj tst;
//    for (int i = 0;i < 10000000;i++)
//      tst.Inc();
  okres &= TestSolver(str);
  for (int i = 0;i < 9;i++)
    okres &= TestSolver(str,false);
  time_t after = time(NULL);
  if (okres)
  cout << endl << "Tests successfully executed!" << endl << endl;
  str << endl << "Tests successfully executed!" << endl << endl;
  str << "Time : " << difftime(after,before) << " s" << endl;
//    TestBitArrObj();
  // Test-Mode zum Testen des Solvers mit speziellen funktionen
  if ((argc > 1) && !strcmp(argv[1],"-s"))
    do
    {
      EFDaten Daten(NumListFormat(_TEXT(";"),_TEXT(","),0));
      Daten.LoadTextDatei(_TEXT("tstsolv/func6.txt"));
      ParserDoubleObj P(NumListFormat(_TEXT(";"),_TEXT(","),0));
      MGPStr DebugOut;
      TstSolverObj Solver(Daten.Funktion,_TEXT("us"),
                          _TEXT("G;Rl;Ri;Rs;xci;xcr;xco;Rli;Rg"),_TEXT(""),
                          SolverObj::SolverTeilen+SolverObj::SolverDebug,&P,&DebugOut);
    //  TstSolverObj Solver(Daten.Funktion,"us","G;Rl;Ri;Rs;Rg;xci;xcr;xco;xce;Aoppos;Aopneg;uldiv","",SolverObj::SolverDebug+SolverObj::SolverKlKuerzen,&P,&DebugOut);
    //  TstSolverObj Solver(Daten.Funktion,"Erg","A;B;C;D;E","",SolverObj::SolverDebug+SolverObj::SolverKlKuerzen,&P,&DebugOut);
      Solver.StartWerte();

      try
      {
        while (Solver.ElimLabel())
          cout << "\n" << Solver.GetErgString() << "\n";
      }
      catch (MGPStr tmp)
      {
        cout << "Fehler :" << tmp.c_str() << endl;
      }
      Daten.Funktion->SetAkEint(1);
      cout << (*Daten.Funktion)->Namen << " = " << (*Daten.Funktion)->Term;
    }
    while (true);
}

